/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef __CK_DEF__
#define __CK_DEF__

#include <BIOS.H>
#include <CONIO.H>

#include "ID_HEADS.H"

/*
=============================================================================

						 GLOBAL CONSTANTS & MACROS

=============================================================================
*/

#define MAXACTORS 100

#define GAMELEVELS 25

#define FULLCLIP 2

#define CONVERT_GLOBAL_TO_TILE(x)  ((x)>>(G_T_SHIFT))
#define CONVERT_TILE_TO_GLOBAL(x)  ((x)<<(G_T_SHIFT))
#define CONVERT_GLOBAL_TO_PIXEL(x) ((x)>>(G_P_SHIFT))
#define CONVERT_PIXEL_TO_GLOBAL(x) ((x)<<(G_P_SHIFT))
#define CONVERT_PIXEL_TO_TILE(x)   ((x)>>(P_T_SHIFT))
#define CONVERT_TILE_TO_PIXEL(x)   ((x)<<(P_T_SHIFT))

#define SPAWN_ADJUST_Y(y, h) (CONVERT_TILE_TO_GLOBAL(y) + (CONVERT_PIXEL_TO_GLOBAL(16-(h))))

#define ARRAYLENGTH(x) (sizeof(x)/sizeof(*(x)))

#define CA_UnmarkGrChunk(num) (grneeded[num] &= ~ca_levelbit)

#define SetPalette(pal) {_ES=FP_SEG(pal); _DX=FP_OFF(pal); _AX=0x1002; geninterrupt(0x10);}
#define SetPaletteEx(pal) {(pal)[16] = bordercolor; SetPalette(pal);}

//HACKS:
#define statusBtnHit gravisButtonDown[3]
#define gravisJumpDown gravisButtonDown[0]
#define gravisPogoDown gravisButtonDown[1]
#define gravisFireDown gravisButtonDown[2]

void RFL_InitAnimList(void);
void CA_FreeGraphics(void);
void CA_SetGrPurge(void);

/*
Note:

The ID software memory manager doesn't care about the different purge levels.
Using PURGE_FIST is identical to using PURGE_LAST.
*/
#define PURGE_FIRST 3
#define PURGE_LAST 1

#define PLATFORMBLOCK 31
#define DIRARROWSTART 91
#define DIRARROWEND   (DIRARROWSTART+arrow_None)

/*
=============================================================================

						   GLOBAL TYPES

=============================================================================
*/

//SDL-style integer types - just to make future SDL ports easier
typedef unsigned int Uint16;
typedef signed int Sint16;
typedef unsigned char Uint8;
typedef signed char Sint8;
typedef unsigned long Uint32;
typedef signed long Sint32;
//Note: only the game code (CK_*.C, K?_*.C) uses these!

//some compile-time checks to make sure the ints have the correct size
#if (sizeof(Uint16) != 2)
#error 'Uint16' has wrong size
#elif (sizeof(Sint16) != 2)
#error 'Sint16' has wrong size
#elif (sizeof(Uint8) != 1)
#error 'Uint8' has wrong size
#elif (sizeof(Sint8) != 1)
#error 'Sint8' has wrong size
#elif (sizeof(Uint32) != 4)
#error 'Uint32' has wrong size
#elif (sizeof(Sint32) != 4)
#error 'Sint32' has wrong size
#endif

typedef enum {
	arrow_North,     // 0
	arrow_East,      // 1
	arrow_South,     // 2
	arrow_West,      // 3
	arrow_NorthEast, // 4
	arrow_SouthEast, // 5
	arrow_SouthWest, // 6
	arrow_NorthWest, // 7
	arrow_None       // 8
} arrowdirtype;

typedef enum {
	ex_stillplaying, //  0
	ex_died,         //  1
	ex_completed,    //  2
	ex_rescued,      //  3, only in Keen 4
	ex_warped,       //  4
	ex_resetgame,    //  5
	ex_loadedgame,   //  6
	ex_onfoot,       //  7, only in Keen 4
	ex_abort,        //  8
	ex_gotsandwich,  //  9, only in Keen 6
	ex_gotrope,      // 10, only in Keen 6
	ex_gotpasscard,  // 11, only in Keen 6
	ex_gotmolly,     // 12, only in Keen 6
	ex_teleported,   // 13, only in Keen 5
	ex_brokefuses,   // 14, only in Keen 5
	ex_brokeQED,     // 15, only in Keen 5
	NUMEXITTYPES
} exittype;

typedef enum
{
	INTILE_NOTHING,        //  0
	INTILE_POLE,           //  1
	INTILE_DOOR,           //  2
	INTILE_DEADLY,         //  3
	INTILE_LIFECRUMB,      //  4
	INTILE_SWITCH0,        //  5
	INTILE_SWITCH1,        //  6
	INTILE_GEMSOCKET0,     //  7
	INTILE_GEMSOCKET1,     //  8
	INTILE_GEMSOCKET2,     //  9
	INTILE_GEMSOCKET3,     // 10
	INTILE_SHORESOUTH,     // 11
	INTILE_SHOREWEST,      // 12
	INTILE_SHORENORTH,     // 13
	INTILE_SHOREEAST,      // 14
	INTILE_BRIDGESWITCH,   // 15
	INTILE_MOON,           // 16
	INTILE_DIRARROW,       // 17 (not used in the code, but assigned to tiles in Keen 5 & 6)
	INTILE_BRIDGE,         // 18
	INTILE_FORCEFIELD,     // 19
	INTILE_TELEPORT,       // 20
	INTILE_BONUS100,       // 21
	INTILE_BONUS200,       // 22
	INTILE_BONUS500,       // 23
	INTILE_BONUS1000,      // 24
	INTILE_BONUS2000,      // 25
	INTILE_BONUS5000,      // 26
	INTILE_EXTRALIFE,      // 27
	INTILE_AMMO,           // 28
	INTILE_29,             // 29 (never used)
	INTILE_FORCEFIELDEND,  // 30
	INTILE_AMPTONCOMPUTER, // 31
	INTILE_KEYCARDDOOR,    // 32
	INTILE_ELEVATORLEFT,   // 33
	INTILE_ELEVATORRIGHT,  // 34

	INTILE_FOREGROUND = 0x80
} intiletype;

#define INTILE_TYPEMASK ((byte)(~INTILE_FOREGROUND))

typedef enum
{
	nothing,         //  0
	decoobj,         //  1
	keenobj,         //  2
	shotobj,         //  3
#if defined KEEN4
	bonusobj,        //  4
	slugobj,         //  5
	oracleobj,       //  6
	classtype_7,     //  7, never used
	eggobj,          //  8
	madmushroomobj,  //  9
	arachnutobj,     // 10
	skypestobj,      // 11
	wormouthobj,     // 12
	thundercloudobj, // 13
	berkeloidobj,    // 14
	bounderobj,      // 15
	inchwormobj,     // 16
	footobj,         // 17
	lickobj,         // 18
	mimrockobj,      // 19
	platformobj,     // 20
	dopefishobj,     // 21
	schoolfishobj,   // 22
	spriteobj,       // 23
	lindseyobj,      // 24
	lightningobj,    // 25
	treasureeaterobj,// 26
	eagleobj,        // 27
	classtype_28,    // 28, never used
	classtype_29,    // 29, never used
	wetsuitobj,      // 30
	spriteshotobj,   // 31
	mineobj,         // 32
	stunnedobj,      // 33
	flagobj,         // 34
	badoingobj,
	witchdocobj,
	janitorobj,
	stingaruobj,
#elif defined KEEN5
	enemyshotobj,    //  4
	bonusobj,        //  5
	platformobj,     //  6
	stunnedobj,      //  7
	flagobj,         //  8
	sparkyobj,       //  9
	mineobj,         // 10
	slicestarobj,    // 11
	roboredobj,      // 12
	spiroobj,        // 13
	amptonobj,       // 14
	turretobj,       // 15
	volteobj,        // 16
	classtype_17,    // 17, never used
	spindredobj,     // 18
	masterobj,       // 19
	shikadiobj,      // 20
	shockshundobj,   // 21
	spherefulobj,    // 22
	korathobj,       // 23
	teleportfxobj,   // 24
	qedobj,          // 25
#elif defined KEEN6
	enemyshotobj,    //  4
	bonusobj,        //  5
	platformobj,     //  6
	bloogobj,        //  7
	bloogletobj,     //  8
	classtype_9,     //  9, never used
	fleexobj,        // 10
	classtype_11,    // 11, never used
	mollyobj,        // 12
	babobbaobj,      // 13
	bobbaobj,        // 14
	classtype_15,    // 15
	nospikeobj,      // 16
	gikobj,          // 17
	turretobj,       // 18
	orbatrixobj,     // 19
	bipobj,          // 20
	flectobj,        // 21
	blorbobj,        // 22
	ceilickobj,      // 23
	blooguardobj,    // 24
	stunnedobj,      // 25
	bipshipobj,      // 26
	sandwichobj,     // 27
	ropeobj,         // 28
	passcardobj,     // 29
	grabbiterobj,    // 30
	rocketobj,       // 31
	grapplespotobj,  // 32
	satelliteobj,    // 33
	satellitestopobj,// 34
	flagobj,         // 35
#endif
	NUMCLASSTYPES
} classtype;

typedef struct objstruct objtype;	//a dummy definition so that the function pointers in statetype can use the correct types

typedef struct statestruct
{
	Sint16 leftshapenum, rightshapenum;
	enum {step,slide,think,stepthink,slidethink} progress;
	boolean skippable;
	enum {push_none, push_down, push_up} pushtofloor;
	Sint16 tictime;
	Sint16 xmove;
	Sint16 ymove;
	void (*think) (objtype*);
	void (*contact) (objtype*, objtype*);
	void (*react) (objtype*);
	struct statestruct *nextstate;
} statetype;

typedef struct objstruct	//actual type definition
{
	classtype obclass;
	enum {no, yes, allways, removable} active;
	boolean needtoreact;
	Uint16 needtoclip;
	Uint16 nothink;
	Uint16 x, y;
	Sint16 xdir, ydir;
	Sint16 xmove, ymove;
	Sint16 xspeed, yspeed;
	Sint16 ticcount;
	statetype *state;
	Uint16 shapenum;
	Uint16 priority;
	Uint16 left, top, right, bottom, midx;
	Uint16 tileleft, tiletop, tileright, tilebottom, tilemidx;
	Sint16 hitnorth, hiteast, hitsouth, hitwest;
	Sint16 temp1, temp2, temp3, temp4;
	void *sprite;
	struct objstruct *next, *prev;
} objtype;

typedef struct
{
	Uint16 worldx, worldy;
	boolean leveldone[GAMELEVELS];
	Sint32 score, nextextra, oldscore;	// RTTS addition: oldscore
	Sint16 ammo, lifecrumbs;
#if defined KEEN4
	Sint16 wetsuit;
	Sint16 rescued;
#elif defined KEEN5
	boolean keycard, _unused_;
	Sint16 fuses;
#elif defined KEEN6
	Sint16 sandwichstate, ropestate, passcardstate, rocketstate;
#endif
	Sint16 keys[4];
	Sint16 mapon;
	Sint16 lives;
	Sint16 difficulty;
	objtype *platform;
} gametype;

/*
=============================================================================

						CK_MAIN DEFINITIONS

=============================================================================
*/

extern char str[80], str2[20];
extern boolean singlestep, jumpcheat, godmode, tedlevel;
extern boolean nopan, demo, joypad, fixjerky;

void MeasureTextBlock(char *text, Uint16 *width, Uint16 *height);

/*
=============================================================================

						CK_DEMO DEFINITIONS

=============================================================================
*/

extern boolean highscoredemo;

void CheckMenu(void);
void Intro(void);
void ShowTextCrawl(void);
void TitleScreen(void);
void RunDemo(Sint16 num);
void DrawHighScores(void);
void CheckHighScore(Sint32 score, Sint16 completed);
void RunHighscoreDemo(void);

/*
=============================================================================

						CK_GAME DEFINITIONS

=============================================================================
*/

void FreeUpMemory(void);
void NewGame(void);
boolean SaveTheGame(Sint16 handle);
boolean LoadTheGame(Sint16 handle);
void ResetGame(void);
void SetupGameLevel(boolean loadnow);
void DrawCacheBox(char *title, Uint16 numcache);
void UpdateCacheBox(void);
void FinishCacheBox(void);
void StartDemoRecord(void);
void EndDemoRecord(void);
void GameLoop(void);
void HandleDeath(void);

/*
=============================================================================

						CK_PLAY DEFINITIONS

=============================================================================
*/

extern exittype playstate;
extern gametype gamestate;
extern boolean debugmode, noscrolling;
extern objtype *scoreobj, *new, *player;
extern ControlInfo c;

extern Uint16 originxtilemax;
extern Uint16 originytilemax;

extern boolean upKeyHeld;
extern boolean
	button0down, button0held,
	button1down, button1held,
	button2down, button2held;

extern boolean oldshooting, showscorebox;
extern Sint16 invulnerable;

void CheckKeys(void);
void StatusWindow(void);
void CenterPlayer(objtype *ob);
void ScrollScreenWorld(objtype *ob);
void ScrollScreen(objtype *ob);
void InitObjList(void);
Sint16 GetNewObj(boolean usedummy);
void RemoveObj(objtype *ob);
void GivePoints(Uint16 points);
void StopMusic(void);
void StartMusic(Uint16 num);
void PlayLoop(void);

/*
=============================================================================

						CK_TEXT DEFINITIONS

=============================================================================
*/

void HelpScreens(void);
void EndText(void);

/*
=============================================================================

						CK_STATE DEFINITIONS

=============================================================================
*/

extern Sint16 wallclip[8][16];

extern Sint16 tmp_xmove;
extern Sint16 tmp_ymove;
extern boolean pushPlayer;

void MoveObjVert(objtype *ob, Sint16 ymove);
void MoveObjHoriz(objtype *ob, Sint16 xmove);
void ClipToNorthWall(objtype *ob);
void ClipToSouthWall(objtype *ob);
void ClipToEnds(objtype *ob);
void ClipToEastWestWalls(objtype *ob);
boolean PositionOkay(objtype *ob);
boolean StateOkay(objtype *ob, statetype *state);

#ifdef KEEN5
void CalcBounds(objtype *ob);
#endif

void ClipToWalls(objtype *ob);
void FullClipToWalls(objtype *ob);
void PushClipToWalls(objtype *ob);
void ClipToSpriteSide(objtype *push, objtype *solid);
void ClipToSpriteTop(objtype *push, objtype *solid);
void ClipToSprite(objtype *push, objtype *solid, boolean squish);
Sint16 DoActor(objtype *ob, Sint16 numtics);
void StateMachine(objtype *ob);
void NewState(objtype *ob, statetype *state);
void ChangeState(objtype *ob, statetype *state);
boolean ObjectVisible(objtype *ob);
void DoGravity4(objtype *ob);
void DoGravity3(objtype *ob);
void DoGravity1(objtype *ob);
void AccelerateX(objtype *ob, Sint16 dir, Sint16 maxspeed);
void AccelerateX_0(objtype *ob, Sint16 dir, Sint16 maxspeed);
void AccelerateY(objtype *ob, Sint16 dir, Sint16 maxspeed);
void FrictionX(objtype *ob);
void FrictionY(objtype *ob);
void StunContact(objtype *ob, objtype *shot, statetype *newstate);
void ProjectileThink4(objtype *ob);
void ProjectileThink3(objtype *ob);
void ProjectileThink1(objtype *ob);
void VelocityThink(objtype *ob);
void NeedToReactThink(objtype *ob);
void NeedToReactThink2(objtype *ob);
void HazardContact(objtype *ob, objtype *hit);
void DrawReact(objtype *ob);
void WalkReact(objtype *ob);
void WalkReact2(objtype *ob);
void BadState(objtype *ob);
void StunnedReact(objtype *ob);

extern statetype s_scorebox;
extern statetype s_badstate;

/*
=============================================================================

						CK_KEEN DEFINITIONS

=============================================================================
*/

extern Uint16 bounceangle[8][8];
#ifndef KEEN4
extern arrowdirtype oppositearrow[];
#endif

extern statetype s_keenstand;
extern statetype s_keenidlelookup;
extern statetype s_keenshrug1;
extern statetype s_keenshrug2;
extern statetype s_keenshrug3;
extern statetype s_keenshrug4;
extern statetype s_keenshrug5;
extern statetype s_keenshrug6;
extern statetype s_keenmoon1;
extern statetype s_keenmoon2;
extern statetype s_keenmoon3;
extern statetype s_keenread1;
extern statetype s_keenread2;
extern statetype s_keenread3;
extern statetype s_keenread4;
extern statetype s_keenread5;
extern statetype s_keenread6;
extern statetype s_keenread7;
extern statetype s_keenread8;
extern statetype s_keenread9;
extern statetype s_keenread10;
extern statetype s_keenlookup1;
extern statetype s_keenlookup2;
extern statetype s_keenduck1;
extern statetype s_keenduck2;
extern statetype s_keenduck3;
extern statetype s_keenafterduck;
extern statetype s_keenUNUSED1;
extern statetype s_keenUNUSED2;
extern statetype s_keendead1;
extern statetype s_keendead2;
extern statetype s_scubakeendead1;
extern statetype s_scubakeendead2;
extern statetype s_keenshoot1;
extern statetype s_keenshoot2;
extern statetype s_keenshootup1;
extern statetype s_keenshootup2;
extern statetype s_keenuseswitch;
extern statetype s_keenafteruse;
extern statetype s_keenusekeyholder;
extern statetype s_keenslide;
extern statetype s_keenenterdoor1;
extern statetype s_keenenterdoor2;
extern statetype s_keenenterdoor3;
extern statetype s_keenenterdoor4;
extern statetype s_keenenterdoor5;
extern statetype s_keenpole;
extern statetype s_keenpoleclimb1;
extern statetype s_keenpoleclimb2;
extern statetype s_keenpoleclimb3;
extern statetype s_keenpoleslide1;
extern statetype s_keenpoleslide2;
extern statetype s_keenpoleslide3;
extern statetype s_keenpoleslide4;
extern statetype s_keenpoleshoot1;
extern statetype s_keenpoleshoot2;
extern statetype s_keenpoleshootup1;
extern statetype s_keenpoleshootup2;
extern statetype s_keenpoleshootdown1;
extern statetype s_keenpoleshootdown2;
extern statetype s_keenwalk1;
extern statetype s_keenwalk2;
extern statetype s_keenwalk3;
extern statetype s_keenwalk4;
extern statetype s_keenpogo1;
extern statetype s_keenpogo2;
extern statetype s_keenpogo3;
extern statetype s_keenjump1;
extern statetype s_keenjump2;
extern statetype s_keenfall1;
extern statetype s_keenfall2;
extern statetype s_keenairshoot1;
extern statetype s_keenairshoot2;
extern statetype s_keenairshoot3;
extern statetype s_keenairshootup1;
extern statetype s_keenairshootup2;
extern statetype s_keenairshootup3;
extern statetype s_keenairshootdown1;
extern statetype s_keenairshootdown2;
extern statetype s_keenairshootdown3;
extern statetype s_keengrabedge1;
extern statetype s_keengrabedge2;
extern statetype s_keenclimbedge1;
extern statetype s_keenclimbedge2;
extern statetype s_keenclimbedge3;
extern statetype s_keenclimbedge4;
extern statetype s_keenclimbedge5;

extern Sint16 slopespeed[8];
extern Sint16 polespeed[3];

extern Sint16 ammotable[4];
extern Sint16 bonussound[];
extern Sint16 bonuspoints[];
extern Sint16 pickupshape[];

extern Uint16 emptytile;

extern Sint16 mooningstate;
extern Sint32 leavepoletime;
extern Sint16 jumptime;

void SpawnKeen(Sint16 x, Sint16 y, Sint16 dir);
boolean CheckGrabPole(objtype *ob);
boolean CheckInteractiveTiles(objtype *ob);
void KeenEnterDoor1(objtype *ob);
void KeenEnterDoor2(objtype *ob);
void KeenStandThink(objtype *ob);
void KeenIdleThink(objtype *ob);
void KeenReadThink(objtype *ob);
void KeenLookUpThink(objtype *ob);
void KeenDuckThink(objtype *ob);
void KeenWalkThink(objtype *ob);
void KeenSlideThink(objtype *ob);
void KeenDoorThink(objtype *ob);
void KeenUseSwitchThink(objtype *ob);
void KeenUseKeyholder(objtype *ob);
void KeenAirThink(objtype *ob);
void KeenPogoThink1(objtype *ob);
void KeenPogoThink(objtype *ob);
void PoleActions(objtype *ob);
void KeenPoleThink(objtype *ob);
void KeenPoleClimbThink(objtype *ob);
void KeenPoleSlideThink(objtype *ob);
void KeenClipThink(objtype *ob);
void KeenGrabEdgeThink(objtype *ob);
void KeenShootThink(objtype *ob);
void KeenClimbEdge1Think(objtype *ob);
void KeenClimbEdge2Think(objtype *ob);
void KeenClimbEdge3Think(objtype *ob);
void KeenClimbEdge4Think(objtype *ob);
void KeenDeadThink(objtype *ob);
void KillKeen(void);
void KeenContact(objtype *ob, objtype *hit);
void KeenContact2(objtype *ob, objtype *hit);
void FollowPlatform(objtype *ob);
void CollectBonusTile(Uint16 x, Uint16 y, Uint16 bonus);
void CollectLifecrumbTile(Uint16 x, Uint16 y);
void CheckTiles(objtype *ob);
void KeenDrawReact(objtype *ob);
void KeenStandReact(objtype *ob);
void KeenWalkReact(objtype *ob);
void KeenAirReact(objtype *ob);
void KeenPogoReact(objtype *ob);
void KeenPoleSlideReact(objtype *ob);


/*
=============================================================================

						CK_KEEN2 DEFINITIONS

=============================================================================
*/

extern statetype s_score;
extern statetype s_demo;
void SpawnScore(void);
void UpdateScorebox(objtype *ob);
void UpdateDemosprite(objtype *ob);

extern statetype s_worldkeen;
extern statetype s_worldkeenwave1;
extern statetype s_worldkeenwave2;
extern statetype s_worldkeenwave3;
extern statetype s_worldkeenwave4;
extern statetype s_worldkeenwave5;
extern statetype s_worldkeenwalk;
void SpawnWorldKeen(Sint16 x, Sint16 y);
#ifdef KEEN5
void SpawnWorldKeenTeleported(Uint16 tileX, Uint16 tileY);
#endif
void CheckEnterLevel(objtype *ob);
void KeenWorldThink(objtype *ob);
void KeenWorldWalk(objtype *ob);
void CheckWorldTiles(objtype *ob);

#ifdef KEEN4
extern statetype s_worldkeenride1;
extern statetype s_worldkeenride2;
extern statetype s_worldkeenswim;
void KeenWorldRideThink(objtype *ob);
void KeenWorldSwimThink(objtype *ob);
#endif

extern statetype s_flag1;
extern statetype s_flag2;
extern statetype s_flag3;
extern statetype s_flag4;
void PlaceFlag(Sint16 x, Sint16 y);

#ifndef KEEN5
extern statetype s_flagspin1;
extern statetype s_flagspin2;
extern statetype s_flagspin3;
extern statetype s_flagspin4;
extern statetype s_flagspin5;
extern statetype s_flagspin6;
extern statetype s_flagspin7;
void SpawnFlag(Sint16 x, Sint16 y);
void FlagThink1(objtype *ob);
void FlagThink2(objtype *ob);
void FlagLandThink(objtype *ob);
#endif

extern statetype s_keenshot1;
extern statetype s_keenshot2;
extern statetype s_keenshot3;
extern statetype s_keenshot4;
extern statetype s_keenshotexplode1;
extern statetype s_keenshotexplode2;
void SpawnShot(Uint16 x, Uint16 y, Direction dir);
void ExplodeShot(objtype *ob);
void KeenShotThink(objtype *ob);
void KeenShotReact(objtype *ob);

extern statetype s_gemdoor1;
extern statetype s_gemdoor2;
extern statetype s_gemdoor3;
void GemDoorThink(objtype *ob);

#ifdef KEEN5
extern statetype s_dooropener;
void DoorOpenerThink(objtype *ob);
#endif

/*
=============================================================================

						OTHER DEFINITIONS

=============================================================================
*/

#if defined KEEN4
#include "K4_DEF.H"
#elif defined KEEN5
#include "K5_DEF.H"
#elif defined KEEN6
#include "K6_DEF.H"
#endif

#endif