/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "CK_DEF.H"

/////////////////////////////////////////////////////////////////////////////
// initialized variables:
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// uninitialized variables:
/////////////////////////////////////////////////////////////////////////////

//Sint16 mapon;	//is actually declared in ID_CA.C
Sint32 cacheStep, cachesPerStep, cachesIdle;
Sint16 oldmapon;
static Uint16 fadecount;

/////////////////////////////////////////////////////////////////////////////
// local prototypes:
/////////////////////////////////////////////////////////////////////////////

static void FadeAndUnhook(void);

/////////////////////////////////////////////////////////////////////////////
// functions:
/////////////////////////////////////////////////////////////////////////////

void FreeUpMemory(void)
{
	register Sint16 i;
	for (i=DEMOSPR; i<STARTSPRITES+NUMSPRITES; i++)	// RTTS mod: don't free paddle war sprites
	{
		if (grsegs[i])
		{
			MM_SetPurge(&grsegs[i], PURGE_LAST);
		}
	}
	for (i=STARTTILE16; i<STARTEXTERNS; i++)
	{
		if (grsegs[i])
		{
			MM_SetPurge(&grsegs[i], PURGE_LAST);
		}
	}
}

void NewGame(void)
{
	memset(&gamestate, 0, sizeof(gamestate));
	gamestate.nextextra = 20000;
	gamestate.lives = 3;
	gamestate.ammo = 5;
}

#ifndef KEEN5
static void GameOver(void)
{
	VW_FixRefreshBuffer();
	US_CenterWindow(16, 3);
	US_PrintCentered("Game Over!");
	VW_UpdateScreen();
	IN_ClearKeysDown();
	IN_UserInput(4*TickBase, false);
}
#endif

#define RLETAG 0xABCD

boolean SaveTheGame(Sint16 handle)
{
	Uint16	i,size,compressed,expanded;
	objtype	*ob;
	memptr	bigbuffer;

	gamestate.platform = NULL;

	if (!CA_FarWrite(handle, (byte far *)&gamestate, sizeof(gamestate)))
		return false;

	expanded = mapwidth * mapheight * 2;
	MM_GetPtr(&bigbuffer, expanded);

	for (i = 0; i < 3; i++)
	{
		compressed = CA_RLEWCompress(mapsegs[i], expanded, (Uint16 huge *)bigbuffer+1, RLETAG);
		*(Uint16 huge *)bigbuffer = compressed;
		if (!CA_FarWrite(handle, bigbuffer, compressed+2))
		{
			MM_FreePtr(&bigbuffer);
			return false;
		}
	}
	for (ob = player; ob; ob=ob->next)
	{
		if (!CA_FarWrite(handle, (byte far *)ob, sizeof(objtype)))
		{
			MM_FreePtr(&bigbuffer);
			return false;
		}
	}
	MM_FreePtr(&bigbuffer);
	return true;
}

boolean LoadTheGame(Sint16 handle)
{
	Uint16	i,j,size;
	Sint16	orgx,orgy;
	objtype	*prev,*next,*followed;
	Uint16	compressed,expanded;
	memptr	bigbuffer;
#ifdef KEEN5
	Sint16	fuses;
#endif

	if (!CA_FarRead(handle, (byte far *)&gamestate, sizeof(gamestate)))
		return false;

#ifdef KEEN5
	fuses = gamestate.fuses;
#endif
	ca_levelbit >>= 1;
	ca_levelnum--;
	SetupGameLevel(false);
	if (mmerror)
	{
		mmerror = false;
		US_CenterWindow(20, 8);
		PrintY += 20;
		US_CPrint("Not enough memory\nto load game!");
		VW_UpdateScreen();
		IN_Ack();
		return false;
	}
	ca_levelbit <<= 1;
	ca_levelnum++;

	expanded = mapwidth * mapheight * 2;
	MM_BombOnError(true);	//BUG: this should use false to avoid an instant crash
	MM_GetPtr(&bigbuffer, expanded);
	MM_BombOnError(false);	//BUG: this should use true to force an instant crash
	if (mmerror)
	{
		mmerror = false;
		US_CenterWindow(20, 8);
		PrintY += 20;
		US_CPrint("Not enough memory\nto load game!");
		VW_UpdateScreen();
		IN_Ack();
		return false;
	}
	for (i = 0; i < 3; i++)
	{
		if (!CA_FarRead(handle, (byte far *)&compressed, sizeof(compressed)))
		{
			MM_FreePtr(&bigbuffer);
			return false;
		}
		if (!CA_FarRead(handle, (byte far *)bigbuffer, compressed))
		{
			MM_FreePtr(&bigbuffer);
			return false;
		}
		CA_RLEWexpand(bigbuffer, mapsegs[i], expanded, RLETAG);
	}
	MM_FreePtr(&bigbuffer);

	InitObjList();
	new = player;
	prev = new->prev;
	next = new->next;
	if (!CA_FarRead(handle, (byte far *)new, sizeof(objtype)))
	{
		return false;
	}
	new->prev = prev;
	new->next = next;
	new->needtoreact = true;
	new->sprite = NULL;
	new = scoreobj;
	while (true)
	{
		prev = new->prev;
		next = new->next;
		if (!CA_FarRead(handle, (byte far *)new, sizeof(objtype)))
		{
			return false;
		}
		followed = new->next;
		new->prev = prev;
		new->next = next;
		new->needtoreact = true;
		new->sprite = NULL;
		if (new->obclass == stunnedobj)
		{
			new->temp3 = 0;	//clear sprite ptr for the stars
		}
#if defined KEEN4
		else if (new->obclass == platformobj)
		{
			new->temp2 = new->temp3 = 0;	//clear sprite ptrs
		}
#elif defined KEEN5
		else if (new->obclass == mineobj)
		{
			new->temp4 = 0;	//clear sprite ptr
		}
		else if (new->obclass == spherefulobj)
		{
			new->temp1 = new->temp2 = new->temp3 = new->temp4 = 0;	//clear sprite ptrs
		}
#elif defined KEEN6
		else if (new->obclass == platformobj)
		{
			new->temp3 = 0;	//clear sprite ptr
		}
#endif
		if (followed)
		{
			GetNewObj(false);
		}
		else
		{
			break;
		}
	}
	scoreobj->temp2 = -1;
	scoreobj->temp1 = -1;
	scoreobj->temp3 = -1;
	scoreobj->temp4 = -1;
#ifdef KEEN5
	gamestate.fuses = fuses;
#endif
	return true;
}

void ResetGame(void)
{
	NewGame();
	ca_levelnum--;
	ca_levelbit >>= 1;
	CA_ClearMarks();
	ca_levelbit <<= 1;
	ca_levelnum++;
}

static void PatchWorldMap(void)
{
	Uint16 x, y, planeoff, info, level, tag;
	Uint16 far *infoptr;

	planeoff = 0;
	infoptr = mapsegs[2];
	for (y = 0; y < mapheight; y++)
	{
		for (x = 0; x < mapwidth; x++, infoptr++, planeoff++)
		{
			info = *infoptr;
			level = info & 0xFF;
			if (level >= MINDONELEVEL && level <= MAXDONELEVEL && gamestate.leveldone[level])
			{
				tag = info >> 8;
				*infoptr = 0;
				if (tag == 0xD0)
				{
					mapsegs[1][planeoff] = 0;
				}
				else if (tag == 0xF0)
				{
#ifdef KEEN5
					PlaceFlag(x, y);
#else
					if (oldmapon == level)
					{
						SpawnFlag(x, y);
					}
					else
					{
						PlaceFlag(x, y);
					}
#endif
				}
			}
		}
	}
}

static void FadeOut(void)
{
	VW_FadeOut();
	fadecount = 0;
	RF_SetRefreshHook(&FadeAndUnhook);
}

static void FadeAndUnhook(void)
{
	if (++fadecount == 2)
	{
		VW_FadeIn();
		RF_SetRefreshHook(NULL);
		TimeCount = lasttimecount;
	}
}

void SetupGameLevel(boolean loadnow)
{
	if (DemoMode)
	{
		US_InitRndT(false);
		gamestate.difficulty = gd_Normal;
	}
	else
	{
		US_InitRndT(true);
	}
	FreeUpMemory();	// RTTS addition
	CA_CacheMap(gamestate.mapon);
	RF_NewMap();
	CA_ClearMarks();
	ScanInfoPlane();
	if (mapon == 0)
	{
		PatchWorldMap();
	}
	RF_MarkTileGraphics();
	MM_BombOnError(false);
	CA_LoadAllSounds();
	if (loadnow)
	{
		if (highscoredemo)
		{
			CA_CacheMarks(NULL);
		}
		else if (DemoMode)
		{
			CA_CacheMarks("DEMO");
		}
#ifdef KEEN5
		else if (mapon == 0 && player->tiletop > 100)
		{
			CA_CacheMarks("Keen steps out\nonto Korath III");
		}
#endif
		else
		{
			_fstrcpy(str, levelmessages[mapon]);
			CA_CacheMarks(str);
		}
	}
	MM_BombOnError(true);
	if (!mmerror && loadnow)
	{
		FadeOut();
	}
}

void DrawCacheBox(char *title, Uint16 numcache)
{
	Sint16 i;
	Uint16 width, height;
	Sint32 totalfree;

	totalfree = MM_TotalFree();
	if (totalfree < 2048)
	{
		cacheStep = 5;
	}
	else
	{
		cacheStep = 0;
		for (i = 0; i < 6; i++)
		{
			CA_CacheGrChunk(i+LOADING0PIC);
			CA_UnmarkGrChunk(i+LOADING0PIC);
			if (grsegs[i+LOADING0PIC])
			{
				MM_SetPurge(&grsegs[i+LOADING0PIC], PURGE_FIRST);
			}
			else
			{
				mmerror = false;
				cacheStep = 5;
				break;
			}
		}
	}
	US_CenterWindow(26, 8);
	if (grsegs[LOADING0PIC])
	{
		VWB_DrawPic(WindowX, WindowY, LOADING0PIC);
	}
	else
	{
		cacheStep = 5;
	}
	CA_UnmarkGrChunk(LOADING0PIC);	//redundant
	WindowW -= 48;
	WindowX += 48;
	MeasureTextBlock(title, &width, &height);
	PrintY += (WindowH-height)/2 - 4;
	US_CPrint(title);
	VW_UpdateScreen();
	cachesPerStep = cachesIdle = numcache / 6;
	if (!cachesPerStep && !cacheStep)
	{
		if (grsegs[LOADING5PIC])
			VWB_DrawPic(WindowX-24, WindowY+40, LOADING5PIC);
		VW_UpdateScreen();
	}
}

void UpdateCacheBox(void)
{
	cachesIdle--;
	if (!cachesIdle && cacheStep <= 4)
	{
		cachesIdle = cachesPerStep;
		if (grsegs[cacheStep+LOADING1PIC])
		{
			VWB_DrawPic(WindowX-24, WindowY+40, cacheStep+LOADING1PIC);
		}
		VW_UpdateScreen();
		cacheStep++;
	}
}

void FinishCacheBox(void)
{
	//this is empty
}

void StartDemoRecord(void)
{
	Sint16 level;
	boolean esc;

	VW_FixRefreshBuffer();
	US_CenterWindow(30, 3);
	PrintY += 6;
	US_Print("  Record a demo from level(0-21):");
	VW_UpdateScreen();
	esc = !US_LineInput(px, py, str, NULL, true, 2, 0);
	if (!esc)
	{
		level = atoi(str);
		if (level >= 0 && level <= 21)
		{
			gamestate.mapon = level;
			playstate = ex_warped;
			IN_StartDemoRecord(0x1000);
		}
	}
}

void EndDemoRecord(void)
{
	Sint16 handle;
	boolean esc;
	char filename[] = "DEMO?."EXTENSION;

	IN_StopDemo();
	VW_FixRefreshBuffer();
	US_CenterWindow(22, 3);
	PrintY += 6;
	US_Print("  Save as demo #(0-9):");
	esc = !US_LineInput(px, py, str, NULL, true, 2, 0);
	if (!esc && str[0] >= '0' && str[0] <= '9')
	{
		filename[4] = str[0];
		handle = open(filename, O_BINARY|O_WRONLY|O_CREAT, S_IFREG|S_IREAD|S_IWRITE);
		if (handle == -1)
		{
			Quit("EndDemoRecord:  Cannot write demo file!");
		}
		write(handle, &mapon, sizeof(mapon));
		write(handle, &DemoOffset, sizeof(DemoOffset));
		CA_FarWrite(handle, DemoBuffer, DemoOffset);
		close(handle);
	}
	IN_FreeDemoBuffer();
}

void HandleDeath(void)
{
	Uint16 y, color, top, bottom, selection, w, h;

	_fstrcpy(str, levelnames[mapon]);
	MeasureTextBlock(str, &w, &h);

	memset(gamestate.keys, 0, sizeof(gamestate.keys));
	gamestate.lives--;
	if (gamestate.lives >= 0)
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(20, 8);
		PrintY += 3;
		US_CPrint("You didn't make it past");
		top = PrintY+22;
		if (h < 15)
			PrintY += 4;
		US_CPrint(str);
		PrintY = top+2;
		US_CPrint("Try Again");
		PrintY += 4;
		bottom = PrintY-2;
		US_CPrint("Exit to "WORLDMAPNAME);

		IN_ClearKeysDown();
		selection = 0;
		while (true)
		{
			if (selection)
			{
				y = bottom;
			}
			else
			{
				y = top;
			}
			if ((TimeCount / 16) & 1)
			{
				color = SECONDCOLOR;
			}
			else
			{
				color = FIRSTCOLOR;
			}
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y, color);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+1, color);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+12, color);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+13, color);
			VWB_Vlin(y+1, y+11, WindowX+4, color);
			VWB_Vlin(y+1, y+11, WindowX+5, color);
			VWB_Vlin(y+1, y+11, WindowX+WindowW-4, color);
			VWB_Vlin(y+1, y+11, WindowX+WindowW-5, color);
			VW_UpdateScreen();
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y, WHITE);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+1, WHITE);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+12, WHITE);
			VWB_Hlin(WindowX+4, WindowX+WindowW-4, y+13, WHITE);
			VWB_Vlin(y+1, y+11, WindowX+4, WHITE);
			VWB_Vlin(y+1, y+11, WindowX+5, WHITE);
			VWB_Vlin(y+1, y+11, WindowX+WindowW-4, WHITE);
			VWB_Vlin(y+1, y+11, WindowX+WindowW-5, WHITE);

			if (LastScan == sc_Escape)
			{
				gamestate.mapon = 0;
				IN_ClearKeysDown();
				return;
			}
			IN_ReadControl(0, &c);
			if (c.button0 || c.button1 || LastScan == sc_Return || LastScan == sc_Space)
			{
				if (selection)
					gamestate.mapon = 0;
				return;
			}
			if (c.yaxis == -1 || LastScan == sc_UpArrow)
			{
				selection = 0;
			}
			else if (c.yaxis == 1 || LastScan == sc_DownArrow)
			{
				selection = 1;
			}
		}
	}
}

void GameLoop(void)
{
	Uint16 temp;

#ifdef KEEN6
	if (!demo)
	{
		if (!US_ManualCheck())
		{
			loadedgame = false;
			restartgame = gd_Continue;
			return;
		}
	}
#endif

	if (playstate == ex_loadedgame)
	{
		goto loaded;
	}
reset:
	gamestate.difficulty = restartgame;
	restartgame = gd_Continue;
	do
	{
		gamestate.oldscore = gamestate.score;	// RTTS addition

		SetupGameLevel(true);
		if (mmerror)
		{
			if (gamestate.mapon != 0)
			{
				mmerror = false;
				US_CenterWindow(20, 8);
				PrintY += 20;
				US_CPrint("Insufficient memory\nto load level!");
				VW_UpdateScreen();
				IN_Ack();
				gamestate.mapon = 0;
				SetupGameLevel(true);
			}
			if (mmerror)
			{
				Quit("GameLoop: Insufficient memory to load world map!");
			}
		}
loaded:
		noscrolling = false;
		SD_WaitSoundDone();
		PlayLoop();
		if (playstate != ex_loadedgame)
		{
			memset(gamestate.keys, 0, sizeof(gamestate.keys));
#ifdef KEEN5
			gamestate.keycard = false;
#endif
		}
		VW_FixRefreshBuffer();
		if (tedlevel)
		{
			if (playstate == ex_loadedgame)
			{
				goto loaded;
			}
			else if (playstate == ex_died)
			{
				continue;
			}
			else
			{
				TEDDeath();
			}
		}
		oldmapon = -1;
		switch (playstate)
		{
		case ex_died:
			gamestate.score = gamestate.oldscore;	// RTTS addition
			HandleDeath();
			break;
		case ex_resetgame:
			goto reset;
		case ex_loadedgame:
			goto loaded;
#if defined KEEN4
		case ex_rescued:
			if (mapon != 0)
			{
				SD_PlaySound(SND_LEVELDONE);
			}
			oldmapon = mapon;
			gamestate.leveldone[mapon] = true;
			OracleDialog();
			if (gamestate.rescued != 8)
			{
				gamestate.mapon = 0;
				break;
			}
			FreeUpMemory();
			RF_FixOfs();
			VW_FixRefreshBuffer();
			EndText();
			CheckHighScore(gamestate.score, gamestate.rescued);
			return;
#elif defined KEEN5
		case ex_brokefuses:
			SD_PlaySound(SND_LEVELDONE);
			oldmapon = mapon;
			gamestate.leveldone[mapon] = ex_brokefuses;
			FuseDialog();
			gamestate.mapon = 0;
			break;
		case ex_brokeQED:
			FreeUpMemory();
			RF_FixOfs();
			VW_FixRefreshBuffer();
			EndText();
			CheckHighScore(gamestate.score, 0);
			return;
#elif defined KEEN6
		case ex_gotrope:
			RopeDialog();
			goto completed;
		case ex_gotsandwich:
			SandwichDialog();
			goto completed;
		case ex_gotpasscard:
			PasscardDialog();
			goto completed;
		case ex_gotmolly:
			FreeUpMemory();
			RF_FixOfs();
			VW_FixRefreshBuffer();
			EndText();
			goto check_score;
#endif
		case ex_completed:
		case ex_onfoot:
		case ex_teleported:
completed:
			if (mapon != 0)
			{
				SD_PlaySound(SND_LEVELDONE);
				gamestate.mapon = 0;
				oldmapon = mapon;
				gamestate.leveldone[mapon] = true;
				if (demo && mapon == 2)
				{
					IN_ClearKeysDown();
					return;
				}
			}
			else
			{
				temp = bufferofs;
				bufferofs = displayofs;
				US_CenterWindow(26, 8);
				PrintY += 25;
				US_CPrint("One moment");
				bufferofs = temp;
			}
			break;
		case ex_abort:
			IN_ClearKeysDown();
			return;
		}
	} while (gamestate.lives >= 0);
	GameOver();
check_score:
#if defined KEEN4
	CheckHighScore(gamestate.score, gamestate.rescued);
#else
	temp = 0;
#if defined KEEN6
	{
		Uint16 i;

		for (i=0; i<GAMELEVELS; i++)
		{
			if (gamestate.leveldone[i])
				temp++;
		}
	}
#endif
	CheckHighScore(gamestate.score, temp);
#endif
}