/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
CK_KEEN2.C
==========

Contains the following actor types (in this order):

- Score Box & Demo sprites
- Keen (world map)
- Flags (world map)
- Neural Stunner Shots
- Gem Door Opener

*/

#include "CK_DEF.H"

Direction oppositedir[8] = {dir_South, dir_SouthWest, dir_West, dir_NorthWest, dir_North, dir_NorthEast, dir_East, dir_SouthEast};

//============================================================================
//============================================================================

statetype s_score         = {  0,   0, think, false, push_none, 0, 0, 0, NULL, NULL, NULL, NULL};
statetype s_demo          = {DEMOSPR, DEMOSPR, think, false, push_none, 0, 0, 0, NULL, NULL, NULL, NULL};

void SpawnScore(void)
{
	scoreobj->obclass = decoobj;
	scoreobj->priority = 3;
	scoreobj->active = allways;
	scoreobj->needtoclip = false;
	scoreobj->temp2 = -1;
	scoreobj->temp1 = -1;
	scoreobj->temp3 = -1;
	scoreobj->temp4 = -1;
	if (highscoredemo)
	{
		scoreobj->state = &s_scorebox;
	}
	else if (!DemoMode)
	{
		NewState(scoreobj, &s_score);
	}
	else
	{
		NewState(scoreobj, &s_demo);
		CA_MarkGrChunk(DEMOSPR);
	}
}

// Taken from Keen Dreams: MemDrawChar and ShiftScore

static void MemDrawChar(Sint16 char8, Uint8 far *dest, Uint16 width, Uint16 planesize)
{
	Uint16 source = (Uint16)grsegs[STARTTILE8];	//KDB: this differs from Keen Dreams source

asm	mov	si,[char8]
asm	shl	si,1
asm	shl	si,1
asm	shl	si,1
asm	shl	si,1
asm	shl	si,1		// index into char 8 segment

asm	mov	ds,[WORD PTR source]	//KDB: this differs from Keen Dreams source
asm	mov	es,[WORD PTR dest+2]

asm	mov	cx,4		// draw four planes
asm	mov	bx,[width]
asm	dec	bx

planeloop:

asm	mov	di,[WORD PTR dest]

asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb
asm	add	di,bx
asm	movsb

asm	mov	ax,[planesize]
asm	add	[WORD PTR dest],ax

asm	loop	planeloop

asm	mov	ax,ss
asm	mov	ds,ax
}

static void ShiftScore (void)
{
	spritetabletype far *spr;
	spritetype _seg *dest;

	spr = &spritetable[SCOREBOXSPR-STARTSPRITES];
	dest = (spritetype _seg *)grsegs[SCOREBOXSPR];

	CAL_ShiftSprite (FP_SEG(dest),dest->sourceoffset[0],
		dest->sourceoffset[1],spr->width,spr->height,2);

	CAL_ShiftSprite (FP_SEG(dest),dest->sourceoffset[0],
		dest->sourceoffset[2],spr->width,spr->height,4);

	CAL_ShiftSprite (FP_SEG(dest),dest->sourceoffset[0],
		dest->sourceoffset[3],spr->width,spr->height,6);
}

void UpdateScorebox(objtype *ob)
{
	char		str[10],*ch;
	spritetype	_seg	*block;
	Uint8		far *dest;
	Uint16	i, length, width, planesize, number;
	boolean changed;

	if (highscoredemo)
		return;

	if (DemoMode)
	{
		UpdateDemosprite(ob);
		return;
	}

	if (!showscorebox)
		return;

	changed = false;

//code below is a combination of ScoreThink and ScoreReact from Keen Dreams with minor changes

//
// score changed
//
	if ((gamestate.score>>16) != ob->temp1
		|| (Uint16)gamestate.score != ob->temp2 )
	{
		block = (spritetype _seg *)grsegs[SCOREBOXSPR];
		width = block->width[0];
		planesize = block->planesize[0];
		dest = (Uint8 far *)grsegs[SCOREBOXSPR]+block->sourceoffset[0]
			+ planesize + width*4;

		ltoa (gamestate.score,str,10);

		// erase leading spaces
		length = strlen(str);
		for (i=9;i>length;i--)
			MemDrawChar (41,dest+=CHARWIDTH,width,planesize);

		// draw digits
		ch = str;
		while (*ch)
			MemDrawChar (*ch++ - '0'+42,dest+=CHARWIDTH,width,planesize);

#if GRMODE == EGAGR
		ShiftScore ();
#endif
		ob->needtoreact = true;
		ob->temp1 = gamestate.score>>16;
		ob->temp2 = gamestate.score;

		changed = true;
	}

//
// ammo changed
//
	number = gamestate.ammo;
	if (number != ob->temp3)
	{
		block = (spritetype _seg *)grsegs[SCOREBOXSPR];
		width = block->width[0];
		planesize = block->planesize[0];
		dest = (byte far *)grsegs[SCOREBOXSPR]+block->sourceoffset[0]
			+ planesize + width*20 + 7*CHARWIDTH;

		if (number > 99)
			strcpy (str,"99");
		else
			ltoa (number,str,10);

		// erase leading spaces
		length = strlen(str);
		for (i=2;i>length;i--)
			MemDrawChar (41,dest+=CHARWIDTH,width,planesize);

		// draw digits
		ch = str;
		while (*ch)
			MemDrawChar (*ch++ - '0'+42,dest+=CHARWIDTH,width,planesize);

#if GRMODE == EGAGR
		ShiftScore ();
#endif
		ob->needtoreact = true;
		ob->temp3 = gamestate.ammo;

		changed = true;
	}

//
// lives changed
//
	if (gamestate.lives != ob->temp4)
	{
		block = (spritetype _seg *)grsegs[SCOREBOXSPR];
		width = block->width[0];
		planesize = block->planesize[0];
		dest = (byte far *)grsegs[SCOREBOXSPR]+block->sourceoffset[0]
			+ planesize + width*20 + 2*CHARWIDTH;

		if (gamestate.lives > 99)
			strcpy (str,"99");
		else
			ltoa (gamestate.lives,str,10);

		// erase leading spaces
		length = strlen(str);
		for (i=2;i>length;i--)
			MemDrawChar (41,dest+=CHARWIDTH,width,planesize);

		// draw digits
		ch = str;
		while (*ch)
			MemDrawChar (*ch++ - '0'+42,dest+=CHARWIDTH,width,planesize);

#if GRMODE == EGAGR
		ShiftScore ();
#endif
		ob->needtoreact = true;
		ob->temp4 = gamestate.lives;

		changed = true;
	}

/*
Note:
-----

It would be more efficient to use

	if (changed)
		ShiftScore();

here instead of the individual ShiftScore() calls above. Because if the player
gains a life by collecting points items, both the score and lives numbers need
to be updated, which means the sprite would be shifted twice. And if the player
fires a shot during the same frame, the ammo number also needs to be updated,
leading to up to three shifts in one frame.
*/

	if (ob->x != originxglobal || ob->y != originyglobal)
	{
		ob->x = originxglobal;
		ob->y = originyglobal;
		changed = true;
	}

	if (changed)
#if GRMODE == EGAGR
		RF_PlaceSprite(&ob->sprite, ob->x+4*PIXGLOBAL, ob->y+4*PIXGLOBAL, SCOREBOXSPR, 0, 3);
#elif GRMODE == CGAGR
		RF_PlaceSprite(&ob->sprite, ob->x+8*PIXGLOBAL, ob->y+8*PIXGLOBAL, SCOREBOXSPR, 0, 3);
#endif
}

void UpdateDemosprite(objtype *ob)
{
	if (ob->x != originxglobal || ob->y != originyglobal)
	{
		ob->x = originxglobal;
		ob->y = originyglobal;
		RF_PlaceSprite(&ob->sprite, ob->x + 160*PIXGLOBAL - 32*PIXGLOBAL, ob->y + 8*PIXGLOBAL, DEMOSPR, 0, 3);
	}
}

//============================================================================
//============================================================================

#ifdef KEEN4
statetype s_worldkeenride1 = {WOLRDKEENRIDE1SPR, WOLRDKEENRIDE1SPR, stepthink, false, push_none, 30, 0, 0, KeenWorldRideThink, NULL, DrawReact, &s_worldkeenride2};
statetype s_worldkeenride2 = {WOLRDKEENRIDE2SPR, WOLRDKEENRIDE2SPR, stepthink, false, push_none, 30, 0, 0, KeenWorldRideThink, NULL, DrawReact, &s_worldkeenride1};
statetype s_worldkeenswim = {0, 0, slide, true, push_none, 6, 16, 16, KeenWorldSwimThink, NULL, DrawReact, &s_worldkeenswim};
#endif

statetype s_worldkeen     = {0, 0, stepthink, false, push_none, 360, 0, 0, KeenWorldThink, NULL, DrawReact, &s_worldkeenwave1};

statetype s_worldkeenwave1 = {WORLDKEENWAVE1SPR, WORLDKEENWAVE1SPR, stepthink, false, push_none, 20, 0, 0, KeenWorldThink, NULL, DrawReact, &s_worldkeenwave2};
statetype s_worldkeenwave2 = {WORLDKEENWAVE2SPR, WORLDKEENWAVE2SPR, stepthink, false, push_none, 20, 0, 0, KeenWorldThink, NULL, DrawReact, &s_worldkeenwave3};
statetype s_worldkeenwave3 = {WORLDKEENWAVE1SPR, WORLDKEENWAVE1SPR, stepthink, false, push_none, 20, 0, 0, KeenWorldThink, NULL, DrawReact, &s_worldkeenwave4};
statetype s_worldkeenwave4 = {WORLDKEENWAVE2SPR, WORLDKEENWAVE2SPR, stepthink, false, push_none, 20, 0, 0, KeenWorldThink, NULL, DrawReact, &s_worldkeenwave5};
statetype s_worldkeenwave5 = {WORLDKEENWAVE1SPR, WORLDKEENWAVE1SPR, stepthink, false, push_none, 20, 0, 0, KeenWorldWalk, NULL, DrawReact, &s_worldkeen};

statetype s_worldkeenwalk = {0, 0, slide, true, push_none, 4, 24, 24, KeenWorldWalk, NULL, DrawReact, &s_worldkeenwalk};

Sint16 worldshapes[8] = {WORLDKEENU1SPR-1, WORLDKEENUR1SPR-1, WORLDKEENR1SPR-1, WORLDKEENDR1SPR-1, WORLDKEEND1SPR-1, WORLDKEENDL1SPR-1, WORLDKEENL1SPR-1, WORLDKEENUL1SPR-1};	//-1 to everything because worldanims values are 1-based
Sint16 worldanims[4] = {2, 3, 1, 3};
#ifdef KEEN4
Sint16 worldswimshapes[8] = {WORLDKEENSWIMU1SPR, WORLDKEENSWIMUR1SPR, WORLDKEENSWIMR1SPR, WORLDKEENSWIMDR1SPR, WORLDKEENSWIMD1SPR, WORLDKEENSWIMDL1SPR, WORLDKEENSWIML1SPR, WORLDKEENSWIMUL1SPR};
#endif
Sint16 shoredir[4] = {dir_South, dir_West, dir_North, dir_East};

void SpawnWorldKeen(Sint16 x, Sint16 y)
{
#ifdef KEEN4
	if (playstate == ex_onfoot)
	{
		player->needtoclip = false;
		player->obclass = keenobj;
		player->x = gamestate.worldx;
		player->y = gamestate.worldy;
		player->active = allways;
		player->priority = 3;
		player->xdir = 0;
		player->ydir = 0;

		// Note: this has been modified for RTTS:
		if (gamestate.worldy > 30*TILEGLOBAL)
		{
			// fly to the Lost Longhouse:
			player->temp1 = 280;
			player->xspeed = (Sint16)(32*TILEGLOBAL - player->x)/280 + 1;
			player->yspeed = (Sint16)(24*TILEGLOBAL - player->y)/280 + 1;
		}
		else
		{
			// fly to the forest southeast of the Pyramid of Footsteps:
			player->temp1 = 140;
			player->xspeed = (Sint16)(32*TILEGLOBAL - player->x)/140 + 1;
			player->yspeed = (Sint16)(45*TILEGLOBAL - player->y)/140 + 1;
		}

		NewState(player, &s_worldkeenride1);
		return;
	}
#endif

	player->obclass = keenobj;
	if (gamestate.worldx == 0)
	{
		player->x = CONVERT_TILE_TO_GLOBAL(x);
		player->y = CONVERT_TILE_TO_GLOBAL(y);
	}
	else
	{
		player->x = gamestate.worldx;
		player->y = gamestate.worldy;
	}
	player->active = allways;
	player->priority = 1;
	player->xdir = 0;
	player->ydir = 0;
	player->temp1 = dir_West;
	player->temp2 = 3;
	player->temp3 = 0;
	player->shapenum = WORLDKEENL3SPR;
	NewState(player, &s_worldkeen);
}

#ifdef KEEN5
void SpawnWorldKeenTeleported(Uint16 tileX, Uint16 tileY)
{
	player->obclass = keenobj;
	player->x = CONVERT_TILE_TO_GLOBAL(tileX);
	player->y = CONVERT_TILE_TO_GLOBAL(tileY);
	player->active = allways;
	player->priority = 1;
	player->xdir = 0;
	player->ydir = 0;
	player->temp1 = dir_West;
	player->temp2 = 3;
	player->temp3 = 0;
	player->shapenum = WORLDKEENL3SPR;
	NewState(player, &s_worldkeen);
}
#endif

void CheckEnterLevel(objtype *ob)
{
	Uint16 x, y, info;

	for (y = ob->tiletop; y <= ob->tilebottom; y++)
	{
		for (x = ob->tileleft; x <= ob->tileright; x++)
		{
			info = mapsegs[2][mapbwidthtable[y]/2 + x];
			if (info > 0xC000 && info <= (0xC000 + 20))	// RTTS mod: allow up to 20 enterable levels
			{
				gamestate.worldx = ob->x;
				gamestate.worldy = ob->y;
				gamestate.mapon = info - 0xC000;
				playstate = ex_completed;
				SD_PlaySound(SND_ENTERLEVEL);
			}
		}
	}
}

void KeenWorldThink(objtype *ob)
{
	if (c.dir != dir_None)
	{
		ob->state = &s_worldkeenwalk;
		ob->temp2 = 0;
		KeenWorldWalk(ob);
	}
	if (button0down || button1down || button2down)
	{
		CheckEnterLevel(ob);
	}
}

void KeenWorldWalk(objtype *ob)
{
	if (ob->temp3)
	{
		ob->temp3 -= 4;
		if (ob->temp3 < 0)
			ob->temp3 = 0;
	}
	else
	{
		ob->xdir = c.xaxis;
		ob->ydir = c.yaxis;
		if (button1down || button2down || button0down)
		{
			CheckEnterLevel(ob);
		}
		if (c.dir == dir_None)
		{
			ob->state = &s_worldkeen;
			ob->shapenum = worldshapes[ob->temp1] + 3;
			return;
		}
		ob->temp1 = c.dir;
	}
	ob->temp2++;
	if (ob->temp2 == 4)
		ob->temp2 = 0;
	ob->shapenum = worldshapes[ob->temp1] + worldanims[ob->temp2];

	if (ob->temp2 == 1)
	{
		SD_PlaySound(SND_WORLDWALK1);
	}
	else if (ob->temp2 == 3)
	{
		SD_PlaySound(SND_WORLDWALK2);
	}
}

#ifdef KEEN4
void KeenWorldRideThink(objtype *ob)
{
	ob->temp1 = ob->temp1 - tics;
	tmp_xmove = ob->xspeed * tics;
	tmp_ymove = ob->yspeed * tics;
	if (ob->temp1 <= 0)
	{
		tmp_xmove -= ob->xspeed * -ob->temp1;
		tmp_ymove -= ob->yspeed * -ob->temp1;
		ob->priority = 1;
		ob->temp1 = dir_West;
		ob->temp2 = 3;
		ob->temp3 = 0;
		player->xdir = 0;
		player->ydir = 0;
		ob->state = &s_worldkeen;
		ob->shapenum = WORLDKEENL3SPR;
		ob->needtoclip = true;
	}
}

void KeenWorldSwimThink(objtype *ob)
{
	if (ob->temp3)
	{
		ob->temp3 -= 6;
		if (ob->temp3 < 0)
			ob->temp3 = 0;
	}
	else
	{
		ob->xdir = c.xaxis;
		ob->ydir = c.yaxis;
		if (c.xaxis || c.yaxis)
			ob->temp1 = c.dir;
	}
	ob->shapenum = worldswimshapes[ob->temp1] + ob->temp2;
	ob->temp2++;
	if (ob->temp2 == 2)
		ob->temp2 = 0;

	if (ob->temp2 == 0)
	{
		SD_PlaySound(SND_SWIM1);
	}
	else
	{
		SD_PlaySound(SND_SWIM2);
	}
}

#else	// NOT Keen 4 (i.e. Keen 5 & 6):

void EnterWorldTeleporter(Uint16 tileX, Uint16 tileY)
{
	Uint16 tile, globalx, globaly, duration, move;
	objtype *o;
	objtype *ob = player;

	//
	// enter the teleporter
	//
	SD_PlaySound(SND_TELEPORT);
	globalx = CONVERT_TILE_TO_GLOBAL(tileX);
	globaly = CONVERT_TILE_TO_GLOBAL(tileY);

	for (duration = 0; duration < 130; )
	{
		RF_Refresh();
		move = tics*2;
		duration += tics;

		if (ob->x == globalx && ob->y == globaly)
			break;

		if (ob->y < globaly)
		{
			ob->y += move;
			if (ob->y > globaly)
				ob->y = globaly;
		}
		else if (ob->y > globaly)
		{
			ob->y -= move;
			if (ob->y < globaly)
				ob->y = globaly;
		}

		if (ob->x < globalx)
		{
			ob->x += move;
			if (ob->x > globalx)
				ob->x = globalx;
		}
		else if (ob->x > globalx)
		{
			ob->x -= move;
			if (ob->x < globalx)
				ob->x = globalx;
		}

		ob->shapenum = ((TimeCount >> 3) % 3) + WORLDKEENU1SPR;
		RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, spritedraw, ob->priority);

		tile = ((TimeCount >> 2) & 1) + TELEPORTERTILE1;
		RF_MemToMap(&tile, 1, tileX, tileY, 1, 1);
	}

	tile = TELEPORTERTILE2;
	RF_MemToMap(&tile, 1, tileX, tileY, 1, 1);

	//
	// teleport to new location
	//
	tile = mapsegs[2][mapbwidthtable[tileY]/2 + tileX];
	tileX = tile >> 8;
	tileY = tile & 0xFF;
	ob->x = CONVERT_TILE_TO_GLOBAL(tileX);
	ob->y = CONVERT_TILE_TO_GLOBAL(tileY);
	ob->xdir = 0;
	ob->ydir = 1;
	ob->temp1 = dir_South;
	NewState(ob, ob->state);
	CenterPlayer(ob);

	//
	// draw flags/signs for new location
	//
	for (o=player->next; o; o=o->next)
	{
		if (!o->active && o->obclass == flagobj
			&& o->tileright >= originxtile-1 && o->tileleft <= originxtilemax+1
			&& o->tiletop <= originytilemax+1 && o->tilebottom >= originytile+1)
		{
			o->needtoreact = true;
			o->active = yes;
			RF_PlaceSprite(&o->sprite, o->x, o->y, o->shapenum, spritedraw, o->priority);
		}
	}
	UpdateScorebox(scoreobj);
	RF_Refresh();
	RF_Refresh();

	//
	// leave teleporter
	//
	SD_PlaySound(SND_TELEPORT);

	for (duration = 0; duration < 90; )
	{
		RF_Refresh();
		duration += tics;
		ob->y += tics*2;

		ob->shapenum = ((TimeCount >> 3) % 3) + WORLDKEEND1SPR;
		RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, spritedraw, ob->priority);

		tile = ((TimeCount >> 2) & 1) + TELEPORTERTILE3;
		RF_MemToMap(&tile, 1, tileX, tileY, 1, 1);
	}

	tile = TELEPORTERTILE4;
	RF_MemToMap(&tile, 1, tileX, tileY, 1, 1);
	tmp_xmove = tmp_ymove = 0;
	ClipToWalls(ob);
}

#ifdef KEEN5

void KeenWorldLeaveElevator(objtype *ob)
{
	Sint16 i, x, y;
	Uint16 tx, ty;
	Uint16 tiles[4];

	tmp_ymove = ob->ydir * 64 * tics;
	if (ob->x != ob->temp2)
	{
		tmp_xmove = ob->xdir * 12 * tics;
		if ( (ob->xdir == 1 && ob->x + tmp_xmove > ob->temp2)
			|| (ob->xdir == -1 && ob->x + tmp_xmove < ob->temp2) )
		{
			tmp_xmove = ob->temp2 - ob->x;
		}
	}

	//
	// Keen has no sprite in this state, so we need to update the hitbox manually
	// to avoid issues (the screen scrolling routines use left/right/top/bottom)
	//
	ob->left = ob->x + tmp_xmove;
	ob->right = ob->left + TILEGLOBAL-1;
	ob->top = ob->y + tmp_ymove;
	ob->bottom = ob->top + TILEGLOBAL-1;

	if (ob->ydir == 1)
	{
		if (ob->y + tmp_ymove < ob->temp1)
			return;
	}
	else
	{
		if (ob->y + tmp_ymove > ob->temp1)
			return;
	}

	//
	// the invisible Keen has arrived at its destination
	//
	tmp_ymove = 0;
	tmp_xmove = 0;
	ob->x = ob->temp2;
	ob->y = ob->temp1;
	ob->priority = 1;
	ob->temp1 = 4;
	ob->temp2 = 3;
	ob->temp3 = 0;
	player->xdir = 0;
	player->ydir = 0;
	ob->state = &s_worldkeen;
	ob->shapenum = WORLDKEEND3SPR;
	ob->needtoclip = true;
	tx = CONVERT_GLOBAL_TO_TILE(ob->x);
	ty = CONVERT_GLOBAL_TO_TILE(ob->y);
	ScrollScreenWorld(ob);
	UpdateScorebox(scoreobj);
	RF_Refresh();
	RF_Refresh();

	ob->y -= TILEGLOBAL;
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, spritedraw, ob->priority);

	//
	// open the elevator door
	//
	SD_PlaySound(SND_ELEVATORDOOR);
	for (i=0; i<=5; i++)
	{
		for (y=0; y<2; y++)
		{
			for (x=0; x<2; x++)
			{
				tiles[2*y + x] = mapsegs[1][mapbwidthtable[y]/2 + i*2 + x];
			}
		}
		RF_MemToMap(tiles, 1, tx, ty-2, 2, 2);
		RF_Refresh();
		VW_WaitVBL(8);
	}

	//
	// make Keen walk out of the elevator
	//
	for (y=0; y<32; y++)
	{
		ob->y += 8;	// move half a pixel every frame for 32 frames -> move down 16 pixels total
		ob->shapenum = (y / 4) % 3 + WORLDKEEND1SPR;
		RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, spritedraw, ob->priority);
		RF_Refresh();
	}
	ob->needtoclip = true;	// redundant, but doesn't do any harm
}

statetype s_worldkeenhidden = {-1, -1, think, true, push_none, 6, 16, 16, KeenWorldLeaveElevator, NULL, DrawReact, NULL};

void EnterWorldElevator(Uint16 tileX, Uint16 tileY, Sint16 dir)
{
	Uint16 info, globalx, globaly, duration, move;
	Sint16 i, x, y;
	Uint16 tiles[4];
	objtype *ob = player;

	globalx = CONVERT_TILE_TO_GLOBAL(tileX);
	globaly = CONVERT_TILE_TO_GLOBAL(tileY);

	//
	// make Keen walk into the elevator
	//
	for (duration = 0; duration < 130; )
	{
		CalcBounds(ob);
		ScrollScreenWorld(ob);
		UpdateScorebox(scoreobj);
		RF_Refresh();

		move = tics * 2;
		duration += tics;

		if (ob->x == globalx && ob->y == globaly)
			break;

		if (ob->y < globaly)
		{
			ob->y += move;
			if (ob->y > globaly)
				ob->y = globaly;
		}
		else if (ob->y > globaly)
		{
			ob->y -= move;
			if (ob->y < globaly)
				ob->y = globaly;
		}

		if (ob->x < globalx)
		{
			ob->x += move;
			if (ob->x > globalx)
				ob->x = globalx;
		}
		else if (ob->x > globalx)
		{
			ob->x -= move;
			if (ob->x < globalx)
				ob->x = globalx;
		}

		ob->shapenum = ((duration / 8) % 3) + WORLDKEENU1SPR;
		RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, spritedraw, ob->priority);
	}

	//
	// close the elevator door
	//
	SD_PlaySound(SND_ELEVATORDOOR);
	for (i=5; i >= 0; i--)
	{
		for (y=0; y<2; y++)
		{
			for (x=0; x<2; x++)
			{
				tiles[y*2 + x] = mapsegs[1][mapbwidthtable[y]/2 + i*2 + x];
			}
		}
		RF_MemToMap(tiles, 1, tileX+dir, tileY-1, 2, 2);
		RF_Refresh();
		VW_WaitVBL(8);
	}

	//
	// make Keen invisible (and not clipping) and send him to the destination
	//
	RF_RemoveSprite(&ob->sprite);
	info = mapsegs[2][mapbwidthtable[tileY]/2 + tileX];
	ob->temp2 = CONVERT_TILE_TO_GLOBAL(info >> 8);
	ob->temp1 = CONVERT_TILE_TO_GLOBAL((info & 0xFF) + 1);
	if (ob->temp1 < ob->y)
	{
		ob->ydir = -1;
	}
	else
	{
		ob->ydir = 1;
	}
	if (ob->temp2 < ob->x)
	{
		ob->xdir = -1;
	}
	else
	{
		ob->xdir = 1;
	}
	ob->needtoclip = false;
	ob->state = &s_worldkeenhidden;
}

#endif	//ifdef KEEN5

#endif	//ifdef KEEN4 ... else ...

void CheckWorldTiles(objtype *ob)
{
	Uint16 tx, ty, intile;

	if (ob->temp3)
		return;

	tx = ob->tilemidx;
	ty = CONVERT_GLOBAL_TO_TILE(ob->top + (ob->bottom-ob->top)/2);
	intile = tinf[INTILE + mapsegs[1][mapbwidthtable[ty]/2+tx]];
#if defined KEEN4
	if (intile == INTILE_SHORESOUTH || intile == INTILE_SHORENORTH
		|| intile == INTILE_SHOREEAST || intile == INTILE_SHOREWEST)
	{
		if (!gamestate.wetsuit)
		{
			SD_PlaySound(SND_NOWAY);
			CantSwimMessage();
			RF_ForceRefresh();
			tmp_xmove = -ob->xmove;
			tmp_ymove = -ob->ymove;
			ob->xdir = 0;
			ob->ydir = 0;
			ClipToWalls(ob);
		}
		else
		{
			ob->temp1 = shoredir[intile-INTILE_SHORESOUTH];
			if (ob->state == &s_worldkeenswim)
			{
				ob->temp1 = oppositedir[ob->temp1];
			}
			switch (ob->temp1)
			{
			case dir_North:
				ob->xdir = 0;
				ob->ydir = -1;
				break;
			case dir_East:
				ob->xdir = 1;
				ob->ydir = 0;
				break;
			case dir_South:
				ob->xdir = 0;
				ob->ydir = 1;
				break;
			case dir_West:
				ob->xdir = -1;
				ob->ydir = 0;
				break;
			}
			ob->temp2 = 0;
			ob->temp3 = 18;
			if (ob->state == &s_worldkeenswim)
			{
				ChangeState(ob, &s_worldkeenwalk);
			}
			else
			{
				ChangeState(ob, &s_worldkeenswim);
			}
		}
	}
#elif defined KEEN5
	switch (intile)
	{
	case INTILE_TELEPORT:
		EnterWorldTeleporter(tx, ty);
		break;
	case INTILE_ELEVATORLEFT:
		EnterWorldElevator(tx, ty, 0);
		break;
	case INTILE_ELEVATORRIGHT:
		EnterWorldElevator(tx, ty, -1);
		break;
	}
#elif defined KEEN6
	switch (intile)
	{
	case INTILE_TELEPORT:
		EnterWorldTeleporter(tx, ty);
		break;
	}
#endif
}

//============================================================================
//============================================================================

statetype s_flag1         = {FLAG1SPR, FLAG1SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_flag2};
statetype s_flag2         = {FLAG2SPR, FLAG2SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_flag3};
statetype s_flag3         = {FLAG3SPR, FLAG3SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_flag4};
statetype s_flag4         = {FLAG4SPR, FLAG4SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_flag1};

#ifndef KEEN5
statetype s_flagspin1     = {FLAGTHROW1SPR, FLAGTHROW1SPR, think, false, push_none, 6, 0, 0, FlagThink1, NULL, DrawReact, &s_flagspin2};
statetype s_flagspin2     = {FLAGTHROW1SPR, FLAGTHROW1SPR, stepthink, false, push_none, 12, 0, 0, FlagThink2, NULL, DrawReact, &s_flagspin3};
statetype s_flagspin3     = {FLAGTHROW2SPR, FLAGTHROW2SPR, stepthink, false, push_none, 12, 0, 0, FlagThink2, NULL, DrawReact, &s_flagspin4};
statetype s_flagspin4     = {FLAGTHROW3SPR, FLAGTHROW3SPR, stepthink, false, push_none, 12, 0, 0, FlagThink2, NULL, DrawReact, &s_flagspin5};
statetype s_flagspin5     = {FLAGTHROW4SPR, FLAGTHROW4SPR, stepthink, false, push_none, 12, 0, 0, FlagThink2, NULL, DrawReact, &s_flagspin6};
statetype s_flagspin6     = {FLAGTHROW6SPR, FLAGTHROW6SPR, stepthink, false, push_none, 12, 0, 0, FlagThink2, NULL, DrawReact, &s_flagspin7};
statetype s_flagspin7     = {FLAGTHROW6SPR, FLAGTHROW6SPR, stepthink, true, push_none, 1, 0, 0, FlagLandThink, NULL, DrawReact, &s_flag1};
#endif

Point flagpath[30], flagpos;

void PlaceFlag(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->needtoclip = false;
	new->priority = 3;
	new->obclass = flagobj;
	new->active = yes;
#if defined KEEN4
	new->x = CONVERT_TILE_TO_GLOBAL(x) + 6*PIXGLOBAL;
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -30*PIXGLOBAL;
#elif defined KEEN5
	new->x = CONVERT_TILE_TO_GLOBAL(x) + -5*PIXGLOBAL;
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -30*PIXGLOBAL;
#elif defined KEEN6
	new->x = CONVERT_TILE_TO_GLOBAL(x) + 6*PIXGLOBAL;		// useless!
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -30*PIXGLOBAL;	// useless!
	new->x = CONVERT_TILE_TO_GLOBAL(x) + 14*PIXGLOBAL;
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -26*PIXGLOBAL;
	{
		Uint16 tile = mapsegs[1][mapbwidthtable[y]/2 + x] + 1;
		RF_MemToMap(&tile, 1, x, y, 1, 1);
	}
#endif
	new->ticcount = US_RndT() / 16;
	NewState(new, &s_flag1);
}

#ifndef KEEN5
void SpawnFlag(Sint16 x, Sint16 y)
{
	Sint16 i;
	Sint32 xdist, ydist;

	GetNewObj(false);
	new->needtoclip = false;
	new->priority = 3;
	new->obclass = flagobj;
	new->active = allways;
	new->x = gamestate.worldx - 16*PIXGLOBAL;
	new->y = gamestate.worldy - 16*PIXGLOBAL;
#if defined KEEN4
	new->temp1 = CONVERT_TILE_TO_GLOBAL(x) + 6*PIXGLOBAL;
	new->temp2 = CONVERT_TILE_TO_GLOBAL(y) + -38*PIXGLOBAL;
#elif defined KEEN6
	flagpos.x = x;
	flagpos.y = y;
	new->temp1 = CONVERT_TILE_TO_GLOBAL(x) + 14*PIXGLOBAL;
	new->temp2 = CONVERT_TILE_TO_GLOBAL(y) + -34*PIXGLOBAL;
#endif
	xdist = (Sint32)new->temp1 - (Sint32)new->x;
	ydist = (Sint32)new->temp2 - (Sint32)new->y;
	for (i = 0; i < 30; i++)
	{
		flagpath[i].x = new->x + (xdist * min(i, 24))/24;
		flagpath[i].y = new->y + (ydist * i)/30;
		if (i < 10)
		{
			flagpath[i].y -= i*3*PIXGLOBAL;
		}
		else if (i < 15)
		{
			flagpath[i].y -= i*PIXGLOBAL + 20*PIXGLOBAL;
		}
		else if (i < 20)
		{
			flagpath[i].y -= (20-i)*PIXGLOBAL + 30*PIXGLOBAL;
		}
		else
		{
			flagpath[i].y -= (29-i)*3*PIXGLOBAL;
		}
	}
	NewState(new, &s_flagspin1);
}

void FlagThink1(objtype *ob)
{
	if (screenfaded)
		return;

	SD_StopSound();
	SD_PlaySound(SND_FLAGSPIN);
	ob->state = ob->state->nextstate;
}

void FlagThink2(objtype *ob)
{
	ob->temp3 = ob->temp3 + tics;
	if (ob->temp3 > 58)
		ob->temp3 = 58;

	ob->x = flagpath[ob->temp3/2].x;
	ob->y = flagpath[ob->temp3/2].y;
	ob->needtoreact = true;
	if (ob->temp1 == 0)
	{
		SD_PlaySound(SND_FLAGSPIN);
	}
}

void FlagLandThink(objtype *ob)
{
	ob->x = ob->temp1;
	ob->y = ob->temp2 + 8*PIXGLOBAL;
	SD_PlaySound(SND_FLAGLAND);
#ifdef KEEN6
	{
		Uint16 tile = mapsegs[1][mapbwidthtable[flagpos.y]/2 + flagpos.x] + 1;
		RF_MemToMap(&tile, 1, flagpos.x, flagpos.y, 1, 1);
	}
#endif
}
#endif

//============================================================================
//============================================================================

statetype s_keenshot1     = {KEENSHOT1SPR, KEENSHOT1SPR, slide, false, push_none, 6, 64, 64, KeenShotThink, NULL, KeenShotReact, &s_keenshot2};
statetype s_keenshot2     = {KEENSHOT2SPR, KEENSHOT2SPR, slide, false, push_none, 6, 64, 64, KeenShotThink, NULL, KeenShotReact, &s_keenshot3};
statetype s_keenshot3     = {KEENSHOT3SPR, KEENSHOT3SPR, slide, false, push_none, 6, 64, 64, KeenShotThink, NULL, KeenShotReact, &s_keenshot4};
statetype s_keenshot4     = {KEENSHOT4SPR, KEENSHOT4SPR, slide, false, push_none, 6, 64, 64, KeenShotThink, NULL, KeenShotReact, &s_keenshot1};

statetype s_keenshotexplode1 = {KEENSHOTEXPLODE1SPR, KEENSHOTEXPLODE1SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, &s_keenshotexplode2};
statetype s_keenshotexplode2 = {KEENSHOTEXPLODE2SPR, KEENSHOTEXPLODE2SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, NULL};

void SpawnShot(Uint16 x, Uint16 y, Direction dir)
{
	if (!gamestate.ammo)
	{
		SD_PlaySound(SND_USESWITCH);
		return;
	}

	gamestate.ammo--;
	GetNewObj(true);
	new->x = x;
	new->y = y;
	new->priority = 0;
	new->obclass = shotobj;
	new->active = allways;
	SD_PlaySound(SND_KEENFIRE);
	switch (dir)
	{
	case dir_North:
		new->xdir = 0;
		new->ydir = -1;
		break;
	case dir_East:
		new->xdir = 1;
		new->ydir = 0;
		break;
	case dir_South:
		new->xdir = 0;
		new->ydir = 1;
		break;
	case dir_West:
		new->xdir = -1;
		new->ydir = 0;
		break;
	default:
		Quit("SpawnShot: Bad dir!");
		break;
	}
	NewState(new, &s_keenshot1);

#ifdef KEEN6
	{
		objtype *ob;

		for (ob=player->next; ob; ob=ob->next)
		{
			if (ob->active
				&& new->right > ob->left && new->left < ob->right
				&& new->top < ob->bottom && new->bottom > ob->top
				&& ob->state->contact)
			{
				ob->state->contact(ob, new);
				return;
			}
		}
	}
#endif
}

void ExplodeShot(objtype *ob)
{
	ob->obclass = decoobj;
	ChangeState(ob, &s_keenshotexplode1);
	SD_PlaySound(SND_SHOTEXPLODE);
}

void KeenShotThink(objtype *ob)
{
	objtype *ob2;

	if (ob->tileright >= originxtile && ob->tilebottom >= originytile
		&& ob->tileleft <= originxtilemax && ob->tiletop <= originytilemax)
	{
		//object is visible, so do nothing
		return;
	}

	if (ob->tileright+10 < originxtile
		|| ob->tileleft-10 > originxtilemax
		|| ob->tilebottom+6 < originytile
		|| ob->tiletop+6 > originytilemax)
	{
		//shot is off-screen by more than half a screen, so remove it
		RemoveObj(ob);
		return;
	}

	//check for collisions with INACTIVE objects
	for (ob2 = player; ob2; ob2 = ob2->next)
	{
		if (!ob2->active && ob->right > ob2->left && ob->left < ob2->right
			&& ob->top < ob2->bottom && ob->bottom > ob2->top)
		{
			if (ob2->state->contact)
			{
				ob2->state->contact(ob2, ob);
				ob2->needtoreact = true;
				ob2->active = yes;
			}

			if (ob->obclass == nothing)	//BUG: obclass is 'decoobj' for the exploded shot
				break;
		}
	}
}

void KeenShotReact(objtype *ob)
{
	Uint16 tile;

	if (ob->hitnorth == 1 && ob->tileleft != ob->tileright)
	{
		tile = mapsegs[1][mapbwidthtable[ob->tiletop-1]/2+ob->tileright];
		if (tinf[NORTHWALL+tile] == 17)
		{
			ob->hitnorth = 17;
			ob->x += 0x100 - (ob->x & 0xFF);
		}
	}
	else if (ob->hitnorth == 17 && ob->tileleft != ob->tileright)
	{
		ob->x &= 0xFF00;
	}
	if (ob->hitsouth == 1 && ob->tileleft != ob->tileright)
	{
		tile = mapsegs[1][mapbwidthtable[ob->tiletop-1]/2+ob->tileright];
		if (tinf[SOUTHWALL+tile] == 17)
		{
			ob->hitsouth = 17;
			ob->x += 0x100 - (ob->x & 0xFF);
		}
	}
	else if (ob->hitsouth == 17 && ob->tileleft != ob->tileright)
	{
		ob->x &= 0xFF00;
	}
	if (ob->hitnorth == 17 || ob->hitsouth == 17)
	{
		tmp_ymove = ob->state->ymove * tics * ob->ydir;
		ob->y += tmp_ymove;
		ob->top += tmp_ymove;
		ob->bottom += tmp_ymove;
		ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
		ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
	}
	else if (ob->hitnorth || ob->hitsouth || ob->hiteast || ob->hitwest)
	{
		ExplodeShot(ob);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

//============================================================================
//============================================================================

statetype s_gemdoor1    = {0, 0, step, false, push_none, 10, 0, 0, GemDoorThink, NULL, NULL, &s_gemdoor2};
statetype s_gemdoor2    = {0, 0, step, false, push_none, 10, 0, 0, GemDoorThink, NULL, NULL, &s_gemdoor3};
statetype s_gemdoor3    = {0, 0, step, false, push_none, 10, 0, 0, GemDoorThink, NULL, NULL, NULL};

void GemDoorThink(objtype *ob)
{
	Sint16 i;
	Uint16 far *map;
	Uint16 tiles[50];

	map = mapsegs[1] + mapbwidthtable[ob->y]/2 + ob->x;
	for (i=0; i < ob->temp1+2; i++, map+=mapwidth)
	{
		tiles[i] = *map + 1;
	}
	RF_MemToMap(tiles, 1, ob->x, ob->y, 1, ob->temp1+2);
}

#ifdef KEEN5
//============================================================================
//============================================================================

statetype s_dooropener    = {0, 0, step, false, push_none, 15, 0, 0, DoorOpenerThink, NULL, NULL, &s_dooropener};

void DoorOpenerThink(objtype *ob)
{
	Uint16 x, y;
	Uint16 far *map;
	Uint16 tiles[16], *tileptr;

	tileptr = tiles;
	map = mapsegs[1] + mapbwidthtable[ob->y]/2 + ob->x;
	for (y=0; y<4; y++, map+=mapwidth)
	{
		for (x=0; x<4; x++)
		{
			*tileptr++ = map[x]-4;
		}
	}
	RF_MemToMap(tiles, 1, ob->x, ob->y, 4, 4);

	if (++ob->temp1 == 3)
		ob->state = NULL;
}

#endif