/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "CK_DEF.H"

/////////////////////////////////////////////////////////////////////////////
// initialized variables:
/////////////////////////////////////////////////////////////////////////////

ScanCode firescan = sc_Space;

static Sint16 songs[] =
{
#if defined KEEN4
	// this is the RTTS music list!
	WORLD_MUS,
	GOODTIME_MUS,
	SLACK_MUS,
	K4S4C_MUS,
	SLACK_MUS,
	SLACK_MUS,
	K4S4C_MUS,
	GOODTIME_MUS,
	SLACK_MUS,
	SLOWBURN_MUS,
	SLOWBURN_MUS,
	SLOWBURN_MUS,
	GOODTIME_MUS,
	SLOWBURN_MUS,
	SLOWBURN_MUS,
	K4S4C_MUS,
	SLACK_MUS,
	DOPEY_MUS,
	SLACK_MUS,
	GOODTIME_MUS,
	WONDER_MUS
#elif defined KEEN5
	TRACK11_MUS,
	BEACH_MUS,
	TRACK07_MUS,
	SPHEREFUL_MUS,
	TIGHTER_MUS,
	SPHEREFUL_MUS,
	TIGHTER_MUS,
	SPHEREFUL_MUS,
	TIGHTER_MUS,
	SPHEREFUL_MUS,
	TIGHTER_MUS,
	SNOOPIN1_MUS,
	TRACK13_MUS,
	BAGPIPES_MUS,
	FNFARE01_MUS,
	TRACK02_MUS,
	TRACK06_MUS,
	TRACK01_MUS,
	CAMEIN_MUS,
	TRACK08_MUS,
#elif defined KEEN6
	ALIENATE_MUS,
	FASTER_MUS,
	BRERTAR_MUS,
	MAMSNAKE_MUS,
	MAMSNAKE_MUS,
	MAMSNAKE_MUS,
	METAL_MUS,
	TOFUTURE_MUS,
	METAL_MUS,
	BRERTAR_MUS,
	FASTER_MUS,
	TOFUTURE_MUS,
	BRERTAR_MUS,
	SPACFUNK_MUS,
	SPACFUNK_MUS,
	OMINOUS_MUS,
	TOFUTURE_MUS,
	WONDER_MUS,
	WONDER_MUS,
	WONDER_MUS
#endif
};

/////////////////////////////////////////////////////////////////////////////
// uninitialized variables:
/////////////////////////////////////////////////////////////////////////////

exittype playstate;
gametype gamestate;
boolean debugmode, noscrolling;

Sint16 pogotics;
#ifdef KEEN6
Sint16 shaketics;
#endif

boolean statushide;
Sint16 statusy;
Uint16 statusbufferofs;

Uint16 extravbls;

ControlInfo c;

objtype objarray[MAXACTORS];
objtype dummyobj;
objtype *objfreelist;
objtype *lastobj;
objtype *new;
objtype *player;
objtype *scoreobj;
objtype *obj;
Uint16 objectcount;

Uint16 originxtilemax;
Uint16 originytilemax;
Sint16 inactivateleft;
Sint16 inactivateright;
Sint16 inactivatetop;
Sint16 inactivatebottom;
Uint16 camyoff;

boolean upKeyHeld;
boolean button0down, button0held, button1down, button1held, button2down, button2held;

boolean oldshooting, showscorebox;

Sint16 invulnerable;

/////////////////////////////////////////////////////////////////////////////
// functions:
/////////////////////////////////////////////////////////////////////////////

static void CountObjects(void)
{
	Uint16 activeobjects, inactiveobjects;
	objtype *ob;

	activeobjects = inactiveobjects = 0;
	for (ob=player; ob; ob=ob->next)
	{
		if (ob->active)
		{
			activeobjects++;
		}
		else
		{
			inactiveobjects++;
		}
	}
	VW_FixRefreshBuffer();
	US_CenterWindow(18, 4);
	PrintY += 7;
	US_Print("Active Objects :");
	US_PrintUnsigned(activeobjects);
	US_Print("\nInactive Objects:");
	US_PrintUnsigned(inactiveobjects);
	VW_UpdateScreen();
	IN_Ack();
}

static void DebugMemory(void)
{
	VW_FixRefreshBuffer();
	US_CenterWindow(16, 7);
	US_CPrint("Memory Usage");
	US_CPrint("------------");
	US_Print("Total     :");
	US_PrintUnsigned((mminfo.mainmem+mminfo.EMSmem+mminfo.XMSmem)/1024);
	US_Print("k\nFree      :");
	US_PrintUnsigned(MM_UnusedMemory()/1024);
	US_Print("k\nWith purge:");
	US_PrintUnsigned(MM_TotalFree()/1024);
	US_Print("k\n");
	VW_UpdateScreen();
	IN_Ack();
	MM_ShowMemory();
}

static void TestSprites(void)
{
	Uint16 infox, infoy;
	Sint16 chunk, oldchunk;
	Sint16 shift;
	Uint16 infobottom, drawx;
	spritetabletype far *info;
	Uint8 _seg *block;
	Uint16 size;
	Uint16 scan;
	Uint32 totalsize;

	VW_FixRefreshBuffer();
	US_CenterWindow(30, 17);
	totalsize = 0;
	US_CPrint("Sprite Test");
	US_CPrint("-----------");
	infoy = PrintY;
	infox = (PrintX + 56) & SCREENXMASK;
	drawx = infox + 40;
	US_Print("Chunk:\nWidth:\nHeight:\nOrgx:\nOrgy:\nXl:\nYl:\nXh:\nYh:\nShifts:\nMem:\n");
	infobottom = PrintY;
	chunk = STARTSPRITES;
	shift = 0;
	while (1)
	{
		if (chunk >= STARTSPRITES+NUMSPRITES)
		{
			chunk = STARTSPRITES+NUMSPRITES-1;
		}
		else if (chunk < STARTSPRITES)
		{
			chunk = STARTSPRITES;
		}
		info = &spritetable[chunk-STARTSPRITES];
		block = grsegs[chunk];
		VWB_Bar(infox, infoy, 40, infobottom-infoy, WHITE);
		PrintX = infox;
		PrintY = infoy;
		US_PrintUnsigned(chunk);
		US_Print("\n");
		PrintX = infox;
		US_PrintUnsigned(info->width);
		US_Print("\n");
		PrintX = infox;
		US_PrintUnsigned(info->height);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->orgx);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->orgy);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->xl);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->yl);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->xh);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->yh);
		US_Print("\n");
		PrintX = infox;
		US_PrintSigned(info->shifts);
		US_Print("\n");
		PrintX = infox;
		if (!block)
		{
			US_Print("-----");
		}
		else
		{
			size = ((spritetype far *)block)->sourceoffset[3] + ((spritetype far *)block)->planesize[3]*5;
			size = (size + 15) & ~15;	//round up to multiples of 16
			totalsize += size;	//useless: the value stored in 'totalsize' is never used
			US_PrintUnsigned(size);
			US_Print("=");
		}
		oldchunk = chunk;
		do
		{
			VWB_Bar(drawx, infoy, 110, infobottom-infoy, WHITE);
			if (block)
			{
				PrintX = drawx;
				PrintY = infoy;
				US_Print("Shift:");
				US_PrintUnsigned(shift);
				US_Print("\n");
				VWB_DrawSprite(drawx + 2*shift + 16, PrintY, chunk);
			}
			VW_UpdateScreen();
			scan = IN_WaitForKey();
			switch (scan)
			{
			case sc_UpArrow:
				chunk++;
				break;
			case sc_DownArrow:
				chunk--;
				break;
			case sc_PgUp:
				chunk += 10;
				if (chunk >= STARTSPRITES+NUMSPRITES)
				{
					chunk = STARTSPRITES+NUMSPRITES-1;
				}
				break;
			case sc_PgDn:
				chunk -= 10;
				if (chunk < STARTSPRITES)
				{
					chunk = STARTSPRITES;
				}
				break;
			case sc_LeftArrow:
				shift--;
				if (shift == -1)
				{
					shift = 3;
				}
				break;
			case sc_RightArrow:
				shift++;
				if (shift == 4)
				{
					shift = 0;
				}
				break;
			case sc_Escape:
				return;
			}
		} while (chunk == oldchunk);
	}
}

static void PicturePause(void)
{
	Uint16 offset;
	Sint16 y;

	IN_ClearKeysDown();
	while (!LastScan);
	if (LastScan != sc_Enter)
	{
		IN_ClearKeysDown();
		return;
	}
	SD_PlaySound(SND_JUMP);
	SD_WaitSoundDone();
	offset = displayofs + panadjust;
	VW_ColorBorder(WHITE);
	VW_SetLineWidth(40);
	VW_SetScreen(0, 0);
	if (offset < 0xCE00)
	{
		for (y=0; y<200; y++)
		{
			VW_ScreenToScreen(offset+y*64, y*40, 40, 1);
		}
	}
	else
	{
		for (y=199; y>=0; y--)
		{
			VW_ScreenToScreen(offset+y*64, y*40, 40, 1);
		}
	}
	IN_Shutdown();
	SD_PlaySound(SND_EXTRAKEEN);
	SD_WaitSoundDone();
	US_Shutdown();
	SD_Shutdown();
	IN_Shutdown();
	RF_Shutdown();
	CA_Shutdown();
	MM_Shutdown();
	while (((bioskey(0) >> 8) & 0xFF) != sc_Escape);
	VW_Shutdown();
	exit(0);
}

static void MergeTile16M(Uint16 dest, Uint16 source)
{
	Uint16 _seg *sourceseg;
	Uint16 _seg *destseg;
	Uint16 sourceval, sourcemask, maskindex;
	Sint16 i;

	sourceseg = grsegs[STARTTILE16M+source];
	destseg = grsegs[STARTTILE16M+dest];
	for (i=0; i<64; i++)
	{
		maskindex = i & 15;
		sourceval = sourceseg[16+i];
		sourcemask = sourceseg[maskindex];
		destseg[maskindex] &= sourcemask;
		destseg[16+i] &= sourcemask;
		destseg[16+i] |= sourceval;
	}
}

static void WallDebug(void)
{
	Sint16 i, val;

	// Note: the -18 adjustments are just for Return To The Shadowlands

	VW_FixRefreshBuffer();
	US_CenterWindow(24, 3);
	US_PrintCentered("WORKING");
	VW_UpdateScreen();
	for (i=STARTTILE16M+108-18; i<STARTTILE16M+124-18; i++)
	{
		CA_CacheGrChunk(i);
	}
	for (i=0; i<NUMTILE16M; i++)
	{
		if (!grsegs[STARTTILE16M+i])
		{
			continue;
		}
		val = tinf[i+NORTHWALL] & 7;
		if (val)
		{
			MergeTile16M(i, val+107-18);
		}
		val = tinf[i+SOUTHWALL] & 7;
		if (val)
		{
			MergeTile16M(i, val+115-18);
		}
		val = tinf[i+EASTWALL] & 7;
		if (val > 1)
		{
			strcpy(str, "WallDebug: East wall other than 1:");
			itoa(i, str2, 10);
			strcat(str, str2);
			Quit(str);
		}
		if (val)
		{
			MergeTile16M(i, val+114-18);	//Note: val is always 1 here, so you could use 115 as 2nd arg
		}
		val = tinf[i+WESTWALL] & 7;
		if (val > 1)
		{
			strcpy(str, "WallDebug: West wall other than 1:");
			itoa(i, str2, 10);
			strcat(str, str2);
			Quit(str);
		}
		if (val)
		{
			MergeTile16M(i, val+122-18);	//Note: val is always 1 here, so you could use 123 as 2nd arg
		}
	}
}

static boolean DebugKeys(void)
{
	Sint16 level, i, esc;

	if (Keyboard[sc_B] && ingame)
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(24, 3);
		PrintY += 6;
		US_Print(" Border color (0-15):");
		VW_UpdateScreen();
		esc = !US_LineInput(px, py, str, NULL, true, 2, 0);
		if (!esc)
		{
			level = atoi(str);
			if (level >= 0 && level <= 15)
			{
				VW_ColorBorder(level);
			}
		}
		return true;
	}
	if (Keyboard[sc_C] && ingame)
	{
		CountObjects();
		return true;
	}
	if (Keyboard[sc_D] && ingame)
	{
		if (DemoMode == demo_Off)
		{
			StartDemoRecord();
		}
		else if (DemoMode == demo_Record)
		{
			EndDemoRecord();
			playstate = ex_completed;
		}
		return true;
	}
	if (Keyboard[sc_E] && ingame)
	{
		if (tedlevel)
		{
			TEDDeath();
		}
		playstate = ex_completed;
		//BUG? there is no return in this branch (should return false)
	}
	if (Keyboard[sc_G] && ingame)
	{
		US_CenterWindow(12, 2);
		if (godmode)
		{
			US_PrintCentered("God mode OFF");
		}
		else
		{
			US_PrintCentered("God mode ON");
		}
		VW_UpdateScreen();
		IN_Ack();
		godmode ^= true;
		return true;
	}
	else if (Keyboard[sc_I])
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(12, 3);
		US_PrintCentered("Free items!");
		for (i=0; i<4; i++)
		{
			gamestate.keys[i] = 99;
		}
		gamestate.ammo = 99;
#if defined KEEN4
		gamestate.wetsuit = true;
#elif defined KEEN5
		gamestate.keycard = true;
#elif defined KEEN6
		gamestate.passcardstate=gamestate.ropestate=gamestate.sandwichstate = 1;
#endif
		VW_UpdateScreen();
		IN_Ack();
		//GivePoints(3000);	// RTTS mod: don't give points
		return true;
	}
	else if (Keyboard[sc_J] && ingame)
	{
		jumpcheat ^= true;
		VW_FixRefreshBuffer();
		US_CenterWindow(18, 3);
		if (jumpcheat)
		{
			US_PrintCentered("Jump cheat ON");
		}
		else
		{
			US_PrintCentered("Jump cheat OFF");
		}
		VW_UpdateScreen();
		IN_Ack();
		return true;
	}
	else if (Keyboard[sc_M])
	{
		DebugMemory();
		return true;
	}
	else if (Keyboard[sc_N])
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(18, 3);
		if (player->needtoclip)
		{
			US_PrintCentered("No clipping ON");
			player->needtoclip = false;
		}
		else
		{
			US_PrintCentered("No clipping OFF");
			player->needtoclip = true;
		}
		VW_UpdateScreen();
		IN_Ack();
		return true;
	}
	else if (Keyboard[sc_P])
	{
		IN_ClearKeysDown();
		PicturePause();
		return true;
	}
	else if (Keyboard[sc_S] && ingame)
	{
		singlestep ^= true;
		VW_FixRefreshBuffer();
		US_CenterWindow(18, 3);
		if (singlestep)
		{
			US_PrintCentered("Slow motion ON");
		}
		else
		{
			US_PrintCentered("Slow motion OFF");
		}
		VW_UpdateScreen();
		IN_Ack();
		return true;
	}
	else if (Keyboard[sc_T])
	{
		TestSprites();
		return true;
	}
	else if (Keyboard[sc_V])
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(30, 3);
		PrintY += 6;
		US_Print("  Add how many extra VBLs(0-8):");
		VW_UpdateScreen();
		esc = !US_LineInput(px, py, str, NULL, true, 2, 0);
		if (!esc)
		{
			level = atoi(str);
			if (level >= 0 && level <= 8)
			{
				extravbls = level;
			}
		}
		return true;
	}
	else if (Keyboard[sc_W] && ingame)
	{
		VW_FixRefreshBuffer();
		US_CenterWindow(26, 3);
		PrintY += 6;
		US_Print("  Warp to which level(1-20):");	// RTTS mod: allow up to 20 levels
		VW_UpdateScreen();
		esc = !US_LineInput(px, py, str, NULL, true, 2, 0);
		if (!esc)
		{
			level = atoi(str);
			if (level > 0 && level <= 20)	// RTTS mod: allow up to 20 levels
			{
				gamestate.mapon = level;
				playstate = ex_warped;
			}
		}
		return true;
	}
	else if (Keyboard[sc_Y])
	{
		WallDebug();
		return true;
	}
	else if (Keyboard[sc_Z])
	{
		gamestate.lives = 0;
		KillKeen();
		return false;
	}
	return false;
}

static void CheatOption(void)
{
	Sint16 i;

	for (i=sc_A; i<=sc_Z; i++)	//Note: this does NOT check the keys in alphabetical order!
	{
		if (i != sc_B && i != sc_A && i != sc_T && Keyboard[i])
		{
			return;
		}
	}
	US_CenterWindow(20, 9);
	PrintY += 2;
	US_CPrint(
		"Cheat Option!\n"
		"\n"
		"You just got all\n"
		"the keys, 99 shots\n"
		"and an extra keen!");
	VW_UpdateScreen();
	IN_Ack();
	RF_ForceRefresh();
	gamestate.ammo = 99;
	gamestate.lives++;
#ifdef KEEN5
	gamestate.keycard = true;
#endif
	gamestate.keys[0] = gamestate.keys[1] = gamestate.keys[2] = gamestate.keys[3] = 1;
}

void CheckKeys(void)
{
	if (screenfaded)
	{
		return;
	}
	if (Keyboard[sc_Enter] || (gravisOn && statusBtnHit))
	{
		StatusWindow();
		IN_ClearKeysDown();
		RF_ForceRefresh();
		lasttimecount = TimeCount;
	}
	if (Paused)
	{
		SD_MusicOff();
		VW_FixRefreshBuffer();
		US_CenterWindow(8, 3);
		US_PrintCentered("PAUSED");
		VW_UpdateScreen();
		IN_Ack();
		RF_ForceRefresh();
		Paused = false;
		SD_MusicOn();
	}
#if 0
//video memory dump hack:
if (Keyboard[sc_D])
{
	Sint16 handle;
	while (Keyboard[sc_D]);
	handle = open("mem.dat", O_WRONLY|O_BINARY|O_CREAT, S_IWRITE);
	if (handle != -1)
	{
		Sint16 i;
		outp(0x3CE, 4);
		for (i=0; i<4; i++)
		{
			outp(0x3CF, i);
			CA_FarWrite(handle, MK_FP(0xA000, 0), 0x8000);
			CA_FarWrite(handle, MK_FP(0xA000, 0x8000), 0x8000);
		}
		close(handle);
	}
}
#endif
#ifndef KEEN6
	if (LastScan == sc_F1)
	{
		StopMusic();
		HelpScreens();
		StartMusic(gamestate.mapon);
		if (showscorebox)
		{
			scoreobj->temp2 = -1;
			scoreobj->temp1 = -1;
			scoreobj->temp3 = -1;
			scoreobj->temp4 = -1;
		}
		RF_ForceRefresh();
	}
#endif
	if (!demo)
	{
		if (LastScan >= sc_F2 && LastScan <= sc_F7 || LastScan == sc_Escape)
		{
			VW_FixRefreshBuffer();
			StopMusic();
			US_ControlPanel();
			RF_FixOfs();
			StartMusic(gamestate.mapon);
			if (!showscorebox && scoreobj->sprite)
			{
				RF_RemoveSprite(&scoreobj->sprite);
			}
			if (showscorebox)
			{
				scoreobj->temp2 = -1;
				scoreobj->temp1 = -1;
				scoreobj->temp3 = -1;
				scoreobj->temp4 = -1;
			}
			IN_ClearKeysDown();
			if (restartgame)
			{
				playstate = ex_resetgame;
			}
			else if (!loadedgame)
			{
				RF_ForceRefresh();
			}
			if (abortgame)
			{
				abortgame = false;
				playstate = ex_abort;
			}
			if (loadedgame)
			{
				playstate = ex_loadedgame;
			}
			lasttimecount = TimeCount;
		}
		if (LastScan == sc_F9)
		{
			VW_Shutdown();
			SD_MusicOff();
			cputs("C:>");
			IN_ClearKeysDown();
			while (LastScan != sc_Escape);
			VW_SetScreenMode(GRMODE);
			VW_ColorBorder(bordercolor);
			RF_ForceRefresh();
			IN_ClearKeysDown();
			lasttimecount = TimeCount;
			SD_MusicOn();
		}
	}
	if (Keyboard[sc_B] && Keyboard[sc_A] && Keyboard[sc_T])
	{
		CheatOption();
	}
	if (debugmode && Keyboard[sc_F10])
	{
		if (DebugKeys())
		{
			RF_ForceRefresh();
			lasttimecount = TimeCount;
		}
	}
	if (demo && Keyboard[sc_Control] && LastScan == sc_S)
	{
		if (SoundMode != sdm_Off)
		{
			SD_SetSoundMode(sdm_Off);
			SD_SetMusicMode(smm_Off);
		}
		else
		{
			if (AdLibPresent)
			{
				SD_SetSoundMode(sdm_AdLib);
				QuietAdLib = false;
				SD_SetMusicMode(smm_AdLib);
			}
			else
			{
				SD_SetSoundMode(sdm_PC);
				SD_SetMusicMode(smm_Off);
			}
			CA_LoadAllSounds();
		}
	}
	if (Keyboard[sc_Control] && LastScan == sc_Q)
	{
		IN_ClearKeysDown();
		Quit(NULL);
	}
}

static void DrawStatusNumber(Sint16 x, Sint16 y, Sint16 maxlen, Sint16 basetile, Sint32 number)
{
	register Sint16 i;
	Sint16 len;
	char buffer[20];

	ltoa(number, buffer, 10);
	len = strlen(buffer);
	i = maxlen;
	while (i>len)
	{
		VWB_DrawTile8(x, y, basetile);
		i--;
		x += 8;
	}
	while (i>0)
	{
		VWB_DrawTile8(x, y, basetile+buffer[len-i]+(1-'0'));
		i--;
		x += 8;
	}
}

static void DrawStatus(void)
{
	Sint16 off, x, y, w, h, i;
	Uint16 width, height;

	x = 64;
	y = 16;
	w = 184;
	h = 144;
	VWB_DrawTile8(x, y, 54);
	VWB_DrawTile8(x, y+h, 60);
	for (i=x+8; i<=x+w-8; i+=8)
	{
		VWB_DrawTile8(i, y, 55);
		VWB_DrawTile8(i, y+h, 61);
	}
	VWB_DrawTile8(i, y, 56);
	VWB_DrawTile8(i, y+h, 62);
	for (i=y+8; i<=y+h-8; i+=8)
	{
		VWB_DrawTile8(x, i, 57);
		VWB_DrawTile8(x+w, i, 59);
	}
	VWB_Bar(72, 24, 176, 136, LIGHTGRAY);	//TODO: CGA support

	PrintY = 28;
	WindowX = 80;
	WindowW = 160;
	US_CPrint("LOCATION");
	VWB_Bar(79, 38, 162, 20, WHITE);
#ifdef KEEN5
	if (mapon == 0 && player->y > 100*TILEGLOBAL)
		_fstrcpy(str, levelnames[13]);
	else
		_fstrcpy(str, levelnames[gamestate.mapon]);
#else
	_fstrcpy(str, levelnames[gamestate.mapon]);
#endif
	MeasureTextBlock(str, &width, &height);
	PrintY = (20-height)/2+40-2;
	US_CPrint(str);

	PrintY = 61;
	WindowX = 80;
	WindowW = 64;
	US_CPrint("SCORE");
	VWB_Bar(79, 71, 66, 10, BLACK);
	DrawStatusNumber(80, 72, 8, 41, gamestate.score);

	PrintY = 61;
	WindowX = 176;
	WindowW = 64;
	US_CPrint("EXTRA");
	VWB_Bar(175, 71, 66, 10, BLACK);
	DrawStatusNumber(176, 72, 8, 41, gamestate.nextextra);

#if defined KEEN4
	PrintY = 85;
	WindowX = 80;
	WindowW = 64;
	US_CPrint("RESCUED");
	VWB_Bar(79, 95, 66, 10, BLACK);
	for (i = 0; i < gamestate.rescued; i++, off+=8)
	{
		VWB_DrawTile8(i*8 + 80, 96, 40);
	}
#elif defined KEEN5
	PrintY = 92;
	PrintX = 80;
	US_Print("KEYCARD");
	VWB_Bar(135, 91, 10, 10, BLACK);
	if (gamestate.keycard)
	{
		VWB_DrawTile8(136, 92, 40);
	}
#endif

	PrintY = 85;
	WindowX = 176;
	WindowW = 64;
	US_CPrint("LEVEL");
	VWB_Bar(175, 95, 66, 10, WHITE);
	PrintY = 96;
	WindowX = 176;
	WindowW = 64;
	switch (gamestate.difficulty)
	{
	case gd_Easy:
		US_CPrint("Easy");
		break;
	case gd_Normal:
		US_CPrint("Normal");
		break;
	case gd_Hard:
		US_CPrint("Hard");
		break;
	}

#ifdef KEEN6
	PrintX = 80;
	PrintY = 96;
	US_Print("ITEMS");
	VWB_Bar(127, 95, 26, 10, BLACK);
	if (gamestate.sandwichstate == 1)
	{
		VWB_DrawTile8(128, 96, 2);
	}
	else
	{
		VWB_DrawTile8(128, 96, 1);
	}
	if (gamestate.ropestate == 1)
	{
		VWB_DrawTile8(136, 96, 4);
	}
	else
	{
		VWB_DrawTile8(136, 96, 3);
	}
	if (gamestate.passcardstate == 1)
	{
		VWB_DrawTile8(144, 96, 6);
	}
	else
	{
		VWB_DrawTile8(144, 96, 5);
	}
#endif

	PrintX = 80;
	PrintY = 112;
	US_Print("KEYS");
	VWB_Bar(119, 111, 34, 10, BLACK);
	for (i = 0; i < 4; i++)
	{
		if (gamestate.keys[i])
		{
			VWB_DrawTile8(i*8+120, 112, 36+i);
		}
	}

	PrintX = 176;
	PrintY = 112;
	US_Print("AMMO");
	VWB_Bar(215, 111, 26, 10, BLACK);
	DrawStatusNumber(216, 112, 3, 41, gamestate.ammo);

	PrintX = 80;
	PrintY = 128;
	US_Print("KEENS");
	VWB_Bar(127, 127, 18, 10, BLACK);
	DrawStatusNumber(128, 128, 2, 41, gamestate.lives);

	PrintX = 176;
	PrintY = 128;
	US_Print(LIFECRUMBNAME);
	VWB_Bar(224, 127, 16, 10, BLACK);
	DrawStatusNumber(224, 128, 2, 41, gamestate.lifecrumbs);

#ifdef KEEN4
	VWB_Bar(79, 143, 66, 10, WHITE);
	PrintY = 144;
	WindowX = 80;
	WindowW = 64;
	if (gamestate.wetsuit)
	{
		US_CPrint("Wetsuit");
	}
	else
	{
		US_CPrint("???");
	}
#endif

	// draw the tiles for "PRESS A KEY":
	for (i = 0; i < 10; i++)
	{
		VWB_DrawTile8(i*8+STATUS_PRESSKEY_X, 140, i+72);
		VWB_DrawTile8(i*8+STATUS_PRESSKEY_X, 148, i+82);
	}
}

static void StatusRefreshHook(void)
{
	Uint16 source, dest;
	Sint16 height;

	if (statusy > 152)
	{
		height = statusy - 152;
		source = statusbufferofs + panadjust + 8;
		dest = bufferofs + panadjust + 8;
		VW_ScreenToScreen(source, dest, 24, height);
		VW_ClipDrawMPic((pansx+136)/8, -(16-height)+pansy, STATUSTOPPICM);
		source = statusbufferofs + panadjust + 0x400 + 8;
		dest = bufferofs + panadjust + height * 64 + 8;
		height = 152;
	}
	else
	{
		source = statusbufferofs + panadjust + (152-statusy)*64 + 0x400 + 8;
		dest = bufferofs + panadjust + 8;
		height = statusy;
	}
	if (height > 0)
	{
		VW_ScreenToScreen(source, dest, 24, height);
	}
	if (statushide)
	{
		height = 168-statusy;
		source = masterofs + panadjust + statusy*64 + 8;
		dest = bufferofs + panadjust + statusy*64 + 8;
		VW_ScreenToScreen(source, dest, 24, height);
		height = statusy;
		source = statusbufferofs + panadjust + 8 - 3;
		dest = bufferofs + panadjust + 8 - 3;
		if (height > 0)
			VW_ScreenToScreen(source, dest, 3, height);
	}
	else
	{
		height = statusy + -72;
		if (height > 0)
		{
			source = statusbufferofs + panadjust + 8 - 3;
			dest = bufferofs + panadjust + 8 - 3;
			if (height > 0)
				VW_ScreenToScreen(source, dest, 3, height);
		}
	}
	if (statusy >= 72)
	{
		VW_ClipDrawMPic((pansx+40)/8, statusy-168+pansy, STATUSLEFTPICM);
	}
	VW_UpdateScreen();
}

void StatusWindow(void)
{
	Uint16 oldbufferofs;

	WindowX = 0;
	WindowY = 320;
	WindowY = 0;
	WindowH = 200;

	if (Keyboard[sc_A] && Keyboard[sc_2])
	{
		US_CenterWindow(20, 2);
		PrintY += 2;
		US_Print("Debug keys active");
		VW_UpdateScreen();
		IN_Ack();
		debugmode = true;
	}
	RF_Refresh();
	RFL_InitAnimList();
	oldbufferofs = bufferofs;
	bufferofs = statusbufferofs = RF_FindFreeBuffer();
	VW_ScreenToScreen(displayofs, displayofs, 44, 224);
	VW_ScreenToScreen(displayofs, masterofs, 44, 224);
	VW_ScreenToScreen(displayofs, bufferofs, 44, 168);
	DrawStatus();
	bufferofs = oldbufferofs;
	RF_Refresh();
	SD_PlaySound(SND_SHOWSTATUS);
	statusy = 16;
	statushide = false;
	RF_SetRefreshHook(StatusRefreshHook);
	while (true)
	{
		RF_Refresh();
		if (statusy == 168)
			break;
		statusy = statusy + tics*8;
		if (statusy > 168)
			statusy = 168;
	}
	RF_Refresh();
	RF_SetRefreshHook(NULL);
	IN_ClearKeysDown();
	IN_Ack();
	SD_PlaySound(SND_HIDESTATUS);
	statusy -= 16;
	statushide = true;
	RF_SetRefreshHook(StatusRefreshHook);
	while (true)
	{
		RF_Refresh();
		if (statusy == 0)
			break;
		statusy = statusy - tics*8;
		if (statusy < 0)
			statusy = 0;
	}
	RF_SetRefreshHook(NULL);
	scoreobj->x = 0;	//force scorebox to redraw?
}

void CenterPlayer(objtype *ob)
{
	Uint16 orgx, orgy;

	camyoff = 140;
	if (ob->x < 152*PIXGLOBAL)
	{
		orgx = 0;
	}
	else
	{
		orgx = ob->x - 152*PIXGLOBAL;
	}
	if (mapon == 0)
	{
		if (ob->y < 80*PIXGLOBAL)
		{
			orgy = 0;
		}
		else
		{
			orgy = ob->y - 80*PIXGLOBAL;
		}
	}
	else
	{
		if (ob->bottom < 140*PIXGLOBAL)
		{
			orgy = 0;
		}
		else
		{
			orgy = ob->bottom - 140*PIXGLOBAL;
		}
	}
	if (!highscoredemo)
	{
		RF_NewPosition(orgx, orgy);
	}

	originxtilemax = (originxtile + PORTTILESWIDE) - 1;
	originytilemax = (originytile + PORTTILESHIGH) - 1;
	inactivateleft = originxtile - INACTIVATEDIST;
	if (inactivateleft < 0)
	{
		inactivateleft = 0;
	}
	inactivateright = originxtilemax + INACTIVATEDIST;
	if (inactivateright < 0)
	{
		inactivateright = 0;
	}
	inactivatetop = originytile - INACTIVATEDIST;
	if (inactivatetop < 0)
	{
		inactivatetop = 0;
	}
	inactivatebottom = originytilemax + INACTIVATEDIST;
	if (inactivatebottom < 0)
	{
		inactivatebottom = 0;
	}
}

void ScrollScreenWorld(objtype *ob)
{
	Sint16 xscroll, yscroll;

	if (noscrolling)
		return;

	if (ob->left < originxglobal + 9*TILEGLOBAL)
	{
		xscroll = ob->left - (originxglobal + 9*TILEGLOBAL);
	}
	else if (ob->right > originxglobal + 12*TILEGLOBAL)
	{
		xscroll = ob->right + 16 - (originxglobal + 12*TILEGLOBAL);
	}
	else
	{
		xscroll = 0;
	}

	if (ob->top < originyglobal + 5*TILEGLOBAL)
	{
		yscroll = ob->top - (originyglobal + 5*TILEGLOBAL);
	}
	else if (ob->bottom > originyglobal + 7*TILEGLOBAL)
	{
		yscroll = ob->bottom - (originyglobal + 7*TILEGLOBAL);
	}
	else
	{
		yscroll = 0;
	}

	if (!xscroll && !yscroll)
		return;

	if (xscroll >= 0x100)
	{
		xscroll = 0xFF;
	}
	else if (xscroll <= -0x100)
	{
		xscroll = -0xFF;
	}
	if (yscroll >= 0x100)
	{
		yscroll = 0xFF;
	}
	else if (yscroll <= -0x100)
	{
		yscroll = -0xFF;
	}
	RF_Scroll(xscroll, yscroll);

	originxtilemax = (originxtile + PORTTILESWIDE) - 1;
	originytilemax = (originytile + PORTTILESHIGH) - 1;
	inactivateleft = originxtile - INACTIVATEDIST;
	if (inactivateleft < 0)
	{
		inactivateleft = 0;
	}
	inactivateright = originxtilemax + INACTIVATEDIST;
	if (inactivateright < 0)
	{
		inactivateright = 0;
	}
	inactivatetop = originytile - INACTIVATEDIST;
	if (inactivatetop < 0)
	{
		inactivatetop = 0;
	}
	inactivatebottom = originytilemax + INACTIVATEDIST;
	if (inactivatebottom < 0)
	{
		inactivatebottom = 0;
	}
}

void ScrollScreen(objtype *ob)
{
	Sint16 xscroll, yscroll, pix, speed;
	Uint16 bottom;

	if (noscrolling)
		return;

	if (ob->left < originxmin || ob->right > originxmax + 320*PIXGLOBAL)
	{
		playstate = ex_completed;
		return;
	}
	if (ob->bottom > originymax + 13*TILEGLOBAL)
	{
		ob->y -= ob->bottom - (originymax + 13*TILEGLOBAL);
		SD_PlaySound(SND_PLUMMET);
		godmode = false;
		KillKeen();
		return;
	}
	xscroll=yscroll=0;
	if (ob->x < originxglobal + 9*TILEGLOBAL)
	{
		xscroll = ob->x - (originxglobal + 9*TILEGLOBAL);
	}
	else if (ob->x > originxglobal + 12*TILEGLOBAL)
	{
		xscroll = ob->x - (originxglobal + 12*TILEGLOBAL);
	}
	if (ob->state == &s_keenlookup2)
	{
		if (camyoff+tics > 167)
		{
			pix = 167-camyoff;
		}
		else
		{
			pix = tics;
		}
		camyoff += pix;
		yscroll = CONVERT_PIXEL_TO_GLOBAL(-pix);
	}
	else if (ob->state == &s_keenduck3)
	{
		if (camyoff-tics < 33)
		{
			pix = camyoff-33;
		}
		else
		{
			pix = tics;
		}
		camyoff -= pix;
		yscroll = CONVERT_PIXEL_TO_GLOBAL(pix);
	}
#ifdef KEEN6
	if (shaketics)
	{
		static Sint16 shaketable[] = {0,
			 -64,  -64,  -64,  64,  64,  64,
			-200, -200, -200, 200, 200, 200,
			-250, -250, -250, 250, 250, 250,
			-250, -250, -250, 250, 250, 250
		};
		yscroll = yscroll + (bottom - (ob->bottom + shaketable[shaketics]));	// BUG: 'bottom' has not been initialized yet!
	}
	else
#endif
	if ( (ob->hitnorth || !ob->needtoclip || ob->state == &s_keengrabedge1))
	{
		if (  ob->state != &s_keenclimbedge1
			&& ob->state != &s_keenclimbedge2
			&& ob->state != &s_keenclimbedge3
			&& ob->state != &s_keenclimbedge4)
		{
			yscroll += ob->ymove;
			bottom = originyglobal + yscroll + CONVERT_PIXEL_TO_GLOBAL(camyoff);
			if (bottom != ob->bottom)
			{
				if (ob->bottom < bottom)
				{
					pix = bottom - ob->bottom;
				}
				else
				{
					pix = ob->bottom - bottom;
				}
				speed = CONVERT_PIXEL_TO_GLOBAL(pix) >> 7;
				if (speed > 0x30)
				{
					speed = 0x30;
				}
				speed *= tics;
				if (speed < 0x10)
				{
					if (pix < 0x10)
					{
						speed = pix;
					}
					else
					{
						speed = 0x10;
					}
				}
				if (ob->bottom < bottom)
				{
					yscroll -= speed;
				}
				else
				{
					yscroll += speed;
				}
			}
		}
	}
	else
	{
		camyoff = 140;
	}
	pix = (ob->bottom-32*PIXGLOBAL)-(originyglobal+yscroll);
	if (pix < 0)
	{
		yscroll += pix;
	}
	pix = (ob->bottom+32*PIXGLOBAL)-(originyglobal+yscroll+200*PIXGLOBAL);
	if (pix > 0)
	{
		yscroll += pix;
	}
	if (xscroll == 0 && yscroll == 0)
		return;

	if (xscroll >= 0x100)
	{
		xscroll = 0xFF;
	}
	else if (xscroll <= -0x100)
	{
		xscroll = -0xFF;
	}
	if (yscroll >= 0x100)
	{
		yscroll = 0xFF;
	}
	else if (yscroll <= -0x100)
	{
		yscroll = -0xFF;
	}
	RF_Scroll(xscroll, yscroll);

	originxtilemax = (originxtile + PORTTILESWIDE) - 1;
	originytilemax = (originytile + PORTTILESHIGH) - 1;
	inactivateleft = originxtile - INACTIVATEDIST;
	if (inactivateleft < 0)
	{
		inactivateleft = 0;
	}
	inactivateright = originxtilemax + INACTIVATEDIST;
	if (inactivateright < 0)
	{
		inactivateright = 0;
	}
	inactivatetop = originytile - INACTIVATEDIST;
	if (inactivatetop < 0)
	{
		inactivatetop = 0;
	}
	inactivatebottom = originytilemax + INACTIVATEDIST;
	if (inactivatebottom < 0)
	{
		inactivatebottom = 0;
	}
}

void InitObjList(void)
{
	Sint16 i;

	for (i=0; i<MAXACTORS; i++)
	{
		objarray[i].prev = &objarray[i+1];
		objarray[i].next = NULL;
	}
	objarray[MAXACTORS-1].prev = NULL;
	objfreelist = &objarray[0];
	lastobj = NULL;
	objectcount = 0;
	GetNewObj(false);
	player = new;
	GetNewObj(false);
	scoreobj = new;
}

Sint16 GetNewObj(boolean usedummy)
{
	if (!objfreelist)
	{
		if (usedummy)
		{
			new = &dummyobj;
			return -1;
		}
		Quit("GetNewObj: No free spots in objarray!");
	}
	new = objfreelist;
	objfreelist = new->prev;
	memset(new, 0, sizeof(*new));
	if (lastobj)
	{
		lastobj->next = new;
	}
	new->prev = lastobj;
	new->active = yes;
	new->needtoclip = true;
	lastobj = new;
	objectcount++;
	return 0;
}

void RemoveObj(objtype *ob)
{
	if (ob == player)
	{
		Quit("RemoveObj: Tried to remove the player!");
	}
	RF_RemoveSprite(&ob->sprite);
	if (ob->obclass == stunnedobj)
	{
		RF_RemoveSprite((void **)&ob->temp3);
	}

	if (ob == lastobj)
	{
		lastobj = ob->prev;
	}
	else
	{
		ob->next->prev = ob->prev;
	}
	ob->prev->next = ob->next;
	ob->prev = objfreelist;
	objfreelist = ob;
}

void GivePoints(Uint16 points)
{
	gamestate.score += points;
	if (!DemoMode && gamestate.score >= gamestate.nextextra)
	{
		SD_PlaySound(SND_EXTRAKEEN);
		gamestate.lives++;
		gamestate.nextextra *= 2;
	}
}

void GetPlayerInput(void)
{
	IN_ReadControl(0, &c);
	if (c.yaxis != -1)
		upKeyHeld = false;

	if (gravisOn && !DemoMode)
	{
		button0down = gravisJumpDown;
		button1down = gravisPogoDown;
		button2down = gravisFireDown;
		if (!button0down)
			button0held = false;
		if (!button1down)
			button1held = false;
		if (!button2down)
			button2held = false;
	}
	else if (oldshooting || DemoMode)
	{
		if (c.button0 && c.button1)
		{
			button2down = true;
			button0down = button1down = button0held = button1held = false;
		}
		else
		{
			button2down = button2held = false;
			if (c.button0)
			{
				button0down = true;
			}
			else
			{
				button0down = button0held = false;
			}
			if (c.button1)
			{
				if (pogotics < 8)
				{
					pogotics = pogotics + tics;
				}
				else
				{
					button1down = true;
				}
			}
			else
			{
				if (pogotics != 0)
				{
					button1down = true;
				}
				else
				{
					button1down = button1held = false;
				}
				pogotics = 0;
			}
		}
	}
	else
	{
		button0down = c.button0;
		button1down = c.button1;
		button2down = Keyboard[firescan];
		if (!button0down)
			button0held = false;
		if (!button1down)
			button1held = false;
		if (!button2down)
			button2held = false;
	}
}

void StopMusic(void)
{
	Sint16 i;

	SD_MusicOff();
	for (i=0; i<LASTMUSIC; i++)
	{
		if (audiosegs[STARTMUSIC+i])
		{
#ifdef FIX_MUSIC_MEMORY_ISSUES
			//unlock any music blocks so that they can be purged
			MM_SetLock(&(memptr)audiosegs[STARTMUSIC+i], false);
#endif
			MM_SetPurge(&(memptr)audiosegs[STARTMUSIC+i], PURGE_FIRST);
		}
	}
}

void StartMusic(Uint16 num)
{
	Sint16 song;
	boolean wasfaded;

	if (num >= ARRAYLENGTH(songs) && num != 0xFFFF)
	{
		Quit("StartMusic() - bad level number");
	}

#ifdef FIX_MUSIC_MEMORY_ISSUES
	StopMusic();
#else
	SD_MusicOff();
#endif

#ifdef KEEN4
	if (num == 0xFFFF)
	{
		song = WONDER_MUS;
	}
	else
	{
		song = songs[num];
	}
#else
	song = songs[num];
#endif

	if (song == -1 || MusicMode != smm_AdLib)
	{
		return;
	}

	MM_BombOnError(false);
	CA_CacheAudioChunk(STARTMUSIC+song);
	MM_BombOnError(true);
	if (mmerror)
	{
		mmerror = false;
		if (!DemoMode)
		{
			US_CenterWindow(20, 8);
			PrintY += 20;
			US_CPrint("Insufficient memory\nfor background music!");
			VW_UpdateScreen();
			wasfaded = screenfaded;
			if (wasfaded)
			{
				VW_SetDefaultColors();
			}
			IN_ClearKeysDown();
			IN_UserInput(3*TickBase, false);
			if (wasfaded)
			{
				VW_FadeOut();
			}
		}
	}
	else
	{
#ifdef FIX_MUSIC_MEMORY_ISSUES
		//The current music should be locked, so the memory manager will not
		//mess with it when compressing memory blocks in MM_SortMem().
		MM_SetLock(&(memptr)audiosegs[STARTMUSIC+song], true);
#endif
		SD_StartMusic((MusicGroup far *)audiosegs[STARTMUSIC+song]);
	}
}

void PlayLoop(void)
{
	objtype *check;

	StartMusic(gamestate.mapon);
	button0held = upKeyHeld = button1held = button2held = false;
	ingame = true;
	playstate = ex_stillplaying;
	invulnerable = noscrolling = pogotics = 0;
	CenterPlayer(player);
	if (DemoMode)
	{
		US_InitRndT(false);
	}
	else
	{
		US_InitRndT(true);
	}
	TimeCount = lasttimecount = tics = 3;
	do
	{
		GetPlayerInput();
		for (obj=player; obj; obj=obj->next)
		{
			// think
			if (!obj->active && obj->tileright >= originxtile-1
				&& obj->tileleft <= originxtilemax+1 && obj->tiletop <= originytilemax+1
				&& obj->tilebottom >= originytile-1)
			{
				obj->needtoreact = true;
				obj->active = yes;
			}
			if (!obj->active)
			{
				continue;
			}
			if (obj->tileright < inactivateleft
				|| obj->tileleft > inactivateright
				|| obj->tiletop > inactivatebottom
				|| obj->tilebottom < inactivatetop)
			{
				if (obj->active == removable)
				{
					RemoveObj(obj);
					continue;
				}
				else if (obj->active != allways)
				{
					if (US_RndT() < tics*2 || screenfaded || loadedgame)
					{
						RF_RemoveSprite(&obj->sprite);
						if (obj->obclass == stunnedobj)
							RF_RemoveSprite((void **)&obj->temp3);
						obj->active = no;
						continue;
					}
				}
			}
			StateMachine(obj);
		}
		if (gamestate.platform)
		{
			FollowPlatform(player);
		}
		for (obj=player; obj; obj=obj->next)
		{
			// check collisions
			if (!obj->active)
			{
				continue;
			}
			for (check=obj->next; check; check=check->next)
			{
				if (!check->active)
				{
					continue;
				}
				if (obj->right > check->left && obj->left < check->right
					&& obj->top < check->bottom && obj->bottom > check->top)
				{
					if (obj->state->contact)
					{
						obj->state->contact(obj, check);
					}
					if (check->state->contact)
					{
						check->state->contact(check, obj);
					}
					if (obj->obclass == nothing)	//useless -- obclass is NOT set to nothing by RemoveObj
					{
						break;
					}
				}
			}
		}
		if (mapon != 0)
		{
			CheckTiles(player);
		}
		else
		{
			CheckWorldTiles(player);
		}
		for (obj=player; obj; obj=obj->next)
		{
			// react
			if (!obj->active)
			{
				continue;
			}
			if (obj->tilebottom >= mapheight-1)
			{
				if (obj->obclass == keenobj)
				{
					playstate = ex_died;
				}
				else
				{
					RemoveObj(obj);
				}
				continue;
			}
			if (obj->needtoreact && obj->state->react)
			{
				obj->needtoreact = false;
				obj->state->react(obj);
			}
		}
#ifdef KEEN4
		if (mapon != 0 && mapon != 17)
#else
		if (mapon != 0)
#endif
		{
			ScrollScreen(player);
		}
		else
		{
			ScrollScreenWorld(player);
		}
		UpdateScorebox(scoreobj);
		if (loadedgame)
		{
			loadedgame = false;
		}
		RF_Refresh();
		if (invulnerable)
		{
			invulnerable = invulnerable - tics;
			if (invulnerable < 0)
				invulnerable = 0;
		}
#ifdef KEEN6
		if (shaketics)
		{
			shaketics = shaketics - tics;
			if (shaketics < 0)
				shaketics = 0;
		}
#endif
		if (singlestep)
		{
			VW_WaitVBL(14);	//reduces framerate to 5 fps on VGA or 4.3 fps on EGA cards
			lasttimecount = TimeCount;
		}
		if (extravbls)
		{
			VW_WaitVBL(extravbls);
		}
		if (DemoMode == demo_Playback)
		{
			if (!screenfaded && IN_IsUserInput())
			{
				playstate = ex_completed;
				if (LastScan == sc_F1)
				{
					LastScan = sc_Space;
				}
			}
		}
		else if (DemoMode == demo_PlayDone)
		{
			playstate = ex_completed;
		}
		else
		{
			CheckKeys();
		}
		if (Keyboard[sc_E] && Keyboard[sc_N] && Keyboard[sc_D])
		{
#if defined KEEN4
			gamestate.rescued = 7;
			playstate = ex_rescued;
#elif defined KEEN5
			playstate = ex_brokeQED;
#elif defined KEEN6
			playstate = ex_gotmolly;
#endif
		}
	} while (playstate == ex_stillplaying);
	ingame = false;
	StopMusic();
}