/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "CK_DEF.H"

/////////////////////////////////////////////////////////////////////////////
// initialized variables:
/////////////////////////////////////////////////////////////////////////////

Sint16 wallclip[8][16] = {			// the height of a given point in a tile
{ 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256, 256},
{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0},
{   0,0x08,0x10,0x18,0x20,0x28,0x30,0x38,0x40,0x48,0x50,0x58,0x60,0x68,0x70,0x78},
{0x80,0x88,0x90,0x98,0xa0,0xa8,0xb0,0xb8,0xc0,0xc8,0xd0,0xd8,0xe0,0xe8,0xf0,0xf8},
{   0,0x10,0x20,0x30,0x40,0x50,0x60,0x70,0x80,0x90,0xa0,0xb0,0xc0,0xd0,0xe0,0xf0},
{0x78,0x70,0x68,0x60,0x58,0x50,0x48,0x40,0x38,0x30,0x28,0x20,0x18,0x10,0x08,   0},
{0xf8,0xf0,0xe8,0xe0,0xd8,0xd0,0xc8,0xc0,0xb8,0xb0,0xa8,0xa0,0x98,0x90,0x88,0x80},
{0xf0,0xe0,0xd0,0xc0,0xb0,0xa0,0x90,0x80,0x70,0x60,0x50,0x40,0x30,0x20,0x10,   0}
};

/////////////////////////////////////////////////////////////////////////////
// uninitialized variables:
/////////////////////////////////////////////////////////////////////////////

static Sint16 topmoved;
Sint16 tmp_xmove;
Sint16 tmp_ymove;
boolean pushPlayer;
static Sint16 bottommoved;
static Uint16 oldmidx;
static Uint16 oldleft;
static Uint16 oldtop;
static Uint16 oldright;
static Uint16 oldbottom;
static Uint16 oldtilemidx;
static Uint16 oldtileleft;
static Uint16 oldtiletop;
static Uint16 oldtileright;
static Sint16 leftmoved;
static Sint16 midxmoved;
static Sint16 rightmoved;
static Uint16 oldtilebottom;

/////////////////////////////////////////////////////////////////////////////
// functions:
/////////////////////////////////////////////////////////////////////////////

void MoveObjVert(objtype *ob, Sint16 ymove)
{
	ob->y += ymove;
	ob->top += ymove;
	ob->bottom += ymove;
	ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
	ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
}

void MoveObjHoriz(objtype *ob, Sint16 xmove)
{
	//BUG? ob->midx is not adjusted
	ob->x += xmove;
	ob->left += xmove;
	ob->right += xmove;
#ifdef KEEN6
	ob->midx += xmove;	//BUG? ob->tilemidx is not updated
#endif
	ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
	ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
}

void ClipToNorthWall(objtype *ob)
{
	Uint16 far *map;
	Uint16 wall, clip, xpix;
	Sint16 xmove, ymove;

	map = mapsegs[1] + mapbwidthtable[ob->tilebottom-1]/2;
	if (ob->xdir == 1)
	{
		xpix = 0;
		map += ob->tileright;
		xmove = ob->right - ob->midx;
		if (tinf[map[-mapwidth]+WESTWALL] || tinf[map[0]+WESTWALL])
		{
			return;
		}
	}
	else
	{
		xpix = 15;
		map += ob->tileleft;
		xmove = ob->left - ob->midx;
		if (tinf[map[-mapwidth]+EASTWALL] || tinf[map[0]+EASTWALL])
		{
			return;
		}
	}
	if (tinf[map[0]+NORTHWALL])
	{
		return;
	}
	map += mapwidth;
	if ((wall = tinf[map[0]+NORTHWALL]) != 1)
	{
		return;
	}
	clip = wallclip[wall&7][xpix];
	ymove = CONVERT_TILE_TO_GLOBAL(ob->tilebottom) + clip - 1 -ob->bottom;
	if (ymove <= 0 && ymove >= -bottommoved)
	{
		ob->hitnorth = wall;
		MoveObjVert(ob, ymove);
		MoveObjHoriz(ob, xmove);
	}
}

void ClipToSouthWall(objtype *ob)
{
	Uint16 far *map;
	Uint16 xpix, wall, clip;
	Sint16 move;

	map = mapsegs[1] + mapbwidthtable[ob->tiletop+1]/2;
	if (ob->xdir == 1)
	{
		xpix = 0;
		map += ob->tileright;
		if (tinf[map[mapwidth]+WESTWALL] || tinf[map[2*mapwidth]+WESTWALL])
		{
			return;
		}
	}
	else
	{
		xpix = 15;
		map += ob->tileleft;
		if (tinf[map[mapwidth]+EASTWALL] || tinf[map[2*mapwidth]+EASTWALL])
		{
			return;
		}
	}
	if (tinf[map[0]+SOUTHWALL])
	{
		return;
	}
	map -= mapwidth;
	if ((wall = tinf[map[0]+SOUTHWALL]) != 0)
	{
		clip = wallclip[wall&7][xpix];
		move = CONVERT_TILE_TO_GLOBAL(ob->tiletop+1) - clip - ob->top;
		if (move >= 0 && move <= -topmoved)
		{
			ob->hitsouth = wall;
			MoveObjVert(ob, move);
		}
	}
}

void ClipToEnds(objtype *ob)
{
	Uint16 far *map;
	Uint16 wall, y, clip;
	Sint16 totalmove, maxmove, move;
	Uint16 midxpix;
	
	midxpix = CONVERT_GLOBAL_TO_PIXEL(ob->midx & 0xF0);
	maxmove = -abs(midxmoved)-bottommoved-16;
	map = mapsegs[1] + mapbwidthtable[oldtilebottom-1]/2 + ob->tilemidx;
	for (y=oldtilebottom-1; y <= ob->tilebottom; y++,map+=mapwidth)
	{
		wall = tinf[*map + NORTHWALL];
		if (wall)
		{
			clip = wallclip[wall&7][midxpix];
			move = (CONVERT_TILE_TO_GLOBAL(y) + clip)-1-ob->bottom;
			if (move < 0 && move >= maxmove)
			{
				ob->hitnorth = wall;
				MoveObjVert(ob, move);
				return;
			}
		}
	}
	maxmove = abs(midxmoved)-topmoved+16;
	map = mapsegs[1] + mapbwidthtable[oldtiletop+1]/2 + ob->tilemidx;
#if 0
	// BUG: the loop uses unsigned comparison and will never end when tiletop is 0!
	for (y=oldtiletop+1; y >= ob->tiletop; y--,map-=mapwidth)
#else
	// bugfixed version (forced into signed comparison):
	for (y=oldtiletop+1; (Sint16)y >= (Sint16)ob->tiletop; y--,map-=mapwidth)
#endif
	{
		wall = tinf[*map + SOUTHWALL];
		if (wall)
		{
			clip = wallclip[wall&7][midxpix];
			move = CONVERT_TILE_TO_GLOBAL(y+1) - clip - ob->top;
			if (move > 0 && move <= maxmove)
			{
				totalmove = tmp_ymove+move;
				if (totalmove < 0x100 && totalmove > -0x100)
				{
					ob->hitsouth = wall;
					MoveObjVert(ob, move);
					//BUG? no return here
				}
			}
		}
	}
}

void ClipToEastWestWalls(objtype *ob)
{
	Sint16 move, y, top, bottom;
	Uint16 far *map;
	
	top = ob->tiletop;
	if (ob->hitsouth > 1)
	{
		top++;
	}
	bottom = ob->tilebottom;
	if (ob->hitnorth > 1)
	{
		bottom--;
	}
	for (y=top; y<=bottom; y++)
	{
		map = mapsegs[1] + mapbwidthtable[y]/2 + ob->tileleft;
		if ((ob->hiteast = tinf[*map+EASTWALL]) != 0)
		{
			move = CONVERT_TILE_TO_GLOBAL(ob->tileleft+1) - ob->left;
			MoveObjHoriz(ob, move);
			return;
		}
	}
	for (y=top; y<=bottom; y++)
	{
		map = mapsegs[1] + mapbwidthtable[y]/2 + ob->tileright;
		if ((ob->hitwest = tinf[*map+WESTWALL]) != 0)
		{
			move = (CONVERT_TILE_TO_GLOBAL(ob->tileright)-1)-ob->right;
			MoveObjHoriz(ob, move);
			return;
		}
	}
}

boolean PositionOkay(objtype *ob)
{
	Uint16 far *map;
	Uint16 tile, x, y, rowdiff;
	
	map = mapsegs[1] + mapbwidthtable[ob->tiletop]/2 + ob->tileleft;
	rowdiff = mapwidth-(ob->tileright-ob->tileleft+1);
	for (y=ob->tiletop; y<=ob->tilebottom; y++,map+=rowdiff)
	{
		for (x=ob->tileleft; x<=ob->tileright; x++)
		{
			tile = *(map++);
			if (tinf[tile+NORTHWALL] || tinf[tile+EASTWALL] || tinf[tile+SOUTHWALL] || tinf[tile+WESTWALL])
			{
				return false;
			}
		}
	}
	return true;
}

boolean StateOkay(objtype *ob, statetype *state)
{
	spritetabletype far *shape;

	if (ob->xdir > 0)
	{
		ob->shapenum = state->rightshapenum;
	}
	else
	{
		ob->shapenum = state->leftshapenum;
	}
	shape = &spritetable[ob->shapenum-STARTSPRITES];
	ob->left = ob->x + shape->xl;
	ob->right = ob->x + shape->xh;
	ob->top = ob->y + shape->yl;
	ob->bottom = ob->y + shape->yh;
	ob->midx = ob->left + (ob->right-ob->left)/2;
	ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
	ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
	ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
	ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
	ob->tilemidx = CONVERT_GLOBAL_TO_TILE(ob->midx);
	return PositionOkay(ob);
}

#ifdef KEEN5
void CalcBounds(objtype *ob)	//not present in Keen 4 & 6
{
	spritetabletype far *shape;

	shape = &spritetable[ob->shapenum-STARTSPRITES];
	ob->left = ob->x + shape->xl;
	ob->right = ob->x + shape->xh;
	ob->top = ob->y + shape->yl;
	ob->bottom = ob->y + shape->yh;
	ob->midx = ob->left + (ob->right-ob->left)/2;
}
#endif

void ClipToWalls(objtype *ob)
{
	Uint16 oldx, oldy;
	boolean pushed;
	spritetabletype far *shape;

	oldx = ob->x;
	oldy = ob->y;
	pushed = false;
	if (ob->state->pushtofloor)
	{
		if (ob->hitnorth == 25)
		{
			tmp_ymove = 145;
		}
		else
		{
			if (tmp_xmove > 0)
			{
				tmp_ymove = tmp_xmove+16;
			}
			else
			{
				tmp_ymove = -tmp_xmove+16;
			}
			pushed = true;
		}
	}
	if (tmp_xmove > 239)
	{
		tmp_xmove = 239;
	}
	else if (tmp_xmove < -239)
	{
		tmp_xmove = -239;
	}
	if (tmp_ymove > 255)
	{
		tmp_ymove = 255;
	}
	else if (tmp_ymove < -239)
	{
		tmp_ymove = -239;
	}
	ob->x += tmp_xmove;
	ob->y += tmp_ymove;
	ob->needtoreact = true;
	if (!ob->shapenum)
	{
		return;
	}
	shape = &spritetable[ob->shapenum-STARTSPRITES];
	oldtileright = ob->tileright;
	oldtiletop = ob->tiletop;
	oldtileleft = ob->tileleft;
	oldtilebottom = ob->tilebottom;
	oldtilemidx = ob->tilemidx;
	oldright = ob->right;
	oldtop = ob->top;
	oldleft = ob->left;
	oldbottom = ob->bottom;
	oldmidx = ob->midx;
	ob->left = ob->x + shape->xl;
	ob->right = ob->x + shape->xh;
	ob->top = ob->y + shape->yl;
	ob->bottom = ob->y + shape->yh;
	ob->midx = ob->left + (ob->right-ob->left)/2;
	ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
	ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
	ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
	ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
	ob->tilemidx = CONVERT_GLOBAL_TO_TILE(ob->midx);
	ob->hitnorth=ob->hiteast=ob->hitsouth=ob->hitwest=0;
	if (ob->needtoclip)
	{
		leftmoved = ob->left - oldleft;
		rightmoved = ob->right - oldright;
		topmoved = ob->top - oldtop;
		bottommoved = ob->bottom - oldbottom;
		midxmoved = ob->midx - oldmidx;
		ClipToEnds(ob);
		if (ob == player && !pushPlayer)	// zero tolerance near the edge when player gets pushed!
		{
			if (!ob->hitnorth && bottommoved > 0)
			{
				ClipToNorthWall(ob);
			}
			if (!ob->hitsouth && topmoved < 0)
			{
				ClipToSouthWall(ob);
			}
		}
		ClipToEastWestWalls(ob);

#ifdef KEEN6
		//
		// special hack to prevent player from getting stuck on slopes?
		//
		if (ob == player && (ob->hitnorth & 7) > 1 && (ob->hiteast || ob->hitwest))
		{
			Uint16 far *map;
			Uint16 pixx, y, wall, clip, move;

			pixx = CONVERT_GLOBAL_TO_PIXEL(ob->midx & (15*PIXGLOBAL));
			map = mapsegs[1] + mapbwidthtable[oldtilebottom]/2 + ob->tilemidx;

			for (y=oldtilebottom; ob->tilebottom+1 >= y; y++, map+=mapwidth)
			{
				wall = tinf[*map + NORTHWALL];
				if (wall)
				{
					clip = wallclip[wall & 7][pixx];
					move = CONVERT_TILE_TO_GLOBAL(y) + clip - 1 - ob->bottom;
					ob->hitnorth = wall;
					MoveObjVert(ob, move);
					return;
				}
			}
		}
#endif
	}
	if (pushed && !ob->hitnorth)
	{
		ob->y = oldy;
		ob->x = oldx + tmp_xmove;
		ob->left = ob->x + shape->xl;
		ob->right = ob->x + shape->xh;
		ob->top = ob->y + shape->yl;
		ob->bottom = ob->y + shape->yh;
		ob->midx = ob->left + (ob->right-ob->left)/2;
		ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
		ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
		ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
		ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
		ob->tilemidx = CONVERT_GLOBAL_TO_TILE(ob->midx);
	}
	ob->xmove = ob->xmove + (ob->x - oldx);
	ob->ymove = ob->ymove + (ob->y - oldy);
}

void FullClipToWalls(objtype *ob)
{
	Uint16 oldx, oldy, w, h;
	spritetabletype far *shape;

	oldx = ob->x;
	oldy = ob->y;
	if (tmp_xmove > 239)
	{
		tmp_xmove = 239;
	}
	else if (tmp_xmove < -239)
	{
		tmp_xmove = -239;
	}
	if (tmp_ymove > 239)
	{
		tmp_ymove = 239;
	}
	else if (tmp_ymove < -239)
	{
		tmp_ymove = -239;
	}
	ob->x += tmp_xmove;
	ob->y += tmp_ymove;
	ob->needtoreact = true;
	shape = &spritetable[ob->shapenum-STARTSPRITES];
	switch (ob->obclass)
	{
#if defined KEEN4
	case keenobj:
		w = 40*PIXGLOBAL;
		h = 24*PIXGLOBAL;
		break;
	case eagleobj:
		w = 64*PIXGLOBAL;
		h = 32*PIXGLOBAL;
		break;
	case dopefishobj:
		w = 88*PIXGLOBAL;
		h = 64*PIXGLOBAL;
		break;
	case schoolfishobj:
		w = 16*PIXGLOBAL;
		h = 8*PIXGLOBAL;
		break;
#elif defined KEEN5
	case slicestarobj:
	case spherefulobj:
		w = h = 32*PIXGLOBAL;
		break;
#elif defined KEEN6
	case blorbobj:
		w = h = 32*PIXGLOBAL;
		break;
#endif

	default:
		Quit("FullClipToWalls: Bad obclass");
		break;
	}
	ob->right = ob->x + w;
	ob->left = ob->x;
	ob->top = ob->y;
	ob->bottom = ob->y + h;
	ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
	ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
	ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
	ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
	ob->hitnorth=ob->hiteast=ob->hitsouth=ob->hitwest=0;
	if (!PositionOkay(ob))
	{
		MoveObjHoriz(ob, -tmp_xmove);	//undo x movement
		if (PositionOkay(ob))
		{
			if (tmp_xmove > 0)
			{
				ob->hitwest = 1;
			}
			else
			{
				ob->hiteast = 1;
			}
		}
		else
		{
			if (tmp_ymove > 0)
			{
				ob->hitnorth = 1;
			}
			else
			{
				ob->hitsouth = 1;
			}
			MoveObjHoriz(ob, tmp_xmove);	//redo x movement
			MoveObjVert(ob, -tmp_ymove);	//undo y movement
			if (!PositionOkay(ob))
			{
				MoveObjHoriz(ob, -tmp_xmove);	//undo x movement
				if (tmp_xmove > 0)
				{
					ob->hitwest = 1;
				}
				else
				{
					ob->hiteast = 1;
				}
			}
		}
	}
	ob->xmove = ob->xmove + (ob->x - oldx);
	ob->ymove = ob->ymove + (ob->y - oldy);
	ob->left = ob->x + shape->xl;
	ob->right = ob->x + shape->xh;
	ob->top = ob->y + shape->yl;
	ob->bottom = ob->y + shape->yh;
	ob->midx = ob->left + (ob->right-ob->left)/2;
}

void PushClipToWalls(objtype *ob)
{
	Uint16 oldx, oldy;
	spritetabletype far *shape;
	
	oldx = ob->x;
	oldy = ob->y;
	ob->x += tmp_xmove;
	ob->y += tmp_ymove;
	ob->needtoreact = true;
	if (!ob->shapenum)
	{
		return;
	}
	shape = &spritetable[ob->shapenum-STARTSPRITES];
	oldtileright = ob->tileright;
	oldtiletop = ob->tiletop;
	oldtileleft = ob->tileleft;
	oldtilebottom = ob->tilebottom;
	oldtilemidx = ob->tilemidx;
	oldright = ob->right;
	oldtop = ob->top;
	oldleft = ob->left;
	oldbottom = ob->bottom;
	oldmidx = ob->midx;
	ob->left = ob->x + shape->xl;
	ob->right = ob->x + shape->xh;
	ob->top = ob->y + shape->yl;
	ob->bottom = ob->y + shape->yh;
	ob->midx = ob->left + (ob->right-ob->left)/2;
	ob->tileleft = CONVERT_GLOBAL_TO_TILE(ob->left);
	ob->tileright = CONVERT_GLOBAL_TO_TILE(ob->right);
	ob->tiletop = CONVERT_GLOBAL_TO_TILE(ob->top);
	ob->tilebottom = CONVERT_GLOBAL_TO_TILE(ob->bottom);
	ob->tilemidx = CONVERT_GLOBAL_TO_TILE(ob->midx);
	if (ob->needtoclip)
	{
		leftmoved = ob->left - oldleft;
		rightmoved = ob->right - oldright;
		topmoved = ob->top - oldtop;
		bottommoved = ob->bottom - oldbottom;
		midxmoved = ob->midx - oldmidx;
		ClipToEnds(ob);
		ClipToEastWestWalls(ob);
	}
	ob->xmove = ob->xmove + (ob->x - oldx);
	ob->ymove = ob->ymove + (ob->y - oldy);
}

void ClipToSpriteSide(objtype *push, objtype *solid)
{
	Sint16 xmove, leftinto, rightinto;

	xmove = solid->xmove - push->xmove;
	leftinto = solid->right - push->left;
	rightinto = push->right - solid->left;
	if (leftinto > 0 && leftinto <= xmove)
	{
		tmp_xmove = leftinto;
		if (push->state->pushtofloor)
		{
			tmp_ymove = leftinto+16;
		}
		ClipToWalls(push);
		push->hiteast = 1;
	}
	else if (rightinto > 0 && rightinto <= -xmove)
	{
		tmp_xmove = -rightinto;
		if (push->state->pushtofloor)
		{
			tmp_ymove = rightinto+16;
		}
		ClipToWalls(push);
		push->hitwest = 1;
	}
}

void ClipToSpriteTop(objtype *push, objtype *solid)
{
	Sint16 temp, ymove, bottominto;

	ymove = push->ymove - solid->ymove;
	bottominto = push->bottom - solid->top;
	if (bottominto >= 0 && bottominto <= ymove)
	{
		if (push == player)
		{
			gamestate.platform = solid;
		}
		tmp_ymove = -bottominto;
		temp = push->state->pushtofloor;
		push->state->pushtofloor = push_none;
		ClipToWalls(push);
		push->state->pushtofloor = temp;
		if (!push->hitsouth)
		{
			push->hitnorth = 25;
		}
	}
}

void ClipToSprite(objtype *push, objtype *solid, boolean squish)
{
	Sint16 xmove, ymove, leftinto, rightinto, topinto, bottominto;
	
	xmove = solid->xmove - push->xmove;
	tmp_xmove = tmp_ymove = 0;
	leftinto = solid->right - push->left;
	rightinto = push->right - solid->left;
	if (leftinto > 0 && xmove+1 >= leftinto)
	{
		tmp_xmove = leftinto;
		push->xspeed = 0;
		PushClipToWalls(push);
		if (squish && push->hitwest)
		{
			KillKeen();
		}
		push->hiteast = 1;
		return;
	}
	else if (rightinto > 0 && -xmove+1 >= rightinto)
	{
		tmp_xmove = -rightinto;
		push->xspeed = 0;
		PushClipToWalls(push);
		if (squish && push->hiteast)
		{
			KillKeen();
		}
		push->hitwest = 1;
		return;
	}
	ymove = push->ymove - solid->ymove;
	topinto = solid->bottom - push->top;
	bottominto = push->bottom - solid->top;
	if (bottominto >= 0 && bottominto <= ymove)
	{
		if (push == player)
		{
			gamestate.platform = solid;
		}
		tmp_ymove = -bottominto;
		PushClipToWalls(push);
		if (squish && push->hitsouth)
		{
			KillKeen();
		}
		if (!push->hitsouth)
		{
			push->hitnorth = 25;
		}
		return;
	}
	else if (topinto >= 0 && topinto <= ymove)
	{
		tmp_ymove = topinto;
		PushClipToWalls(push);
		if (squish && push->hitnorth)
		{
			KillKeen();
		}
		push->hitsouth = 25;
	}
}

/*
void ClipPushSprite(objtype *push, objtype *solid, boolean squish)	//not present in Keen 4-6
{
	Sint16 xmove, ymove, leftinto, rightinto, topinto, bottominto, var_E;
	
	if ((solid->hiteast && push->xdir == -1) || (solid->hitwest && push->xdir == 1))
	{
		ClipToSprite(push, solid, squish);
		return;
	}
	xmove = solid->xmove - push->xmove;
	tmp_xmove = tmp_ymove = 0;
	ymove = push->ymove - solid->ymove;
	topinto = solid->bottom - push->top;
	bottominto = push->bottom - solid->top;
	if (bottominto >= 0 && bottominto <= ymove)
	{
		if (push == player)
		{
			gamestate.platform = solid;
		}
		tmp_ymove = -bottominto;
		PushClipToWalls(push);
		if (squish && push->hitsouth)
		{
			KillPlayer();
		}
		if (!push->hitsouth)
		{
			push->hitnorth = 23;
		}
		return;
	}
	else if (topinto >= 0 && -ymove >= topinto)
	{
		tmp_ymove = topinto;
		ClipToWalls(push);
		if (squish && push->hitnorth)
		{
			KillPlayer();
		}
		push->hitsouth = 23;
		return;
	}
	leftinto = solid->right - push->left;
	rightinto = push->right - solid->left;
	if (leftinto > 0 && push->left > solid->left)
	{
		var_E = (leftinto-xmove)+1;
		if (var_E < 0)
		{
			var_E = 0;
		}
		push->xspeed = 0;
		tmp_xmove = leftinto+var_E;
		PushClipToWalls(push);
		tmp_xmove = -leftinto/2;
		PushClipToWalls(solid);
		if (squish && push->hitwest)
		{
			KillPlayer();
		}
		return;
	}
	else if (rightinto > 0 && push->right < solid->right)
	{
		var_E = rightinto+xmove+1;
		if (var_E < 0)
		{
			var_E = 0;
		}
		push->xspeed = 0;
		tmp_xmove = -rightinto - var_E;
		PushClipToWalls(push);
		tmp_xmove = rightinto/2;
		PushClipToWalls(solid);
		if (squish && push->hiteast)
		{
			KillPlayer();
		}
		return;
	}
}
*/

Sint16 DoActor(objtype *ob, Sint16 numtics)
{
	Sint16 ticcount, usedtics, excesstics;
	statetype *state;
	
	state = ob->state;
	if (state->progress == think)
	{
		if (state->think)
		{
			if (ob->nothink)
			{
				ob->nothink--;
			}
			else
			{
				state->think(ob);
			}
		}
		return 0;
	}
	ticcount = ob->ticcount+numtics;
	if (state->tictime > ticcount || state->tictime == 0)
	{
		ob->ticcount = ticcount;
		if (state->progress == slide || state->progress == slidethink)
		{
			if (ob->xdir)
			{
				tmp_xmove += ob->xdir == 1? state->xmove*numtics : state->xmove*-numtics;
			}
			if (ob->ydir)
			{
				tmp_ymove += ob->ydir == 1? state->ymove*numtics : state->ymove*-numtics;
			}
		}
		if ((state->progress == slidethink || state->progress == stepthink) && state->think)
		{
			if (ob->nothink)
			{
				ob->nothink--;
			}
			else
			{
				state->think(ob);
			}
		}
		return 0;
	}
	else
	{
		usedtics = state->tictime - ob->ticcount;
		excesstics = ticcount - state->tictime;
		ob->ticcount = 0;
		if (state->progress == slide || state->progress == slidethink)
		{
			if (ob->xdir)
			{
				tmp_xmove += ob->xdir == 1? state->xmove*usedtics : state->xmove*-usedtics;
			}
			if (ob->ydir)
			{
				tmp_ymove += ob->ydir == 1? state->ymove*usedtics : state->ymove*-usedtics;
			}
		}
		else
		{
			if (ob->xdir)
			{
				tmp_xmove += ob->xdir == 1? state->xmove : -state->xmove;
			}
			if (ob->ydir)
			{
				tmp_ymove += ob->ydir == 1? state->ymove : -state->ymove;
			}
		}
		if (state->think)
		{
			if (ob->nothink)
			{
				ob->nothink--;
			}
			else
			{
				state->think(ob);
			}
		}
		if (state == ob->state)
		{
			ob->state = state->nextstate;
		}
		else if (!ob->state)
		{
			return 0;
		}
		return excesstics;
	}
}

void StateMachine(objtype *ob)
{
	Sint16 excesstics, oldshapenum;
	statetype *state;
	
	ob->xmove=ob->ymove=tmp_xmove=tmp_ymove=0;
	oldshapenum = ob->shapenum;
	state = ob->state;
	excesstics = DoActor(ob, tics);
	if (ob->state != state)
	{
		ob->ticcount = 0;
		state = ob->state;
	}
	while (excesstics)
	{
		if (!state->skippable && state->tictime <= excesstics)
		{
			excesstics = DoActor(ob, state->tictime-1);
		}
		else
		{
			excesstics = DoActor(ob, excesstics);
		}
		if (ob->state != state)
		{
			ob->ticcount = 0;
			state = ob->state;
		}
	}
	if (!state)
	{
		RemoveObj(ob);
		return;
	}
	if (state->rightshapenum)
	{
		if (ob->xdir > 0)
		{
			ob->shapenum = state->rightshapenum;
		}
		else
		{
			ob->shapenum = state->leftshapenum;
		}
	}
	if ((Sint16)ob->shapenum == -1)
	{
		ob->shapenum = 0;
	}
	if (tmp_xmove == 0 && tmp_ymove == 0 && ob->shapenum == oldshapenum && ob->hitnorth != 25)
	{
		return;
	}
	if (ob->needtoclip == FULLCLIP)
	{
		FullClipToWalls(ob);
	}
	else
	{
		ClipToWalls(ob);
	}
}

void NewState(objtype *ob, statetype *state)
{
	Sint16 temp;
	
	ob->state = state;
	if (state->rightshapenum)
	{
		if (ob->xdir > 0)
		{
			ob->shapenum = state->rightshapenum;
		}
		else
		{
			ob->shapenum = state->leftshapenum;
		}
	}
	if ((Sint16)ob->shapenum == -1)
	{
		ob->shapenum = 0;
	}
	temp = ob->needtoclip;
	ob->needtoclip = false;
	tmp_xmove=tmp_ymove=0;
	ClipToWalls(ob);
	ob->needtoclip = temp;
	if (ob->needtoclip == FULLCLIP)
	{
		FullClipToWalls(ob);
	}
	else if (ob->needtoclip == true)
	{
		ClipToWalls(ob);
	}
}

void ChangeState(objtype *ob, statetype *state)
{
	ob->state = state;
	ob->ticcount = 0;
	if (state->rightshapenum)
	{
		if (ob->xdir > 0)
		{
			ob->shapenum = state->rightshapenum;
		}
		else
		{
			ob->shapenum = state->leftshapenum;
		}
	}
	if ((Sint16)ob->shapenum == -1)
	{
		ob->shapenum = 0;
	}
	ob->needtoreact = true;
	tmp_xmove=tmp_ymove=0;
	if (ob->hitnorth != 25)
	{
		ClipToWalls(ob);
	}
}

boolean ObjectVisible(objtype *ob)
{
	if (ob->tileright < originxtile || ob->tilebottom < originytile
		|| ob->tileleft > originxtilemax || ob->tiletop > originytilemax)
	{
		return false;
	}
	return true;
}

void DoGravity4(objtype *ob)
{
	Sint32 i;
	for (i = lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i&1)
		{
			if (ob->yspeed < 0 && ob->yspeed >= -4)
			{
				tmp_ymove += ob->yspeed;
				ob->yspeed = 0;
				return;
			}
			ob->yspeed += 4;
			if (ob->yspeed > 70)
			{
				ob->yspeed = 70;
			}
		}
		tmp_ymove += ob->yspeed;
	}
}

void DoGravity3(objtype *ob)
{
	Sint32 i;
	for (i = lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i&1)
		{
			if (ob->yspeed < 0 && ob->yspeed >= -3)
			{
				tmp_ymove += ob->yspeed;
				ob->yspeed = 0;
				return;
			}
			ob->yspeed += 3;
			if (ob->yspeed > 70)
			{
				ob->yspeed = 70;
			}
		}
		tmp_ymove += ob->yspeed;
	}
}

void DoGravity1(objtype *ob)
{
	Sint32 i;
	for (i = lasttimecount-tics; i<lasttimecount; i++)
	{
		if (!i & 3)	//BUG: this is equal to ((!i) & 3), not (!(i & 3))
		{
			ob->yspeed++;
			if (ob->yspeed > 70)
			{
				ob->yspeed = 70;
			}
		}
		tmp_ymove += ob->yspeed;
	}
}

void AccelerateX(objtype *ob, Sint16 dir, Sint16 maxspeed)
{
	Sint32 i;
	Uint16 oldsign;
	
	oldsign = ob->xspeed & 0x8000;
	for (i=lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i & 1)
		{
			ob->xspeed += dir;
			if ((ob->xspeed & 0x8000) != oldsign)
			{
				oldsign = ob->xspeed & 0x8000;
				ob->xdir = oldsign? -1 : 1;
			}
			if (ob->xspeed > maxspeed)
			{
				ob->xspeed = maxspeed;
			}
			else if (ob->xspeed < -maxspeed)
			{
				ob->xspeed = -maxspeed;
			}
		}
		tmp_xmove += ob->xspeed;
	}
}

void AccelerateX_0(objtype *ob, Sint16 dir, Sint16 maxspeed)
{
	Sint32 i;

	for (i=lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i & 1)
		{
			ob->xspeed += dir;
			if (ob->xspeed > maxspeed)
			{
				ob->xspeed = maxspeed;
			}
			else if (ob->xspeed < -maxspeed)
			{
				ob->xspeed = -maxspeed;
			}
		}
		tmp_xmove += ob->xspeed;
	}
}

void AccelerateY(objtype *ob, Sint16 dir, Sint16 maxspeed)
{
	Sint32 i;

	for (i=lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i & 1)
		{
			ob->yspeed += dir;
			if (ob->yspeed > maxspeed)
			{
				ob->yspeed = maxspeed;
			}
			else if (ob->yspeed < -maxspeed)
			{
				ob->yspeed = -maxspeed;
			}
		}
		tmp_ymove += ob->yspeed;
	}
}

void FrictionX(objtype *ob)
{
	Sint16 friction, oldsign;
	Sint32 i;

	oldsign = ob->xspeed & 0x8000;
	if (ob->xspeed > 0)
	{
		friction = -1;
	}
	else if (ob->xspeed < 0)
	{
		friction = 1;
	}
	else
	{
		friction = 0;
	}
	for (i=lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i & 1)
		{
			ob->xspeed += friction;
			if ((ob->xspeed & 0x8000) != oldsign)
			{
				ob->xspeed = 0;
			}
		}
		tmp_xmove += ob->xspeed;
	}
}

void FrictionY(objtype *ob)
{
	Sint16 friction, oldsign;
	Sint32 i;

	if (ob->yspeed > 0)
	{
		friction = -1;
	}
	else if (ob->yspeed < 0)
	{
		friction = 1;
	}
	else
	{
		friction = 0;
	}
	for (i=lasttimecount-tics; i<lasttimecount; i++)
	{
		if (i & 1)
		{
			ob->yspeed += friction;
			if ((ob->yspeed & 0x8000) != oldsign)	//BUG: oldsign is not initialized!
			{
				ob->yspeed = 0;
			}
		}
		tmp_ymove += ob->yspeed;
	}
}

void StunContact(objtype *ob, objtype *shot, statetype *newstate)
{
	ExplodeShot(shot);
	ob->temp1 = ob->temp2 = ob->temp3 = 0;
	ob->temp4 = ob->obclass;
	ChangeState(ob, newstate);
	ob->obclass = stunnedobj;
#ifndef KEEN4
	ob->yspeed -= 24;
	if (ob->yspeed < -48)
		ob->yspeed = -48;
#endif
}

void ProjectileThink4(objtype *ob)
{
	DoGravity4(ob);
	tmp_xmove = ob->xspeed*tics;
}

void ProjectileThink3(objtype *ob)
{
	DoGravity3(ob);
	tmp_xmove = ob->xspeed*tics;
}

/*
void ProjectileThink1(objtype *ob)	//not present in Keen 4-6
{
	DoGravity1(ob);
	tmp_xmove = ob->xspeed*tics;
}
*/

void VelocityThink(objtype *ob)
{
	tmp_xmove = ob->xspeed*tics;
	tmp_ymove = ob->yspeed*tics;
}

void NeedToReactThink(objtype *ob)
{
	ob->needtoreact = true;
}

void NeedToReactThink2(objtype *ob)
{
	ob->temp1 = 0;
	ob->needtoreact = true;
	ob->temp2++;
	if (ob->temp2 == 3)
		ob->temp2 = 0;
}

void HazardContact(objtype *ob, objtype *hit)
{
	ob++;
	if (hit->obclass == keenobj)
	{
		KillKeen();
	}
}

void DrawReact(objtype *ob)
{
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void WalkReact(objtype *ob)
{
	if (ob->xdir == 1 && ob->hitwest)
	{
		ob->x -= ob->xmove;
		ob->xdir = -1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (ob->xdir == -1 && ob->hiteast)
	{
		ob->x -= ob->xmove;
		ob->xdir = 1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (!ob->hitnorth)
	{
		ob->x -= ob->xmove;
		ob->xdir = -ob->xdir;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void WalkReact2(objtype *ob)
{
	if (ob->xdir == 1 && ob->hitwest)
	{
		ob->x -= ob->xmove;
		ob->xdir = -1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (ob->xdir == -1 && ob->hiteast)
	{
		ob->x -= ob->xmove;
		ob->xdir = 1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (!ob->hitnorth || (ob->hitnorth & ~7))
	{
		ob->x -= ob->xmove*2;
		ob->xdir = -ob->xdir;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

#pragma argsused
void BadState(objtype *ob)
{
	Quit("Object with bad state!");
}

void StunnedReact(objtype *ob)
{
	Sint16 dx, dy;

	if (ob->hitwest || ob->hiteast)
		ob->xspeed = 0;

	if (ob->hitsouth)
		ob->yspeed = 0;

	if (ob->hitnorth)
	{
		ob->xspeed = ob->yspeed = 0;
		if (ob->state->nextstate)
			ChangeState(ob, ob->state->nextstate);
	}

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
	dx = dy = 0;
	switch (ob->temp4)
	{
#if defined KEEN4
	case mimrockobj:
		dy = -4*PIXGLOBAL;
		break;
	case eggobj:
		dx = 8*PIXGLOBAL;
		dy = -8*PIXGLOBAL;
		break;
	case treasureeaterobj:
		dx = 8*PIXGLOBAL;
		break;
	case bounderobj:
		dx = 4*PIXGLOBAL;
		dy = -8*PIXGLOBAL;
		break;
	case wormouthobj:
		dx = 4*PIXGLOBAL;
		dy = -350;	// -21.875 pixels (this one is a bit strange)
		break;
	case lickobj:
		dy = -8*PIXGLOBAL;
		break;
	case inchwormobj:
		dx = -4*PIXGLOBAL;
		dy = -8*PIXGLOBAL;
		break;
	case slugobj:
		dy = -8*PIXGLOBAL;
		break;
// Return To The Shaowlands addition:
	case badoingobj:
		dy = -36*PIXGLOBAL;
		dx = -4*PIXGLOBAL;
		break;
#elif defined KEEN5
	case sparkyobj:
		dx += 4*PIXGLOBAL;
		break;
	case amptonobj:
		dy -= 8*PIXGLOBAL;
		break;
	case korathobj:
		dy -= 8*PIXGLOBAL;
		break;
#elif defined KEEN6
	case blooguardobj:
		dx = 16*PIXGLOBAL;
		dy = -4*PIXGLOBAL;
		break;
	case flectobj:
		dx = 4*PIXGLOBAL;
		dy = -4*PIXGLOBAL;
		break;
	case bloogobj:
	case nospikeobj:
		dx = 8*PIXGLOBAL;
		dy = -4*PIXGLOBAL;
		break;
	case bloogletobj:
	case babobbaobj:
		dy = -8*PIXGLOBAL;
		break;
	case fleexobj:
		dx = 16*PIXGLOBAL;
		dy = 8*PIXGLOBAL;
		break;
	case ceilickobj:
		dy = 12*PIXGLOBAL;
		break;
	default:
		Quit("No star spec for object!");
#endif
	}
	ob->temp1 = ob->temp1 + tics;
	if (ob->temp1 > 10)
	{
		ob->temp1 -= 10;
		ob->temp2++;
		if (ob->temp2 == 3)
			ob->temp2 = 0;
	}
	RF_PlaceSprite((void **)&ob->temp3, ob->x+dx, ob->y+dy, ob->temp2+STARS1SPR, 0, 3);
}

statetype s_scorebox = {0, 0, think, false, push_none, 0, 0, 0, NULL, NULL, NULL, NULL};
#pragma warn -sus	//BadState is not a valid contact function. Nobody cares.
statetype s_badstate  = {0, 0, think, false, push_none, 0, 0, 0, &BadState, &BadState, &BadState, NULL};
#pragma warn +sus