/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Wolfenstein 3-D Source Code
 * Copyright (C) 1992 id Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "CK_DEF.H"

/////////////////////////////////////////////////////////////////////////////
// local defines:
/////////////////////////////////////////////////////////////////////////////

#define BACKCOLOR RED

#define WORDLIMIT 80
#define FONTHEIGHT 10
#define TOPMARGIN 10
#define BOTTOMMARGIN 10
#define LEFTMARGIN 10
#define RIGHTMARGIN 10
#define PICMARGIN 8
#define SPACEWIDTH 7
#define TEXTROWS ((200-TOPMARGIN-BOTTOMMARGIN)/FONTHEIGHT)
#define SCREENPIXWIDTH 320
#define SCREENMID (SCREENPIXWIDTH/2)

/////////////////////////////////////////////////////////////////////////////
// uninitialized variables:
/////////////////////////////////////////////////////////////////////////////

static Sint16 helpmenupos;
static boolean layoutdone;
static Sint16 picdelay;
static Sint16 picnum;
static Sint16 picy;
static Sint16 picx;
static Sint16 rowon;
static char far *text;
static Sint16 rightmargin[TEXTROWS];
static Sint16 leftmargin[TEXTROWS];
static Sint16 numpages;
static Sint16 pagenum;

/////////////////////////////////////////////////////////////////////////////
// functions:
/////////////////////////////////////////////////////////////////////////////

static void RipToEOL(void)
{
	while (*text++ != '\n');
}

static Sint16 ParseNumber(void)
{
	char c, buffer[80];
	char *bufptr;

	c = *text;
	while (c < '0' || c > '9')
	{
		text++;
		c = *text;
	}
	bufptr = buffer;
	do
	{
		*bufptr = c;
		bufptr++;
		text++;
		c = *text;
	} while (c >= '0' && c <= '9');
	*bufptr = '\0';
	return atoi(buffer);
}

static void ParsePicCommand(void)
{
	picy = ParseNumber();
	picx = ParseNumber();
	picnum = ParseNumber();
	RipToEOL();
}

static void ParseTimedCommand(void)
{
	picy = ParseNumber();
	picx = ParseNumber();
	picnum = ParseNumber();
	picdelay = ParseNumber();
	RipToEOL();
}

static void TimedPicCommand(void)
{
	ParseTimedCommand();
	VW_WaitVBL(1);
	VW_ScreenToScreen(bufferofs, displayofs, 40, 200);
	TimeCount = 0;
	while (picdelay > TimeCount);
	VWB_DrawPic(picx & ~7, picy, picnum);
}

static void HandleCommand(void)
{
	Sint16 i,margin,top,bottom;
	Sint16 picwidth,picheight,picmid;

	switch (toupper(*(++text)))
	{
	case 'B':
		picy = ParseNumber();
		picx = ParseNumber();
		picwidth = ParseNumber();
		picheight = ParseNumber();
		VWB_Bar(picx, picy, picwidth, picheight, BACKCOLOR);
		RipToEOL();
		break;
	case 'E':
	case 'P':
		layoutdone = true;
		text--;
		break;
	case 'C':
		i = toupper(*(++text));
		if (i >= '0' && i <= '9')
		{
			fontcolor = i + 0 - '0';
		}
		else if (i >= 'A' && i <= 'F')
		{
			fontcolor = i + 10 - 'A';
		}
		fontcolor ^= BACKCOLOR;
		text++;
		break;
	case 'L':
		py = ParseNumber();
		rowon = (py - 10)/10;
		py = rowon * 10 + 10;
		px = ParseNumber();
		while (*(text++) != '\n');
		break;
	case 'T':
		TimedPicCommand();
		break;
	case 'G':
		ParsePicCommand();
		VWB_DrawPic(picx & ~7, picy, picnum);
		picwidth = pictable[picnum-STARTPICS].width * 8;
		picheight = pictable[picnum-STARTPICS].height;
		picmid = picx + picwidth/2;
		if (picmid > SCREENMID)
		{
			margin = picx-PICMARGIN;
		}
		else
		{
			margin = picx+picwidth+PICMARGIN;
		}
		top = (picy-TOPMARGIN)/FONTHEIGHT;
		if (top < 0)
		{
			top = 0;
		}
		bottom = (picy+picheight-TOPMARGIN)/FONTHEIGHT;
		if (bottom >= TEXTROWS)
		{
			bottom = TEXTROWS-1;
		}
		for (i=top; i<=bottom; i++)
		{
			if (picmid > SCREENMID)
			{
				rightmargin[i] = margin;
			}
			else
			{
				leftmargin[i] = margin;
			}
		}
		if (leftmargin[rowon] > px)
		{
			px = leftmargin[rowon];
		}
		break;
	}
}

static void NewLine(void)
{
	char c;

	if (++rowon == TEXTROWS)
	{
		layoutdone = true;
		do
		{
			if (*text == '^')
			{
				c = toupper(text[1]);
				if (c == 'E' || c == 'P')
				{
					layoutdone = true;
					return;
				}
			}
			text++;
		} while (1);
	}
	px = leftmargin[rowon];
	py += FONTHEIGHT;
}

static void HandleCtrls(void)
{
	char c;

	c = *(text++);
	if (c == '\n')
	{
		NewLine();
	}
}

static void HandleWord(void)
{
	Uint16 wwidth, wheight, newpos, wordindex;
	char word[WORDLIMIT];

	word[0] = *(text++);
	wordindex = 1;
	while (*text > ' ')
	{
		word[wordindex] = *(text++);
		if (++wordindex == WORDLIMIT)
		{
			Quit("PageLayout: Word limit exceeded");
		}
	}
	word[wordindex] = '\0';
	VW_MeasurePropString(word, &wwidth, &wheight);
	while (rightmargin[rowon] < px+wwidth)
	{
		NewLine();
		if (layoutdone)
		{
			return;
		}
	}
	newpos = px+wwidth;
	VWB_DrawPropString(word);
	px = newpos;
	while (*text == ' ')
	{
		px += SPACEWIDTH;
		text++;
	}
}

static void PageLayout(boolean shownumber)
{
	Sint16 oldcolor, i;
	char c;

	oldcolor = fontcolor;
	fontcolor = YELLOW^BACKCOLOR;
	VWB_Bar(0, 0, 320, 200, BACKCOLOR);
#ifndef KEEN6
	VWB_DrawPic(  0, 0, TEXTBORDERTOPPIC);
	VWB_DrawPic(  0, 8, TEXTBORDERLEFTPIC);
	VWB_DrawPic(312, 8, TEXTBORDERRIGHTPIC);
	if (shownumber)
	{
		VWB_DrawPic(8, 176, TEXTBORDERBOTTOM1PIC);
	}
	else
	{
		VWB_DrawPic(8, 192, TEXTBORDERBOTTOM2PIC);
	}
#endif
	for (i=0; i<TEXTROWS; i++)
	{
		leftmargin[i] = LEFTMARGIN;
		rightmargin[i] = SCREENPIXWIDTH-RIGHTMARGIN;
	}
	px = LEFTMARGIN;
	py = TOPMARGIN;
	rowon = 0;
	layoutdone = false;
	while (*text <= ' ')
	{
		text++;
	}
	if (*text != '^' || toupper(*(++text)) != 'P')
	{
		Quit("PageLayout: Text not headed with ^P");
	}
	while (*(text++) != '\n');

	do
	{
		c = *text;
		if (c == '^')
		{
			HandleCommand();
		}
		else if (c <= ' ')
		{
			HandleCtrls();
		}
		else
		{
			HandleWord();
		}
	} while (!layoutdone);
	pagenum++;
	if (shownumber)
	{
		strcpy(str, "Pg ");
		itoa(pagenum, str2, 10);
		strcat(str, str2);
		strcat(str, " of ");
		itoa(numpages, str2, 10);
		strcat(str, str2);
		fontcolor = LIGHTRED^BACKCOLOR;
		py = 186;
		px = 218;
		VWB_DrawPropString(str);
	}
	fontcolor = oldcolor;
}

static void BackPage(void)
{
	pagenum--;
	do
	{
		text--;
		if (text[0] == '^' && toupper(text[1]) == 'P')
		{
			return;
		}
	} while (1);
}

static void CacheLayoutGraphics(void)
{
	char	far *bombpoint, far *textstart;
	char	ch;

	textstart = text;
	bombpoint = text+30000;
	numpages = pagenum = 0;
#ifndef KEEN6
	CA_MarkGrChunk(TEXTBORDERTOPPIC);
	CA_MarkGrChunk(TEXTBORDERLEFTPIC);
	CA_MarkGrChunk(TEXTBORDERRIGHTPIC);
	CA_MarkGrChunk(TEXTBORDERBOTTOM1PIC);
	CA_MarkGrChunk(TEXTBORDERBOTTOM2PIC);
#endif
	do
	{
		if (*text == '^')
		{
			ch = toupper(*(++text));
			if (ch == 'P')
			{
				numpages++;
			}
			if (ch == 'E')
			{
				CA_CacheMarks(NULL);
				text = textstart;
				return;
			}
			if (ch == 'G')
			{
				ParsePicCommand();
				CA_MarkGrChunk(picnum);
			}
			if (ch == 'T')
			{
				ParseTimedCommand();
				CA_MarkGrChunk(picnum);
			}
		}
		else
		{
			text++;
		}
	} while (text < bombpoint);
	Quit("CacheLayoutGraphics: No ^E to terminate file!");
}

#ifndef KEEN6

static Sint16 HelpMenu(void)
{
	CursorInfo cursor;
	ControlInfo control;
	Sint16 ydelta;
	Uint16 key;

	VWB_Bar(0, 0, 320, 200, BACKCOLOR);
	CA_CacheGrChunk(HELPMENUPIC);
	CA_CacheGrChunk(HELPCURSORPIC);
	CA_CacheGrChunk(TEXTBORDERTOPPIC);
	CA_CacheGrChunk(TEXTBORDERLEFTPIC);
	CA_CacheGrChunk(TEXTBORDERRIGHTPIC);
	CA_CacheGrChunk(TEXTBORDERBOTTOM2PIC);
	VWB_DrawPic(  0,   0, TEXTBORDERTOPPIC);
	VWB_DrawPic(  0,   8, TEXTBORDERLEFTPIC);
	VWB_DrawPic(312,   8, TEXTBORDERRIGHTPIC);
	VWB_DrawPic(  8, 192, TEXTBORDERBOTTOM2PIC);
	VWB_DrawPic( 96,   8, HELPMENUPIC);
	ydelta = 0;
	IN_ClearKeysDown();
	do
	{
		if (helpmenupos < 0)
		{
			helpmenupos = 0;
		}
		else if (helpmenupos > 4)
		{
			helpmenupos = 4;
		}
		VWB_DrawPic(48, 24*helpmenupos+48, HELPCURSORPIC);
		VW_UpdateScreen();
		VWB_Bar(48, 24*helpmenupos+48, 39, 24, BACKCOLOR);
		IN_ReadControl(0, &control);
		IN_ReadCursor(&cursor);
		if (LastScan)
		{
			key = LastScan;
			IN_ClearKeysDown();
			switch (key)
			{
			case sc_UpArrow:
				helpmenupos--;
				break;
			case sc_DownArrow:
				helpmenupos++;
				break;
			case sc_Enter:
				VW_ClearVideo(BACKCOLOR);
				return helpmenupos;
			case sc_Escape:
				VW_ClearVideo(BACKCOLOR);
				return -1;
			}
		}
		ydelta += cursor.y;
		if (cursor.button0 || cursor.button1 || control.button0 || control.button1)
		{
			VW_ClearVideo(BACKCOLOR);
			return helpmenupos;
		}
		if (ydelta < -40)
		{
			ydelta += 40;
			helpmenupos--;
		}
		else if (ydelta > 40)
		{
			ydelta -= 40;
			helpmenupos++;
		}
	} while (1);
}

void HelpScreens(void)
{
	static Uint16 textchunks[] =
	{
		COMPUTERWRISTTEXT,
		GAMEHELPTEXT,
		STORYTEXT,
		ORDERTEXT,
		ABOUTIDTEXT
	};

	Uint16 olddisplayofs, oldbufferofs, oldfontnumber, temp;
	Sint16 pos;
	boolean newpage;

	oldfontnumber = fontnumber;
	olddisplayofs = displayofs;
	oldbufferofs = bufferofs;
	fontnumber = 0;
	EGAMAPMASK(15);
	CA_UpLevel();
	CA_SetGrPurge();
	VW_ClearVideo(BACKCOLOR);
	RF_FixOfs();
	bufferofs = 0;
	displayofs = 0x8000;
	VW_SetScreen(displayofs, 0);
#ifdef KEEN5
	StartMusic(19);
#endif
	do
	{
		pos = HelpMenu();
		VW_ClearVideo(BACKCOLOR);
		if (pos == -1)
		{
			CA_DownLevel();
			IN_ClearKeysDown();
			bufferofs = oldbufferofs;
			displayofs = olddisplayofs;
			fontnumber = oldfontnumber;
			VW_ClearVideo(BACKCOLOR);
			RF_FixOfs();
#ifdef KEEN5
			StopMusic();	// Note: it's safer to call StopMusic BEFORE CA_DownLevel
#endif
			return;
		}
		pos = textchunks[pos];
		CA_CacheGrChunk(pos);
		MM_SetLock(&grsegs[pos], true);	// RTTS fix
		text = grsegs[pos];
		CA_ClearMarks();	// RTTS fix
		CacheLayoutGraphics();
		newpage = true;
		do
		{
			if (newpage)
			{
				newpage = false;
				PageLayout(true);
				VW_SetScreen(bufferofs, 0);
				temp = displayofs;
				displayofs = bufferofs;
				bufferofs = temp;
			}
			LastScan = 0;
			while (!LastScan);
			switch (LastScan)
			{
			case sc_UpArrow:
			case sc_LeftArrow:
			case sc_PgUp:
				if (pagenum > 1)
				{
					BackPage();
					BackPage();
					newpage = true;
				}
				break;
			case sc_DownArrow:
			case sc_RightArrow:
			case sc_PgDn:
				if (pagenum < numpages)
				{
					newpage = true;
				}
				break;
			}
		} while (LastScan != sc_Escape);
		MM_FreePtr(&grsegs[pos]);
		IN_ClearKeysDown();
	} while (true);
}

#endif

void EndText(void)
{
	char _seg *textseg;
	Sint16 i;

	VW_ClearVideo(BACKCOLOR);
	RF_FixOfs();
	CA_UpLevel();
	CA_ClearMarks();	// RTTS fix
	CA_SetGrPurge();
	CA_CacheGrChunk(TEXTARROW2PIC);
	CA_CacheGrChunk(TEXTARROW1PIC);

#ifdef KEEN5
	if (gamestate.leveldone[13] == ex_brokefuses)
	{
		CA_CacheGrChunk(SECRETENDINGTEXT);
		textseg = grsegs[SECRETENDINGTEXT];
	}
	else
	{
		CA_CacheGrChunk(ENDINGTEXT);
		textseg = grsegs[ENDINGTEXT];
	}
#else
	CA_CacheGrChunk(ENDINGTEXT);
	MM_SetLock(&grsegs[ENDINGTEXT], true);	// RTTS fix
	textseg = grsegs[ENDINGTEXT];
#endif
	text = textseg;
	CacheLayoutGraphics();

	StartMusic(ENDINGMUSIC);

	while (pagenum < numpages)
	{
		PageLayout(false);
		IN_ClearKeysDown();
		VW_SetScreen(bufferofs, 0);
		do {
			VWB_DrawPic(298, 184, TEXTARROW1PIC);
			for (i=0; i<TickBase; i++)
			{
				if (IN_IsUserInput())
				{
					goto nextpage;
				}
				VW_WaitVBL(1);
			}
			VWB_DrawPic(298, 184, TEXTARROW2PIC);
			for (i=0; i<TickBase; i++)
			{
				if (IN_IsUserInput())
				{
					goto nextpage;
				}
				VW_WaitVBL(1);
			}
		} while (1);
nextpage:
	}
	StopMusic();
#ifdef KEEN5
	if (gamestate.leveldone[13] == ex_brokefuses)
	{
		MM_FreePtr(&grsegs[SECRETENDINGTEXT]);
	}
	else
	{
		MM_FreePtr(&grsegs[TEXTARROW1PIC]);	// BUG! this should free ENDINGTEXT, the arrow should be freed after the else branch!
	}
#else
	MM_FreePtr(&grsegs[ENDINGTEXT]);
	MM_FreePtr(&grsegs[TEXTARROW1PIC]);
#endif
	MM_FreePtr(&grsegs[TEXTARROW2PIC]);
	CA_DownLevel();
	IN_ClearKeysDown();
	VW_ClearVideo(BACKCOLOR);
	RF_FixOfs();
	CA_FreeGraphics();
}