/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
K4_ACT1.C
=========

Contains the following actor types (in this order):

- Miragia (world map)
- Bonus Items
- Oracle Member
- Poison Slug
- Mad Mushroom
- Egg & Eagle
- Arachnut
- Skypest

*/

#include "CK_DEF.H"

/*
=============================================================================
	MIRAGIA
=============================================================================
*/
#if 0	// disabled for RTTS
statetype s_miragia1 = {0, 0, step, false, push_none, 300, 0, 0, MiragiaThink1, NULL, NULL, &s_miragia2};
statetype s_miragia2 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink2, NULL, NULL, &s_miragia3};
statetype s_miragia3 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink3, NULL, NULL, &s_miragia4};
statetype s_miragia4 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink4, NULL, NULL, &s_miragia5};
statetype s_miragia5 = {0, 0, step, false, push_none, 300, 0, 0, MiragiaThink5, NULL, NULL, &s_miragia6};
statetype s_miragia6 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink6, NULL, NULL, &s_miragia7};
statetype s_miragia7 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink7, NULL, NULL, &s_miragia8};
statetype s_miragia8 = {0, 0, step, false, push_none, 30, 0, 0, MiragiaThink8, NULL, NULL, &s_miragia1};

void SpawnMiragia(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = decoobj;
	new->active = allways;
	new->x = x;
	new->y = y;
	new->state = &s_miragia1;
}

void MiragiaThink1(objtype *ob)
{
	if (player->tileright >= ob->x && player->tileleft <= ob->x+5
		&& player->tiletop >= ob->y && player->tilebottom <= ob->y+4)
	{
		//don't let miragia appear if player is in the area, so player won't get stuck
		ob->state = &s_miragia8;
	}
	else
	{
		RF_MapToMap(8, 60, ob->x, ob->y, 6, 4);
	}
}

void MiragiaThink2(objtype *ob)
{
	RF_MapToMap(14, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink3(objtype *ob)
{
	RF_MapToMap(20, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink4(objtype *ob)
{
	RF_MapToMap(26, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink5(objtype *ob)
{
	RF_MapToMap(20, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink6(objtype *ob)
{
	RF_MapToMap(14, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink7(objtype *ob)
{
	RF_MapToMap(8, 60, ob->x, ob->y, 6, 4);
}

void MiragiaThink8(objtype *ob)
{
	RF_MapToMap(2, 60, ob->x, ob->y, 6, 4);
}
#endif

/*
=============================================================================
	BONUS ITEMS

	temp1 - bonus item number
	temp2 - first sprite in animation loop
	temp3 - end of animation loop (non-inclusive)
=============================================================================
*/

statetype s_bonus1 = {0, 0, step, false, push_none, 20, 0, 0, BonusThink, NULL, DrawReact, &s_bonus2};
statetype s_bonus2 = {0, 0, step, false, push_none, 20, 0, 0, BonusThink, NULL, DrawReact, &s_bonus1};
statetype s_bonusrise = {0, 0, slide, false, push_none, 40, 0, 8, NULL, NULL, DrawReact, NULL};

statetype s_splash1 = {DROPSPLASH1SPR, DROPSPLASH1SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_splash2};
statetype s_splash2 = {DROPSPLASH2SPR, DROPSPLASH2SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, &s_splash3};
statetype s_splash3 = {DROPSPLASH3SPR, DROPSPLASH3SPR, step, false, push_none, 10, 0, 0, NULL, NULL, DrawReact, NULL};

Uint16 bonusshape[] = {KEYGEMRED1SPR, KEYGEMYELLOW1SPR, KEYGEMBLUE1SPR, KEYGEMGREEN1SPR, BONUS1001SPR, BONUS2001SPR, BONUS5001SPR, BONUS10001SPR, BONUS20001SPR, BONUS50001SPR, EXTRALIFE1SPR, AMMO1SPR};

void SpawnBonus(Sint16 x, Sint16 y, Sint16 type)
{
	GetNewObj(false);
	new->needtoclip = false;
	new->priority = 2;
	new->obclass = bonusobj;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	new->ydir = -1;
	new->temp1 = type;
	new->temp2 = new->shapenum = bonusshape[type];
	new->temp3 = new->temp2 + 2;
	NewState(new, &s_bonus1);
}

void SpawnSplash(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->needtoclip = false;
	new->priority = 3;
	new->obclass = decoobj;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	NewState(new, &s_splash1);
}

void BonusThink(objtype *ob)
{
	ob->shapenum++;
	if (ob->shapenum == ob->temp3)
		ob->shapenum = ob->temp2;
}

/*
=============================================================================
	ORACLE MEMBER
=============================================================================
*/

statetype s_oraclewalk1 = {ORACLEWALK1LSPR, ORACLEWALK1RSPR, step, false, push_down, 10, 64, 0, OracleWalkThink, NULL, WalkReact, &s_oraclewalk2};
statetype s_oraclewalk2 = {ORACLEWALK2LSPR, ORACLEWALK2RSPR, step, false, push_down, 10, 64, 0, OracleWalkThink, NULL, WalkReact, &s_oraclewalk1};
statetype s_oraclestand = {ORACLETHINKLSPR, ORACLETHINKRSPR, step, false, push_none, 120, 128, 0, NULL, NULL, DrawReact, &s_oraclewalk1};

void SpawnOracle(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = oracleobj;
	new->active = yes;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) -0x171;	// weird adjustment
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_oraclewalk1);
}

void OracleWalkThink(objtype *ob)
{
	Uint16 randnum;

	randnum = US_RndT();
	if (tics*8 > randnum)
	{
		tmp_xmove = 0;	// RTTS addition: don't move this time
		ob->state = &s_oraclestand;
	}
}

/*
=============================================================================
	POISON SLUG
=============================================================================
*/

statetype s_slugwalk1 = {SLUGWALKL1SPR, SLUGWALKR1SPR, step, false, push_down, 8, 64, 0, NULL, SlugContact, WalkReact2, &s_slugwalk2};
statetype s_slugwalk2 = {SLUGWALKL2SPR, SLUGWALKR2SPR, step, false, push_down, 8, 64, 0, SlugWalkThink, SlugContact, WalkReact2, &s_slugwalk1};
statetype s_slugpoo = {SLUGPOOLSPR, SLUGPOORSPR, step, false, push_down, 60, 64, 0, SlugPooThink, SlugContact, WalkReact2, &s_slugwalk1};
statetype s_slugstunned1 = {SLUGSTUNNED1SPR, SLUGSTUNNED1SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};
statetype s_slugstunned2 = {SLUGSTUNNED2SPR, SLUGSTUNNED2SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};
statetype s_slugstain1 = {SLUGSTAIN1SPR, SLUGSTAIN1SPR, step, false, push_none, 300, 0, 0, NULL, HazardContact, DrawReact, &s_slugstain2};
statetype s_slugstain2 = {SLUGSTAIN2SPR, SLUGSTAIN2SPR, step, false, push_none, 60, 0, 0, NULL, NULL, DrawReact, NULL};

void SpawnSlug(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = slugobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) - 0x71;
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_slugwalk1);
}

void SlugWalkThink(objtype *ob)
{
	if (US_RndT() < 16)
	{
		if (ob->x < player->x)
		{
			ob->xdir = 1;
		}
		else
		{
			ob->xdir = -1;
		}
		ob->state = &s_slugpoo;
		SD_PlaySound(SND_SLUGPOO);
	}
}

void SlugPooThink(objtype *ob)
{
	GetNewObj(true);
	new->obclass = decoobj;
	new->active = removable;
	new->priority = 0;
	new->x = ob->x;
	new->y = ob->bottom - 8*PIXGLOBAL;
	NewState(new, &s_slugstain1);
}

void SlugContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
	{
		KillKeen();
	}
	else if (hit->obclass == shotobj)
	{
		if (US_RndT() < 0x80)
		{
			StunContact(ob, hit, &s_slugstunned1);
		}
		else
		{
			StunContact(ob, hit, &s_slugstunned2);
		}
		ob->yspeed = -24;
		ob->xspeed = ob->xdir*8;
	}
}

/*
=============================================================================
	MAD MUSHROOM

	temp1 - bounce counter
=============================================================================
*/

statetype s_madmushroom1 = {MADMUSHROOML1SPR, MADMUSHROOMR1SPR, stepthink, false, push_none, 8, 0, 0, MadMushroomThink, MadMushroomContact, MadMushroomReact, &s_madmushroom2};
statetype s_madmushroom2 = {MADMUSHROOML2SPR, MADMUSHROOMR2SPR, stepthink, false, push_none, 8, 0, 0, MadMushroomThink, MadMushroomContact, MadMushroomReact, &s_madmushroom1};

void SpawnMadMushroom(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = madmushroomobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) - 0xF1;
	new->xdir = 1;
	new->ydir = 1;
	NewState(new, &s_madmushroom1);
}

void MadMushroomThink(objtype *ob)
{
	if (player->x < ob->x)
	{
		ob->xdir = -1;
	}
	else
	{
		ob->xdir = 1;
	}
	DoGravity3(ob);
}

void MadMushroomContact(objtype *ob, objtype *hit)
{
	ob++;
	if (hit->obclass == shotobj)
	{
		ExplodeShot(hit);
	}
	else if (hit->obclass == keenobj)
	{
		KillKeen();
	}
}

void MadMushroomReact(objtype *ob)
{
	if (ob->hitsouth)
		ob->yspeed = 0;

	if (ob->hitnorth)
	{
		ob->yspeed = 0;
		ob->temp1++;
		if (ob->temp1 == 3)
		{
			ob->temp1 = 0;
			ob->yspeed = -68;
			SD_PlaySound(SND_BOUNCE2);
		}
		else
		{
			SD_PlaySound(SND_BOUNCE1);
			ob->yspeed = -40;
		}
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

/*
=============================================================================
	EGG & EAGLE
=============================================================================
*/

statetype s_egg           = {EGGSPR, EGGSPR, think, false, push_down, 8, 0, 0, NULL, EggContact, DrawReact, NULL};
statetype s_brokenegg     = {BROKENEGGSPR, BROKENEGGSPR, step, false, push_none, 30000, 0, 0, NULL, NULL, DrawReact, NULL};
statetype s_eggpiece1     = {EGGPIECE1SPR, EGGPIECE1SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, EggPieceReact, NULL};
statetype s_eggpiece2     = {EGGPIECE2SPR, EGGPIECE2SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, EggPieceReact, NULL};
statetype s_eggpiece3     = {EGGPIECE3SPR, EGGPIECE3SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, EggPieceReact, NULL};

statetype s_eaglespawn    = {EAGLEWALK1LSPR, EAGLEWALK1RSPR, step, false, push_down, 120, 128, 0, NULL, EagleHarmlessContact, EagleWalkReact, &s_eaglewalk2};
statetype s_eaglewalk1    = {EAGLEWALK1LSPR, EAGLEWALK1RSPR, step, false, push_down, 7, 128, 0, NULL, EagleContact, EagleWalkReact, &s_eaglewalk2};
statetype s_eaglewalk2    = {EAGLEWALK2LSPR, EAGLEWALK2RSPR, step, false, push_down, 7, 128, 0, NULL, EagleContact, EagleWalkReact, &s_eaglewalk3};
statetype s_eaglewalk3    = {EAGLEWALK3LSPR, EAGLEWALK3RSPR, step, false, push_down, 7, 128, 0, NULL, EagleContact, EagleWalkReact, &s_eaglewalk4};
statetype s_eaglewalk4    = {EAGLEWALK4LSPR, EAGLEWALK4RSPR, step, false, push_down, 7, 128, 0, EagleWalkThink, EagleContact, EagleWalkReact, &s_eaglewalk1};
statetype s_eaglefly1     = {EAGLEFLY1SPR, EAGLEFLY1SPR, slidethink, false, push_none, 8, 0, 0, EagleFlyThink, EagleContact, EagleFlyReact, &s_eaglefly2};
statetype s_eaglefly2     = {EAGLEFLY2SPR, EAGLEFLY2SPR, slidethink, false, push_none, 8, 0, 0, EagleFlyThink, EagleContact, EagleFlyReact, &s_eaglefly3};
statetype s_eaglefly3     = {EAGLEFLY3SPR, EAGLEFLY3SPR, slidethink, false, push_none, 8, 0, 0, EagleFlyThink, EagleContact, EagleFlyReact, &s_eaglefly4};
statetype s_eaglefly4     = {EAGLEFLY4SPR, EAGLEFLY4SPR, slidethink, false, push_none, 8, 0, 0, EagleFlyThink, EagleContact, EagleFlyReact, &s_eaglefly1};
statetype s_eagleUNUSED   = {EAGLEFLY4SPR, EAGLEFLY4SPR, think, false, push_none, 8, 128, 0, ProjectileThink3, EagleContact, EagleUnusedReact, NULL};
statetype s_eaglestunned1 = {EAGLESTUNNEDSPR, EAGLESTUNNEDSPR, stepthink, false, push_none, 240, 0, 0, ProjectileThink4, EagleHarmlessContact, DrawReact, &s_eaglestunned2};
statetype s_eaglestunned2 = {EAGLEWALK1LSPR, EAGLEWALK1RSPR, step, false, push_down, 20, 0, 0, NULL, EagleHarmlessContact, DrawReact, &s_eaglestunned3};
statetype s_eaglestunned3 = {EAGLESTUNNEDSPR, EAGLESTUNNEDSPR, step, false, push_down, 20, 0, 0, NULL, EagleHarmlessContact, DrawReact, &s_eaglestunned4};
statetype s_eaglestunned4 = {EAGLEWALK1LSPR, EAGLEWALK1RSPR, step, false, push_down, 20, 0, 0, NULL, EagleHarmlessContact, DrawReact, &s_eaglestunned5};
statetype s_eaglestunned5 = {EAGLESTUNNEDSPR, EAGLESTUNNEDSPR, step, false, push_down, 20, 0, 0, EagleStunnedThink, EagleHarmlessContact, DrawReact, &s_eaglewalk1};

void SpawnEgg(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = eggobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) - 0x71;
	new->xdir = 1;
	new->ydir = 1;
	NewState(new, &s_egg);
}

void EagleStunnedThink(objtype *ob)
{
	ob->obclass = eagleobj;
}

void SpawnEagle(Sint16 x, Sint16 y)
{
	GetNewObj(true);
	new->obclass = eagleobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) - 0xF1;
	if (new->x < player->x)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_eaglespawn);
}

void EggContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj || hit->obclass == keenobj)
	{
		if (hit->obclass == shotobj)
			ExplodeShot(hit);

		ob->obclass = decoobj;
		ob->active = removable;
		ChangeState(ob, &s_brokenegg);

		GetNewObj(true);
		new->obclass = eagleobj;
		new->active = yes;
		new->x = ob->x;
		new->y = ob->y - 8*PIXGLOBAL;
		if (ob->x < player->x)
		{
			new->xdir = 1;
		}
		else
		{
			new->xdir = -1;
		}
		new->ydir = 1;
		NewState(new, &s_eaglespawn);

		GetNewObj(true);
		new->obclass = decoobj;
		new->active = removable;
		new->x = ob->x;
		new->y = ob->y;
		new->xspeed = -28;
		new->yspeed = -40;
		NewState(new, &s_eggpiece1);

		GetNewObj(true);
		new->obclass = decoobj;
		new->active = removable;
		new->x = ob->x;
		new->y = ob->y;
		new->xspeed = 28;
		new->yspeed = -40;
		NewState(new, &s_eggpiece2);

		GetNewObj(true);
		new->obclass = decoobj;
		new->active = removable;
		new->x = ob->x;
		new->y = ob->y;
		new->xspeed = 0;
		new->yspeed = -56;
		NewState(new, &s_eggpiece3);
	}
}

void EagleWalkThink(objtype *ob)
{
	if (ob->x < player->x)
	{
		ob->xdir = 1;
	}
	else
	{
		ob->xdir = -1;
	}
	if (ob->bottom >= player->bottom + 3*TILEGLOBAL && ob->hitnorth
		&& StateOkay(ob, &s_eaglefly1))	// BUG: StateOkay() only works for normal clipping, not for full clipping
	{
		ob->state = &s_eaglefly1;
		ob->needtoclip = FULLCLIP;
		ob->yspeed = -8;
		ob->temp1 = 0;
	}
}

void EagleFlyThink(objtype *ob)
{
	if (ob->temp1 == 0)
	{
		if (ob->x < player->x)
		{
			ob->xdir = 1;
		}
		else
		{
			ob->xdir = -1;
		}
	}
	AccelerateX_0(ob, ob->xdir, 16);
	if (ob->y < player->y)
	{
		AccelerateY(ob, 1, 16);
	}
	else
	{
		AccelerateY(ob, -1, 16);
	}
}

void EagleContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
	{
		KillKeen();
	}
	else if (hit->obclass == shotobj)
	{
		ob->xspeed = 0;
		ob->needtoclip = true;
		StunContact(ob, hit, &s_eaglestunned1);
	}
}

void EagleHarmlessContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		ob->xspeed = 0;
		StunContact(ob, hit, &s_eaglestunned1);
	}
}

void EagleWalkReact(objtype *ob)
{
	if (ob->xdir == 1 && ob->hitwest)
	{
		ob->x -= ob->xmove;
		ob->xdir = -1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (ob->xdir == -1 && ob->hiteast)
	{
		ob->x -= ob->xmove;
		ob->xdir = 1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	if (!ob->hitnorth)
	{
		ob->yspeed = -16;
		ob->needtoclip = FULLCLIP;
		ChangeState(ob, &s_eaglefly1);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void EagleUnusedReact(objtype *ob)	//never actually used
{
	if (ob->hitnorth)
	{
		ChangeState(ob, &s_eaglewalk1);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void EggPieceReact(objtype *ob)
{
	if (ob->hitnorth)
		ob->xspeed = ob->yspeed = 0;

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void EagleFlyReact(objtype *ob)
{
	statetype *oldstate;

	if ((ob->hitsouth || ob->hitnorth) && !ob->temp1)
		ob->temp1++;

	if (ob->hiteast && ob->xspeed < 0 || ob->hitwest && ob->xspeed > 0)
	{
		ob->xspeed = 0;
		ob->xdir = -ob->xdir;
	}
	if (ob->hitnorth == 1 && player->bottom - ob->bottom < 8*PIXGLOBAL)	// BUG? unsigned comparison!
	{
		oldstate = ob->state;
		ob->needtoclip = true;
		ChangeState(ob, &s_eaglewalk1);
		tmp_xmove = 0;
		tmp_ymove = 8*PIXGLOBAL;
		PushClipToWalls(ob);
		if (!ob->hitnorth)
		{
			ob->needtoclip = FULLCLIP;
			ChangeState(ob, oldstate);
		}
		return;
	}

	if (!ob->hitsouth && !ob->hitnorth)
		ob->temp1 = 0;

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

/*
=============================================================================
	ARACHNUT
=============================================================================
*/

statetype s_arachnutwalk1 = {ARACHNUTWALK1SPR, ARACHNUTWALK4SPR, step, false, push_down, 6, 128, 0, NULL, ArachnutContact, WalkReact, &s_arachnutwalk2};
statetype s_arachnutwalk2 = {ARACHNUTWALK2SPR, ARACHNUTWALK3SPR, step, false, push_down, 6, 128, 0, NULL, ArachnutContact, WalkReact, &s_arachnutwalk3};
statetype s_arachnutwalk3 = {ARACHNUTWALK3SPR, ARACHNUTWALK2SPR, step, false, push_down, 6, 128, 0, NULL, ArachnutContact, WalkReact, &s_arachnutwalk4};
statetype s_arachnutwalk4 = {ARACHNUTWALK4SPR, ARACHNUTWALK1SPR, step, false, push_down, 6, 128, 0, ArachnutThink, ArachnutContact, WalkReact, &s_arachnutwalk1};
statetype s_arachnutstunned1 = {ARACHNUTSTUNNEDSPR, ARACHNUTSTUNNEDSPR, step, false, push_down, 240, 0, 0, NULL, ArachnutStunnedContact, DrawReact, &s_arachnutstunned2};
statetype s_arachnutstunned2 = {ARACHNUTWALK1SPR, ARACHNUTWALK1SPR, step, false, push_down, 20, 0, 0, NULL, ArachnutStunnedContact, DrawReact, &s_arachnutstunned3};
statetype s_arachnutstunned3 = {ARACHNUTSTUNNEDSPR, ARACHNUTSTUNNEDSPR, step, false, push_down, 20, 0, 0, NULL, ArachnutStunnedContact, DrawReact, &s_arachnutstunned4};
statetype s_arachnutstunned4 = {ARACHNUTWALK1SPR, ARACHNUTWALK1SPR, step, false, push_down, 20, 0, 0, NULL, ArachnutStunnedContact, DrawReact, &s_arachnutstunned5};
statetype s_arachnutstunned5 = {ARACHNUTSTUNNEDSPR, ARACHNUTSTUNNEDSPR, step, false, push_down, 20, 0, 0, NULL, ArachnutStunnedContact, DrawReact, &s_arachnutwalk1};

void SpawnArachnut(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = arachnutobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) - 0x171;
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_arachnutwalk1);
}

void ArachnutThink(objtype *ob)
{
	if (ob->x > player->x)
	{
		ob->xdir = -1;
	}
	else
	{
		ob->xdir = 1;
	}
}

void ArachnutContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		StunContact(ob, hit, &s_arachnutstunned1);
	}
	else if (hit->obclass == keenobj)
	{
		KillKeen();
	}
}

void ArachnutStunnedContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		StunContact(ob, hit, &s_arachnutstunned1);
	}
}

/*
=============================================================================
	SKYPEST
=============================================================================
*/

statetype s_skypestfly1   = {SKYPESTFLY1LSPR, SKYPESTFLY1RSPR, stepthink, true, push_none, 5, 0, 0, SkypestFlyThink, SkypestFlyContact, SkypestFlyReact, &s_skypestfly2};
statetype s_skypestfly2   = {SKYPESTFLY2LSPR, SKYPESTFLY2RSPR, stepthink, true, push_none, 5, 0, 0, SkypestFlyThink, SkypestFlyContact, SkypestFlyReact, &s_skypestfly1};
statetype s_skypestsquished = {SKYPESTSQUISHEDSPR, SKYPESTSQUISHEDSPR, think, false, push_none, 0, 0, 0, NULL, NULL, DrawReact, &s_skypestsquished};
statetype s_skypeststand1 = {SKYPESTSIT9SPR, SKYPESTSIT9SPR, step, false, push_none, 100, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand2};
statetype s_skypeststand2 = {SKYPESTSIT1SPR, SKYPESTSIT1SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand3};
statetype s_skypeststand3 = {SKYPESTSIT2SPR, SKYPESTSIT2SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand4};
statetype s_skypeststand4 = {SKYPESTSIT3SPR, SKYPESTSIT3SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand5};
statetype s_skypeststand5 = {SKYPESTSIT4SPR, SKYPESTSIT4SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand6};
statetype s_skypeststand6 = {SKYPESTSIT3SPR, SKYPESTSIT3SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand7};
statetype s_skypeststand7 = {SKYPESTSIT2SPR, SKYPESTSIT2SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand8};
statetype s_skypeststand8 = {SKYPESTSIT1SPR, SKYPESTSIT1SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand9};
statetype s_skypeststand9 = {SKYPESTSIT9SPR, SKYPESTSIT9SPR, step, false, push_none, 60, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand10};
statetype s_skypeststand10 = {SKYPESTSIT5SPR, SKYPESTSIT5SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand11};
statetype s_skypeststand11 = {SKYPESTSIT6SPR, SKYPESTSIT6SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand12};
statetype s_skypeststand12 = {SKYPESTSIT7SPR, SKYPESTSIT7SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand13};
statetype s_skypeststand13 = {SKYPESTSIT8SPR, SKYPESTSIT8SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand14};
statetype s_skypeststand14 = {SKYPESTSIT7SPR, SKYPESTSIT7SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand15};
statetype s_skypeststand15 = {SKYPESTSIT6SPR, SKYPESTSIT6SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand16};
statetype s_skypeststand16 = {SKYPESTSIT5SPR, SKYPESTSIT5SPR, step, false, push_none, 10, 0, 0, NULL, SkypestStandContact, DrawReact, &s_skypeststand17};
statetype s_skypeststand17 = {SKYPESTSIT9SPR, SKYPESTSIT9SPR, step, false, push_none, 100, 0, 0, SkypestStandThink, SkypestStandContact, DrawReact, &s_skypestfly1};

void SpawnSkypest(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = skypestobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	if (US_RndT() < 0x80)
	{
		new->ydir = 1;
	}
	else
	{
		new->ydir = -1;
	}
	NewState(new, &s_skypestfly1);
}

void SkypestFlyThink(objtype *ob)
{
	if (US_RndT() < tics)
		ob->xdir = -ob->xdir;

	if (ob->ydir == -1 && US_RndT() < tics)
		ob->ydir = 1;

	if (ob->ydir == 1 && US_RndT() < tics*2)
		ob->ydir = -ob->ydir;

	AccelerateX(ob, ob->xdir, 20);
	AccelerateY(ob, ob->ydir, 20);
}

void SkypestFlyContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
		KillKeen();

	if (hit->obclass == shotobj)
	{
		if (hit->xdir == 1)
		{
			ob->xspeed = 20;
		}
		else if (hit->xdir == -1)
		{
			ob->xspeed = -20;
		}
		else if (hit->ydir == 1)
		{
			ob->yspeed = 20;
		}
		else if (hit->ydir == -1)
		{
			ob->yspeed = -20;
		}
		ExplodeShot(hit);
	}
}

void SkypestStandContact(objtype *ob, objtype *hit)
{
	if (hit->state == &s_keenpogo1 || hit->state == &s_keenpogo2 || hit->state == &s_keenpogo3)
	{
		ChangeState(ob, &s_skypestsquished);
		SD_PlaySound(SND_SQUISH);
		ob->obclass = decoobj;
	}
}

void SkypestStandThink(objtype *ob)
{
	ob->ydir = -1;
	ob->yspeed = -16;
	tmp_ymove = -144;
}

void SkypestFlyReact(objtype *ob)
{
	if (ob->hitsouth)
	{
		ob->yspeed = 8;
		ob->ydir = 1;
	}
	if (ob->hitnorth && !ob->hiteast && !ob->hitwest)
	{
		ob->y += 8*PIXGLOBAL;
		ChangeState(ob, &s_skypeststand1);
	}
	if (ob->hitwest)
	{
		ob->xspeed = 0;
		ob->xdir = -1;
	}
	if (ob->hiteast)
	{
		ob->xspeed = 0;
		ob->xdir = 1;
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}
