/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
K4_ACT2.C
=========

Contains the following actor types (in this order):

- Wormouth
- Thundercloud & Lightning
- Berkeloid & Fireball
- Inchworm & Foot (in-level)
- Bounder
- Lick
- Platform
- Loose (Falling) Platform

*/

#include "CK_DEF.H"

/*
=============================================================================
	WORMOUTH
=============================================================================
*/

statetype s_wormouth        = {WORMOUTHSPR,         WORMOUTHSPR,         slide, true,  push_down,  4, 16, 0, WormouthThink,          NULL,                  WalkReact,    &s_wormouth};
statetype s_wormouthlook1   = {WORMOUTHLOOK1LSPR,   WORMOUTHLOOK1LSPR,   step,  false, push_none, 20,  0, 0, NULL,                   WormouthLookContact,   DrawReact,    &s_wormouthlook2};
statetype s_wormouthlook2   = {WORMOUTHLOOK2LSPR,   WORMOUTHLOOK2LSPR,   step,  false, push_none,  8,  0, 0, NULL,                   WormouthLookContact,   DrawReact,    &s_wormouthlook3};
statetype s_wormouthlook3   = {WORMOUTHLOOK1LSPR,   WORMOUTHLOOK1LSPR,   step,  false, push_none, 20,  0, 0, WormouthLookLeftThink,  WormouthLookContact,   DrawReact,    &s_wormouthlook4};
statetype s_wormouthlook4   = {WORMOUTHSPR,         WORMOUTHSPR,         step,  false, push_none,  8,  0, 0, NULL,                   WormouthLookContact,   DrawReact,    &s_wormouthlook5};
statetype s_wormouthlook5   = {WORMOUTHLOOK1RSPR,   WORMOUTHLOOK1RSPR,   step,  false, push_none, 20,  0, 0, NULL,                   WormouthLookContact,   DrawReact,    &s_wormouthlook6};
statetype s_wormouthlook6   = {WORMOUTHLOOK2RSPR,   WORMOUTHLOOK2RSPR,   step,  false, push_none,  8,  0, 0, NULL,                   WormouthLookContact,   DrawReact,    &s_wormouthlook7};
statetype s_wormouthlook7   = {WORMOUTHLOOK1RSPR,   WORMOUTHLOOK1RSPR,   step,  false, push_none, 20,  0, 0, WormouthLookRightThink, WormouthLookContact,   DrawReact,    &s_wormouthlook8};
statetype s_wormouthlook8   = {WORMOUTHSPR,         WORMOUTHSPR,         step,  false, push_none,  8,  0, 0, WormouthLookThink,      NULL,                  DrawReact,    &s_wormouth};
statetype s_wormouthattack1 = {WORMOUTHATTACK1LSPR, WORMOUTHATTACK1RSPR, step,  false, push_none,  8,  0, 0, NULL,                   WormouthAttackContact, DrawReact,    &s_wormouthattack2};
statetype s_wormouthattack2 = {WORMOUTHATTACK2LSPR, WORMOUTHATTACK2RSPR, step,  false, push_none,  8,  0, 0, NULL,                   WormouthAttackContact, DrawReact,    &s_wormouthattack3};
statetype s_wormouthattack3 = {WORMOUTHATTACK3LSPR, WORMOUTHATTACK3RSPR, step,  false, push_none, 16,  0, 0, NULL,                   WormouthAttackContact, DrawReact,    &s_wormouthattack4};
statetype s_wormouthattack4 = {WORMOUTHATTACK2LSPR, WORMOUTHATTACK2RSPR, step,  false, push_none,  8,  0, 0, NULL,                   WormouthAttackContact, DrawReact,    &s_wormouthattack5};
statetype s_wormouthattack5 = {WORMOUTHATTACK1LSPR, WORMOUTHATTACK1RSPR, step,  false, push_none,  8,  0, 0, NULL,                   WormouthAttackContact, DrawReact,    &s_wormouth};
statetype s_wormouthstunned = {WORMOUTHSTUNNEDSPR,  WORMOUTHSTUNNEDSPR,  think, false, push_none, 10,  0, 0, ProjectileThink4,       NULL,                  StunnedReact, NULL};

void SpawnWormouth(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = wormouthobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) + 0x8F;
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = -1;
	NewState(new, &s_wormouth);
}

void WormouthLookRightThink(objtype *ob)
{
	if (player->x > ob->x)
	{
		ob->xdir = 1;
		ob->state = &s_wormouth;
	}
}

void WormouthLookThink(objtype *ob)
{
	if (player->x > ob->x)
	{
		ob->xdir = 1;
	}
	else
	{
		ob->xdir = -1;
	}
}

void WormouthLookLeftThink(objtype *ob)
{
	if (player->x < ob->x)
	{
		ob->xdir = -1;
		ob->state = &s_wormouth;
	}
}

void WormouthThink(objtype *ob)
{
	Sint16 xdist, ydist;

	xdist = player->x - ob->x;
	ydist = player->bottom - ob->bottom;
	if ((xdist < -3*TILEGLOBAL || xdist > 3*TILEGLOBAL) && US_RndT() < 6)
	{
		ob->state = &s_wormouthlook1;
	}
	else if (ydist >= -TILEGLOBAL && ydist <= TILEGLOBAL)
	{
		if (ob->xdir == 1 && xdist > 8*PIXGLOBAL && xdist < 24*PIXGLOBAL
			|| ob->xdir == -1 && xdist < -8*PIXGLOBAL && xdist > -32*PIXGLOBAL)
		{
			SD_PlaySound(SND_WORMOUTHATTACK);
			ob->state = &s_wormouthattack1;
		}
	}
}

void WormouthLookContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		StunContact(ob, hit, &s_wormouthstunned);
	}
}

void WormouthAttackContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
	{
		if (ob->xdir == 1 && ob->x > hit->x || ob->xdir == -1 && ob->x < hit->x)
			return;

		KillKeen();
	}
	else
	{
		WormouthLookContact(ob, hit);
	}
}

/*
=============================================================================
	THUNDERCLOUD
=============================================================================
*/

statetype s_thundercloud         = {CLOUDSPR, CLOUDSPR, think,     false, push_none,  20, 0, 0, NULL,                    ThundercloudContact, DrawReact,         NULL};
statetype s_thundercloudactive1  = {CLOUDACTIVESPR, CLOUDACTIVESPR, step,      false, push_none, 100, 0, 0, NULL,                    NULL,                ThundercloudReact, &s_thundercloudactive2};
statetype s_thundercloudactive2  = {CLOUDACTIVESPR, CLOUDACTIVESPR, think,     false, push_none,  20, 0, 0, ThundercloudThink,       NULL,                ThundercloudReact, NULL};
statetype s_thundercloudfloat1   = {CLOUDACTIVESPR, CLOUDACTIVESPR, think,     false, push_none,  20, 0, 0, ThundercloudFloatThink,  NULL,                ThundercloudReact, NULL};
statetype s_thundercloudattack10 = {CLOUDACTIVESPR, CLOUDACTIVESPR, stepthink, false, push_none,  60, 0, 0, VelocityThink,           NULL,                ThundercloudReact, &s_thundercloudactive2};
statetype s_thundercloudattack1  = {CLOUDCHARGESPR, CLOUDCHARGESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack2};
statetype s_thundercloudattack2  = {CLOUDACTIVESPR, CLOUDACTIVESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack3};
statetype s_thundercloudattack3  = {CLOUDCHARGESPR, CLOUDCHARGESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack4};
statetype s_thundercloudattack4  = {CLOUDACTIVESPR, CLOUDACTIVESPR, step,      false, push_none,  10, 0, 0, ThundercloudAttackThink, NULL,                DrawReact,         &s_thundercloudattack5};
statetype s_thundercloudattack5  = {CLOUDCHARGESPR, CLOUDCHARGESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack6};
statetype s_thundercloudattack6  = {CLOUDACTIVESPR, CLOUDACTIVESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack7};
statetype s_thundercloudattack7  = {CLOUDCHARGESPR, CLOUDCHARGESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack8};
statetype s_thundercloudattack8  = {CLOUDACTIVESPR, CLOUDACTIVESPR, step,      false, push_none,  10, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack9};
statetype s_thundercloudattack9  = {CLOUDCHARGESPR, CLOUDCHARGESPR, step,      false, push_none,  48, 0, 0, NULL,                    NULL,                DrawReact,         &s_thundercloudattack10};
statetype s_lightning1    = {LIGHTNING1SPR, LIGHTNING1SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, &s_lightning2};
statetype s_lightning2    = {LIGHTNING2SPR, LIGHTNING2SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, &s_lightning3};
statetype s_lightning3    = {LIGHTNING1SPR, LIGHTNING1SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, &s_lightning4};
statetype s_lightning4    = {LIGHTNING2SPR, LIGHTNING2SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, &s_lightning5};
statetype s_lightning5    = {LIGHTNING1SPR, LIGHTNING1SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, &s_lightning6};
statetype s_lightning6    = {LIGHTNING2SPR, LIGHTNING2SPR, step, false, push_none, 8, 0, 0, NULL, HazardContact, DrawReact, NULL};

void SpawnThundercloud(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = thundercloudobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	new->ydir = new->xdir = 1;
	NewState(new, &s_thundercloud);
}

void ThundercloudThink(objtype *ob)
{
	if (US_RndT() < tics)
		ob->xdir = ob->x > player->x? -1 : 1;

	AccelerateX(ob, ob->xdir, 10);
	if (player->bottom >= ob->bottom && (Sint16)(player->top - ob->bottom) <= 64*PIXGLOBAL
		&& ob->left < player->right && ob->right > player->left)
	{
		ob->state = &s_thundercloudfloat1;
	}
}

void ThundercloudFloatThink(objtype *ob)
{
	AccelerateX(ob, ob->xdir, 10);
	if (tmp_xmove < 0 && (Sint16)(ob->x & 0x7F) + tmp_xmove <= 0)
	{
		tmp_xmove = -(ob->x & 0x7F);
		ob->state = &s_thundercloudattack1;
	}
	if (tmp_xmove > 0 && (ob->x & 0x7F) + tmp_xmove >= 8*PIXGLOBAL)
	{
		tmp_xmove = 8*PIXGLOBAL - (ob->x & 0x7F);
		ob->state = &s_thundercloudattack1;
	}
}

void ThundercloudReact(objtype *ob)
{
	if (ob->hitwest)
	{
		ob->xspeed = 0;
		ob->xdir = -1;
	}
	else if (ob->hiteast)
	{
		ob->xspeed = 0;
		ob->xdir = 1;
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void ThundercloudAttackThink(objtype *ob)
{
	GetNewObj(true);
	new->obclass = lightningobj;
	new->active = removable;
	new->needtoclip = false;
	new->x = ob->x + TILEGLOBAL;
	new->y = ob->y + TILEGLOBAL;
	NewState(new, &s_lightning1);
	SD_PlaySound(SND_THUNDER);
}

void ThundercloudContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
	{
		ChangeState(ob, &s_thundercloudactive1);
	}
}

/*
=============================================================================
	BERKELOID
=============================================================================
*/

statetype s_berkeloidwalk1 = {BERKELOIDWALKL1SPR, BERKELOIDWALKR1SPR, slide, false, push_down, 6, 8, 0, BerkeloidWalkThink, BerkeloidContact, BerkeloidWalkReact, &s_berkeloidwalk2};
statetype s_berkeloidwalk2 = {BERKELOIDWALKL2SPR, BERKELOIDWALKR2SPR, slide, false, push_down, 6, 8, 0, BerkeloidWalkThink, BerkeloidContact, BerkeloidWalkReact, &s_berkeloidwalk3};
statetype s_berkeloidwalk3 = {BERKELOIDWALKL3SPR, BERKELOIDWALKR3SPR, slide, false, push_down, 6, 8, 0, BerkeloidWalkThink, BerkeloidContact, BerkeloidWalkReact, &s_berkeloidwalk4};
statetype s_berkeloidwalk4 = {BERKELOIDWALKL4SPR, BERKELOIDWALKR4SPR, slide, false, push_down, 6, 8, 0, BerkeloidWalkThink, BerkeloidContact, BerkeloidWalkReact, &s_berkeloidwalk1};
statetype s_berkeloidattack1 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack2};
statetype s_berkeloidattack2 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack3};
statetype s_berkeloidattack3 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack4};
statetype s_berkeloidattack4 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack5};
statetype s_berkeloidattack5 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack6};
statetype s_berkeloidattack6 = {BERKELOIDATTACKL1SPR, BERKELOIDATTACKR1SPR, step, false, push_none, 6, 0, 0, BerkeloidAttackThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack7};
statetype s_berkeloidattack7 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack8};
statetype s_berkeloidattack8 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack9};
statetype s_berkeloidattack9 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack10};
statetype s_berkeloidattack10 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack11};
statetype s_berkeloidattack11 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, NeedToReactThink, BerkeloidContact, BerkeloidReact, &s_berkeloidattack12};
statetype s_berkeloidattack12 = {BERKELOIDATTACKL2SPR, BERKELOIDATTACKR2SPR, step, false, push_none, 6, 0, 0, BerkeloidAfterAttackThink, BerkeloidContact, BerkeloidReact, &s_berkeloidwalk1};
statetype s_fireball1     = {FIREBALL1SPR, FIREBALL1SPR, stepthink, false, push_none, 6, 0, 0, ProjectileThink3, HazardContact, FireballReact, &s_fireball2};
statetype s_fireball2     = {FIREBALL2SPR, FIREBALL2SPR, stepthink, false, push_none, 6, 0, 0, ProjectileThink3, HazardContact, FireballReact, &s_fireball1};
statetype s_fireball3     = {FIREBALL1SPR, FIREBALL1SPR, step, false, push_none, 6, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball4};
statetype s_fireball4     = {FIREBALL3SPR, FIREBALL3SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball5};
statetype s_fireball5     = {FIREBALL4SPR, FIREBALL4SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball6};
statetype s_fireball6     = {FIREBALL3SPR, FIREBALL3SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball7};
statetype s_fireball7     = {FIREBALL4SPR, FIREBALL4SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball8};
statetype s_fireball8     = {FIREBALL1SPR, FIREBALL1SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball9};
statetype s_fireball9     = {FIREBALL2SPR, FIREBALL2SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball10};
statetype s_fireball10    = {FIREBALL1SPR, FIREBALL1SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, &s_fireball11};
statetype s_fireball11    = {FIREBALL2SPR, FIREBALL2SPR, step, false, push_none, 12, 0, 0, NULL, BerkeloidContact, DrawReact, NULL};

void SpawnBerkeloid(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = berkeloidobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -2*TILEGLOBAL;
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	new->temp2 = 8;
	NewState(new, &s_berkeloidwalk1);
}


void BerkeloidWalkThink(objtype *ob)
{
	Sint16 xdist, ydist;

	if (US_RndT() < 0x20)
		ob->xdir = player->x < ob->x? -1 : 1;

	if (US_RndT() < 8)
	{
		ChangeState(ob, &s_berkeloidattack1);
	}
	else if (US_RndT() <= 0x40)
	{
		xdist = player->x - ob->x;
		ydist = player->y - ob->y;
		if (ydist >= -TILEGLOBAL && ydist <= TILEGLOBAL
			&& (ob->xdir == 1 && xdist > 0 || ob->xdir == -1 && xdist < 0))
		{
			ob->state = &s_berkeloidattack1;
			SD_PlaySound(SND_BERKELOIDATTACK);
		}
	}
}

void BerkeloidAttackThink(objtype *ob)
{
	GetNewObj(true);
	new->active = removable;
	new->obclass = berkeloidobj;
	new->y = ob->y + 8*PIXGLOBAL;
	new->yspeed = -8;
	if (ob->xdir == 1)
	{
		new->xspeed = 48;
		new->x = ob->x + 32*PIXGLOBAL;
		new->xdir = 1;
	}
	else
	{
		new->xspeed = -48;
		new->x = ob->x - 16*PIXGLOBAL;
		new->xdir = -1;
	}
	NewState(new, &s_fireball1);
	ob->needtoreact = true;
}

void BerkeloidAfterAttackThink(objtype *ob)
{
	ob->nothink = 4;
	ob->needtoreact = true;
}

void BerkeloidContact(objtype *ob, objtype *hit)
{
	ob++;
	if (hit->obclass == shotobj)
	{
		ExplodeShot(hit);
	}
	else if (hit->obclass == keenobj)
	{
		KillKeen();
	}
}

void FireballReact(objtype *ob)
{
	if (ob->hiteast || ob->hitwest)
		ob->xspeed = 0;

	if (ob->hitnorth)
	{
		SD_PlaySound(SND_FIREBALLLAND);
		ChangeState(ob, &s_fireball3);
	}
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

void BerkeloidReact(objtype *ob)
{
	//float up & down:
	ob->temp1 = ob->temp1 + ob->temp2 * tics;
	if (ob->temp1 > 0)
	{
		ob->temp1 = 0;
		ob->temp2 = -8;
	}
	else if (ob->temp1 < -TILEGLOBAL)
	{
		ob->temp1 = -TILEGLOBAL;
		ob->temp2 = 8;
	}

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y+ob->temp1, ob->shapenum, 0, 0);
}

void BerkeloidWalkReact(objtype *ob)
{
	if (ob->xdir == 1 && ob->hitwest)
	{
		ob->x -= ob->xmove;
		ob->xdir = -1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (ob->xdir == -1 && ob->hiteast)
	{
		ob->x -= ob->xmove;
		ob->xdir = 1;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	else if (!ob->hitnorth)
	{
		ob->x -= ob->xmove*2;
		ob->xdir = -ob->xdir;
		ob->nothink = US_RndT() / 32;
		ChangeState(ob, ob->state);
	}
	BerkeloidReact(ob);
}

/*
=============================================================================
	INCHWORM (& FOOT)
=============================================================================
*/

statetype s_smoke1        = {SMOKE1SPR, SMOKE1SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, &s_smoke2};
statetype s_smoke2        = {SMOKE2SPR, SMOKE2SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, &s_smoke3};
statetype s_smoke3        = {SMOKE3SPR, SMOKE3SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, &s_smoke4};
statetype s_smoke4        = {SMOKE4SPR, SMOKE4SPR, step, false, push_none, 12, 0, 0, NULL, NULL, DrawReact, NULL};
statetype s_inchworm1     = {INCHWORML1SPR, INCHWORMR1SPR, step, false, push_down, 30, 128, 0, InchwormThink, InchwormContact, WalkReact, &s_inchworm2};
statetype s_inchworm2     = {INCHWORML2SPR, INCHWORMR2SPR, step, false, push_down, 30, 128, 0, InchwormThink, InchwormContact, WalkReact, &s_inchworm1};
statetype s_footappear    = { -1,  -1, step, false, push_none, 48, 0, 0, NULL, NULL, DrawReact, &s_afoot};	//never used!
statetype s_afoot         = {FOOTSPR, FOOTSPR, think, false, push_none, 30000, 0, 0, NULL, FootContact, DrawReact, NULL};

void SpawnInchworm(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = inchwormobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_inchworm1);
	new->ticcount = US_RndT() / 32;
}

void SpawnFoot(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = footobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y-3);
	NewState(new, &s_afoot);
}

void InchwormThink(objtype *ob)
{
	if (ob->x > player->x)
	{
		ob->xdir = -1;
	}
	else
	{
		ob->xdir = 1;
	}
}

void InchwormContact(objtype *ob, objtype *hit)
{
	objtype *ob2;

	if (hit->obclass != inchwormobj)
		return;

	if (ob->temp1 != (Sint16)lasttimecount)
	{
		ob->temp1 = (Sint16)lasttimecount;
		ob->temp2 = 0;
	}
	ob->temp2++;
	if (ob->temp2 != 11)	//11 instead of 12 since the object can't contact itself
		return;

	//change current inchworm into a foot:
	SD_PlaySound(SND_MAKEFOOT);
	ob->y -= 5*TILEGLOBAL;
	ob->obclass = footobj;
	ChangeState(ob, &s_afoot);

	//Note: It would make more sense to remove the remaining inchworm BEFORE
	//spawning the smoke, just in case there are not enough free spots in the
	//objarray to spawn the smoke. The game won't crash either way, though.

	//spawn smoke:
	GetNewObj(true);
	new->x = ob->x + -8*PIXGLOBAL;
	new->y = ob->y + 16*PIXGLOBAL;
	new->priority = 3;
	NewState(new, &s_smoke1);

	GetNewObj(true);
	new->x = ob->x + 16*PIXGLOBAL;
	new->y = ob->y + 24*PIXGLOBAL;
	new->priority = 3;
	NewState(new, &s_smoke1);

	GetNewObj(true);
	new->x = ob->x + 40*PIXGLOBAL;
	new->y = ob->y + 16*PIXGLOBAL;
	new->priority = 3;
	NewState(new, &s_smoke1);

	GetNewObj(true);
	new->x = ob->x;
	new->y = ob->y - 8*PIXGLOBAL;
	new->priority = 3;
	NewState(new, &s_smoke1);

	//remove ALL inchworm from the level:
	for (ob2 = player->next; ob2; ob2=ob2->next)
	{
		if (ob2->obclass == inchwormobj)
			RemoveObj(ob2);
	}
}

void FootContact(objtype *ob, objtype *hit)	//completely useless
{
	ob++;
	hit++;
}

/*
=============================================================================
	BOUNDER
=============================================================================
*/

statetype s_bounderc1     = {BOUNDER1CSPR, BOUNDER1CSPR, stepthink, false, push_none, 20, 0, 0, ProjectileThink4, BounderContact, BounderReact, &s_bounderc2};
statetype s_bounderc2     = {BOUNDER2CSPR, BOUNDER2CSPR, stepthink, false, push_none, 20, 0, 0, ProjectileThink4, BounderContact, BounderReact, &s_bounderc1};
statetype s_bounder1      = {BOUNDER1LSPR, BOUNDER1RSPR, stepthink, false, push_none, 20, 0, 0, ProjectileThink4, BounderContact, BounderReact, &s_bounder2};
statetype s_bounder2      = {BOUNDER2LSPR, BOUNDER2RSPR, stepthink, false, push_none, 20, 0, 0, ProjectileThink4, BounderContact, BounderReact, &s_bounder1};
statetype s_bounderstunned0 = {BOUNDER1CSPR, BOUNDER1CSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, &s_bounderstunned1};
statetype s_bounderstunned1 = {BOUNDERSTUNNEDSPR, BOUNDERSTUNNEDSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};

void SpawnBounder(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = bounderobj;
	new->active = yes;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) + -8*PIXGLOBAL;
	new->ydir = 1;
	new->xdir = 0;
	NewState(new, &s_bounderc1);
}

void BounderContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		//basically StunContact(), but in different order:
		ob->temp1 = 0;
		ob->temp2 = 0;
		ob->temp3 = 0;
		ob->temp4 = ob->obclass;
		ExplodeShot(hit);
		ChangeState(ob, &s_bounderstunned0);
		ob->obclass = stunnedobj;

		ob->yspeed -= 32;
	}
}

void BounderReact(objtype *ob)
{
	Sint16 randnum;

	if (ob->hitsouth)
		ob->yspeed = 0;

	if (ob->hitnorth)
	{
		ob->temp2++;
		if (ObjectVisible(ob))
			SD_PlaySound(SND_BOUNCE2);

		ob->yspeed = -50;
		if (gamestate.platform == ob)
		{
			ob->temp2 = 0;
			if (player->left < ob->left-4*PIXGLOBAL)
			{
				ob->xdir = -1;
			}
			else if (player->right > ob->right+4*PIXGLOBAL)
			{
				ob->xdir = 1;
			}
			else
			{
				ob->xdir = 0;
			}
			ob->xspeed = ob->xdir * 24;
		}
		else if (ob->temp2 <= 2 || ob->temp1 == 0)
		{
			ob->temp1 = 1;
			ob->xdir = ob->xspeed = 0;
			ChangeState(ob, &s_bounderc1);
		}
		else
		{
			ob->temp1 = 0;
			randnum = US_RndT();
			if (randnum < 100)
			{
				ob->xdir = -1;
			}
			else if (randnum < 200)
			{
				ob->xdir = 1;
			}
			else
			{
				ob->xdir = 0;
			}
			ob->xspeed = ob->xdir * 24;
		}

		if (ob->xdir)
		{
			ChangeState(ob, &s_bounder1);
		}
		else
		{
			ChangeState(ob, &s_bounderc1);
		}
	}

	if (ob->hiteast || ob->hitwest)
	{
		ob->xdir = -ob->xdir;
		ob->xspeed = -ob->xspeed;
	}

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

/*
=============================================================================
	LICK
=============================================================================
*/

statetype s_lick1         = {LICKMOVE1LSPR, LICKMOVE1RSPR, step, false, push_none, 10, 0, 0, LickThink, LickContact, DrawReact, &s_lick2};
statetype s_lick2         = {LICKMOVE2LSPR, LICKMOVE2RSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, LickContact, LickReact, &s_lick3};
statetype s_lick3         = {LICKMOVE3LSPR, LICKMOVE3RSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, LickContact, LickReact, NULL};
statetype s_lickground    = {LICKMOVE4LSPR, LICKMOVE4RSPR, step, false, push_none, 10, 0, 0, NULL, LickContact, DrawReact, &s_lick1};
statetype s_lickattack1   = {LICKATTACK1LSPR, LICKATTACK1RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack2};
statetype s_lickattack2   = {LICKATTACK2LSPR, LICKATTACK2RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack3};
statetype s_lickattack3   = {LICKATTACK3LSPR, LICKATTACK3RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack4};
statetype s_lickattack4   = {LICKATTACK2LSPR, LICKATTACK2RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack5};
statetype s_lickattack5   = {LICKATTACK1LSPR, LICKATTACK1RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack6};
statetype s_lickattack6   = {LICKATTACK2LSPR, LICKATTACK2RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack7};
statetype s_lickattack7   = {LICKATTACK3LSPR, LICKATTACK3RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lickattack8};
statetype s_lickattack8   = {LICKATTACK2LSPR, LICKATTACK2RSPR, step, true, push_none, 4, 0, 0, NULL, LickAttackContact, DrawReact, &s_lick3};
statetype s_lickstunned1  = {LICKSTUNNEDSPR, LICKSTUNNEDSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, &s_lickstunned2};
statetype s_lickstunned2  = {LICKSTUNNEDSPR, LICKSTUNNEDSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};

void SpawnLick(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = lickobj;
	new->active = yes;
	new->priority = 2;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	new->nothink = US_RndT() / 64;
	NewState(new, &s_lick3);
}

void LickThink(objtype *ob)
{
	Sint16 xdist, ydist;

	if (ob->x > player->x)
	{
		ob->xdir = -1;
	}
	else
	{
		ob->xdir = 1;
	}

	xdist = player->x - ob->x;
	ydist = player->y - ob->y;

	if (ydist >= -TILEGLOBAL && ydist <= TILEGLOBAL	&& 
		( ob->xdir == 1 && xdist > -2*PIXGLOBAL && xdist < 24*PIXGLOBAL
		 || ob->xdir == -1 && xdist < 2*PIXGLOBAL && xdist > -24*PIXGLOBAL ) )
	{
		SD_PlaySound(SND_LICKATTACK);
		ChangeState(ob, &s_lickattack1);
	}
	else if (abs(xdist) > 3*TILEGLOBAL)
	{
		ob->xspeed = ob->xdir * 32;
		ob->yspeed = -32;
	}
	else
	{
		ob->xspeed = (ob->xdir * 32)/2;
		ob->yspeed = -16;
	}
}

void LickContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == shotobj)
	{
		StunContact(ob, hit, &s_lickstunned1);
		ob->yspeed -= 16;
	}
}

void LickAttackContact(objtype *ob, objtype *hit)
{
	if (hit->obclass == keenobj)
	{
		if (ob->xdir == 1 && player->x > ob->x 
			|| ob->xdir == -1 && player->x < ob->x)
		{
			KillKeen();
		}
	}
	else
	{
		LickContact(ob, hit);
	}
}

void LickReact(objtype *ob)
{
	if (ob->hitnorth)
		ChangeState(ob, &s_lickground);

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}

/*
=============================================================================
	PLATFORM
=============================================================================
*/

statetype s_platform      = {PLATFORMSPR, PLATFORMSPR, think, false, push_none, 0, 0, 0, PlatformThink, NULL, PlatformReact, NULL};

void SpawnPlatform(Sint16 x, Sint16 y, Sint16 dir)
{
	GetNewObj(false);
	new->obclass = platformobj;
	new->active = allways;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	switch (dir)
	{
	case 0:
		new->xdir = 0;
		new->ydir = -1;
		break;
	case 1:
		new->xdir = 1;
		new->ydir = 0;
		break;
	case 2:
		new->xdir = 0;
		new->ydir = 1;
		break;
	case 3:
		new->xdir = -1;
		new->ydir = 0;
		break;
	}
	NewState(new, &s_platform);
}

void PlatformThink(objtype *ob)
{
	Uint16 newpos, newtile;

	tmp_xmove = ob->xdir * 12 * tics;
	tmp_ymove = ob->ydir * 12 * tics;

	if (ob->xdir == 1)
	{
		newpos = ob->right + tmp_xmove;
		newtile = CONVERT_GLOBAL_TO_TILE(newpos);
		if (ob->tileright != newtile)
		{
			if (mapsegs[2][mapbwidthtable[ob->tiletop]/2+newtile] == 31)
			{
				ob->xdir = -1;
				tmp_xmove = tmp_xmove - (newpos & 0xFF);
			}
		}
	}
	else if (ob->xdir == -1)
	{
		newpos = ob->left + tmp_xmove;
		newtile = CONVERT_GLOBAL_TO_TILE(newpos);
		if (ob->tileleft != newtile)
		{
			if (mapsegs[2][mapbwidthtable[ob->tiletop]/2+newtile] == 31)
			{
				ob->xdir = 1;
				tmp_xmove = tmp_xmove + 0x100 - (newpos & 0xFF);
			}
		}
	}
	else if (ob->ydir == 1)
	{
		newpos = ob->bottom + tmp_ymove;
		newtile = CONVERT_GLOBAL_TO_TILE(newpos);
		if (ob->tilebottom != newtile)
		{
			if (mapsegs[2][mapbwidthtable[newtile]/2+ob->tileleft] == 31)
			{
				if (mapsegs[2][mapbwidthtable[newtile-2]/2+ob->tileleft] == 31)
				{
					tmp_ymove = 0;
					ob->needtoreact = true;
				}
				else
				{
					ob->ydir = -1;
					tmp_ymove = tmp_ymove - (newpos & 0xFF);
				}
			}
		}
	}
	else if (ob->ydir == -1)
	{
		newpos = ob->top + tmp_ymove;
		newtile = CONVERT_GLOBAL_TO_TILE(newpos);
		if (ob->tiletop != newtile)
		{
			if (mapsegs[2][mapbwidthtable[newtile]/2+ob->tileleft] == 31)
			{
				if (mapsegs[2][mapbwidthtable[newtile+2]/2+ob->tileleft] == 31)
				{
					tmp_ymove = 0;
					ob->needtoreact = true;
				}
				else
				{
					ob->ydir = 1;
					tmp_ymove = tmp_ymove + 0x100 - (newpos & 0xFF);
				}
			}
		}
	}
}

void PlatformReact(objtype *ob)
{
	Uint16 frame;

	//place platform sprite:
	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
	
	//place (or remove) thruster sprites:
	frame = (lasttimecount / 4) & 1;
	if (ob->xdir == 1)
	{
		RF_PlaceSprite((void**)&ob->temp2, ob->x-1*PIXGLOBAL, ob->y+3*PIXGLOBAL, frame+PLATSIDETHRUST1SPR, 0, 0);
		if (ob->temp3)
			RF_RemoveSprite((void**)&ob->temp3);
	}
	else if (ob->xdir == -1)
	{
		if (ob->temp2)
			RF_RemoveSprite((void**)&ob->temp2);
		RF_PlaceSprite((void**)&ob->temp3, ob->x+48*PIXGLOBAL, ob->y+5*PIXGLOBAL, frame+PLATSIDETHRUST1SPR, 0, 1);
	}
	else if (ob->ydir == -1)
	{
		RF_PlaceSprite((void**)&ob->temp2, ob->x+2*PIXGLOBAL, ob->y+9*PIXGLOBAL, frame+PLATLTHRUST1SPR, 0, 0);
		RF_PlaceSprite((void**)&ob->temp3, ob->x+46*PIXGLOBAL, ob->y+8*PIXGLOBAL, frame+PLATRTHRUST1SPR, 0, 0);
	}
	else if (ob->ydir == 1)
	{
		if (frame)
		{
			RF_PlaceSprite((void**)&ob->temp2, ob->x+2*PIXGLOBAL, ob->y+9*PIXGLOBAL, frame+PLATLTHRUST1SPR, 0, 0);
			RF_PlaceSprite((void**)&ob->temp3, ob->x+46*PIXGLOBAL, ob->y+8*PIXGLOBAL, frame+PLATRTHRUST1SPR, 0, 0);
		}
		else
		{
			if (ob->temp2)
				RF_RemoveSprite((void**)&ob->temp2);
			if (ob->temp3)
				RF_RemoveSprite((void**)&ob->temp3);
		}
	}
}

/*
=============================================================================
	FALLING PLATFORM
=============================================================================
*/

statetype s_looseplatform = {PLATFORMSPR, PLATFORMSPR, think, false, push_none, 0, 0, 0, LoosePlatformThink, NULL, DrawReact, NULL};
statetype s_platformsink  = {PLATFORMSPR, PLATFORMSPR, think, false, push_none, 0, 0, 0, PlatformSinkThink, NULL, DrawReact, NULL};
statetype s_platformrise  = {PLATFORMSPR, PLATFORMSPR, slidethink, false, push_none, 0, 0, -32, PlatformRiseThink, NULL, DrawReact, NULL};

void SpawnLoosePlatform(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = platformobj;
	new->active = allways;
	new->priority = 0;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = new->temp1 = CONVERT_TILE_TO_GLOBAL(y);
	new->xdir = 0;
	new->ydir = 1;
	new->needtoclip = false;
	NewState(new, &s_looseplatform);
}

void LoosePlatformThink(objtype *ob)
{
	if (gamestate.platform == ob)
	{
		tmp_ymove = tics * 16;
		ob->yspeed = 0;
		if (ob->y + tmp_ymove - ob->temp1 >= 8*PIXGLOBAL)
			ob->state = &s_platformsink;
	}
}

void PlatformSinkThink(objtype *ob)
{
	Uint16 newy, ty;

	DoGravity4(ob);

	if (tmp_ymove >= 15*PIXGLOBAL)
		tmp_ymove = 15*PIXGLOBAL;

	newy = ob->bottom + tmp_ymove;
	ty = CONVERT_GLOBAL_TO_TILE(newy);
	if (ob->tilebottom != ty)
	{
		if (mapsegs[2][mapbwidthtable[ty]/2+ob->tileleft] == 31)
		{
			tmp_ymove = 0xFF - (ob->bottom & 0xFF);
			if (gamestate.platform != ob)
				ChangeState(ob, &s_platformrise);
		}
	}
}

void PlatformRiseThink(objtype *ob)
{
	if (gamestate.platform == ob)
	{
		ob->yspeed = 0;
		ob->state = &s_platformsink;
	}
	else if (ob->y <= ob->temp1)
	{
		tmp_ymove = ob->temp1 - ob->y;
		ob->state = &s_looseplatform;
	}
}
