/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "CK_DEF.H"

/*
=============================================================================
	BADOING

	temp1 - internal state (see enum)
=============================================================================
*/

extern statetype s_badoing;
extern statetype s_badoingstunned;

enum {BD_WAIT, BD_RISE1, BD_RISE2, BD_RISE3, BD_RISE4, BD_TURN, BD_LOOK, BD_VANISH, BD_COOLDOWN};

void SpawnBadoing(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	// Note: temp1 has been set to 0 (= BD_WAIT) by GetNewObj
	new->obclass = badoingobj;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y);
	new->shapenum = BADOINGHIDDENSPR;
	NewState(new, &s_badoing);
}

void BadoingThink(objtype *ob)
{
	ob->ticcount += tics;

	switch (ob->temp1)
	{
	case BD_WAIT:
		if (abs(player->x - ob->x) < (3*TILEGLOBAL + 8*PIXGLOBAL)
			&& player->bottom < ob->bottom+TILEGLOBAL
			&& player->bottom > ob->bottom-40*PIXGLOBAL)
		{
			SD_PlaySound(SND_BADOINGPOPUP);
			ob->shapenum = BADOINGRISE1SPR;
			ob->temp1++;
			ob->ticcount = 0;
		}
		break;

	case BD_RISE1:
	case BD_RISE2:
	case BD_RISE3:
		if (ob->ticcount >= 10)
		{
			ob->shapenum++;
			ob->temp1++;
			ob->ticcount = 0;
		}
		break;

	case BD_RISE4:
		if (ob->ticcount >= TickBase/2)	// half a second
		{
			//ob->xdir = player->x < ob->x? motion_Left : motion_Right;
			//ob->shapenum = BADOINGTURNLSPR + 1 + ob->xdir;
			ob->shapenum = player->x < ob->x? BADOINGTURNLSPR : BADOINGTURNRSPR;
			ob->temp1++;
			ob->ticcount = 0;
		}
		break;

	case BD_TURN:
		if (ob->ticcount >= 10)
		{
			//ob->shapenum = BADOINGLOOKLSPR + 1 + ob->xdir;
			ob->shapenum++;
			ob->temp1++;
			ob->ticcount = 0;
		}
		break;

	case BD_LOOK:
		if (ob->ticcount >= 3*TickBase)	//	3 seconds
		{
			//ob->shapenum = BADOINGTURNLSPR + 1 + ob->xdir;
			ob->shapenum--;
			ob->temp1++;
			ob->ticcount = 0;
			//SD_PlaySound(SND_BADOINGPOPUP);
			SD_PlaySound(SND_SHOWSTATUS);
		}
		break;

	case BD_VANISH:
		if (ob->ticcount >= 10)
		{
			if (ob->shapenum == BADOINGTURNRSPR)
				ob->shapenum = BADOINGRISE4SPR;
			else
				ob->shapenum--;

			if (ob->shapenum < BADOINGRISE1SPR)
			{
				ob->shapenum = BADOINGHIDDENSPR;
				ob->temp1++;
			}
			ob->ticcount = 0;
		}
		break;

	case BD_COOLDOWN:
		if (abs(player->x - ob->x) > 4*TILEGLOBAL)
		{
			ob->temp1 = BD_WAIT;
		}
		break;
	}
}

void BadoingContact(objtype *ob, objtype *hit)
{
	if (ob->shapenum == BADOINGHIDDENSPR)
		return;		// do nothing while hiding

	switch (hit->obclass)
	{
	case keenobj:
		KillKeen();
		break;

	case shotobj:
		StunContact(ob, hit, &s_badoingstunned);
		ob->yspeed = -70;
		break;
	}
}

statetype s_badoing = {0, 0, think, false, push_none, 0, 0, 0, BadoingThink, BadoingContact, DrawReact, NULL};
statetype s_badoingstunned = {BADOINGSTUNNEDSPR, BADOINGSTUNNEDSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};

/*
=============================================================================
	WITCH DOCTOR

	temp1 - internal state (see enum)
	temp2 - frame counter
=============================================================================
*/

extern statetype s_witchdoctor;
extern statetype s_witchdoctorjump;

#define WITCHDOCSTAND1SPR WITCHDOCWALKL2SPR	// just for testing...
#define WITCHDOCJUMPLSPR WITCHDOCWALKL1SPR	// just for testing...
#define WITCHDOCJUMPRSPR WITCHDOCWALKR1SPR	// just for testing...

enum {WD_STAND, WD_WALK, WD_RUN, WD_SEEK};

void SpawnWitchDoctor(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	// Note: temp1 has been set to 0 (= WD_STAND) by GetNewObj
	new->obclass = witchdocobj;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y-1);
	new->xdir = (US_RndT() & 2) - 1;	// set it to either 1 or -1
	NewState(new, &s_witchdoctorjump);
}

void WitchDoctorThink(objtype *ob)
{
	ob->ticcount += tics;

	switch (ob->temp1)
	{
	case WD_RUN:
		if (ob->ticcount >= 5)
		{
			ob->ticcount -= 5;

			// stop when no longer running towards Keen:
			if (ob->xdir > 0 && ob->midx > player->midx+3*TILEGLOBAL || ob->xdir < 0 && ob->midx < player->midx-3*TILEGLOBAL)
			{
				ob->temp2 = 2;
				ob->temp1 = WD_SEEK;
				SD_PlaySound(SND_WITCHDOCSEARCH);
				goto stand;
			}
			goto makestep;
		}
		break;

	case WD_WALK:
		if (ob->ticcount >= 10)
		{
			ob->ticcount -= 10;

			// randomly stop and seek Keen after walking at least 12 steps:
			if (ob->temp2 > 12 && US_RndT() <= 2)
			{
				ob->temp2 = 0;
				ob->temp1 = WD_SEEK;
				SD_PlaySound(SND_WITCHDOCSEARCH);
				goto stand;
			}
makestep:
			tmp_xmove = ob->xdir * 8*PIXGLOBAL;
			ob->shapenum = WITCHDOCWALKL1SPR + 1 + ob->xdir + (++ob->temp2 & 1);
		}
		break;

	case WD_SEEK:
	case WD_STAND:
		if (ob->ticcount >= 10)
		{
			ob->ticcount -= 10;
			if (ob->temp2 >= 4)
			{
				ob->temp2 = 0;

				if (ob->temp1 != WD_STAND)
				{
					ob->xdir = player->midx < ob->midx? -1 : 1;	// walk towards Keen

					// start running if Keen is at about the same ground level
					if (abs(ob->bottom - player->bottom) < 2*TILEGLOBAL)
					{
						SD_PlaySound(SND_WITCHDOCCHARGE);
						ob->temp1 = WD_RUN;
						goto stand;
					}
				}

				ob->temp1 = WD_WALK;
			}
stand:
			ob->shapenum = WITCHDOCSTAND1SPR + (++ob->temp2 & 1);
		}
	}
}

void WitchDoctorContact(objtype *ob, objtype *hit)
{
	switch (hit->obclass)
	{
	case keenobj:
		KillKeen();
		break;

	case shotobj:
		ExplodeShot(hit);
		ob->xdir = player->midx < ob->midx? -1 : 1;	// run towards Keen
		ob->temp1 = WD_RUN;
		break;
	}
}

void WitchDoctorReact(objtype *ob)
{
	if (ob->temp1 == WD_RUN && !ob->hitnorth)
	{
		ob->xspeed = ob->xdir * 48;
		ob->yspeed = -24;	// jump up a bit
		ChangeState(ob, &s_witchdoctorjump);
		goto draw;
	}

	if ((ob->xdir > 0 && ob->hitwest) || (ob->xdir < 0 && ob->hiteast) || (!ob->hitnorth))
	{
		ob->temp1 = WD_STAND;
		ob->temp2 = 0;
		ob->x -= ob->xmove;
		ob->y -= ob->ymove;
		ob->xdir = -ob->xdir;
		ob->shapenum += ob->xdir*2;	// adds 2 if old dir was -1, adds -2 if old dir was 1
		ChangeState(ob, ob->state);	// update hitbox and stuff (also resets ob->ticcount to 0)
	}

draw:
	DrawReact(ob);
}

void WitchDoctorJumpReact(objtype *ob)
{
	if (ob->hiteast || ob->hitwest)	// if hit a wall
		ob->xspeed = 0;	// stop moving left/right

	if (ob->hitsouth)	// if hit a ceiling
		ob->yspeed = 0;	// start falling

	if (ob->hitnorth)	// if hit the ground
	{
		ob->temp1 = WD_SEEK;
		ob->temp2 = 0;
		ob->shapenum = WITCHDOCSTAND1SPR;
		ChangeState(ob, &s_witchdoctor);
	}

	DrawReact(ob);
}

statetype s_witchdoctor = {0, 0, think, false, push_down, 0, 0, 0, WitchDoctorThink, WitchDoctorContact, WitchDoctorReact, NULL};
statetype s_witchdoctorjump = {WITCHDOCJUMPLSPR, WITCHDOCJUMPRSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, WitchDoctorContact, WitchDoctorJumpReact, NULL};

/*
=============================================================================
	ORACLE JANITOR (MIRAGIA REPLACEMENT)
=============================================================================
*/

void JanitorWalkThink(objtype *ob);
void JanitorContact(objtype *ob, objtype* hit);

extern statetype s_janitorwalk1;
extern statetype s_janitorwalk2;
extern statetype s_janitorstand;

statetype s_janitorwalk1 = {JANITORWALK1LSPR, JANITORWALK1RSPR, step, false, push_down, 10, 64, 0, JanitorWalkThink, JanitorContact, WalkReact, &s_janitorwalk2};
statetype s_janitorwalk2 = {JANITORWALK2LSPR, JANITORWALK2RSPR, step, false, push_down, 10, 64, 0, JanitorWalkThink, JanitorContact, WalkReact, &s_janitorwalk1};
statetype s_janitorstand = {JANITORTHINKLSPR, JANITORTHINKRSPR, step, false, push_none, 120, 128, 0, NULL, JanitorContact, DrawReact, &s_janitorwalk1};

void SpawnJanitor(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	new->obclass = janitorobj;
	//new->active = yes;	// RTTS mod: already set to yes by GetNewObj
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) -0x171;	// weird adjustment
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	new->ydir = 1;
	NewState(new, &s_janitorwalk1);
}

void JanitorWalkThink(objtype *ob)
{
	register Uint16 randnum;

	randnum = US_RndT();
	if (tics*8 > randnum)
	{
		tmp_xmove = 0;	// RTTS addition: don't move this time
		ob->state = &s_janitorstand;
	}
}

void JanitorContact(objtype *ob, objtype* hit)
{
	if (hit->obclass == keenobj && hit->hitnorth)	// touching Keen and Keen is touching the ground (i.e. not dead)
	{
		RemoveObj(ob);	// remove the janitor
		JanitorDialog();
		SpawnScore();	// forces scorebox to reset
		UpdateScorebox(scoreobj);	// updates the scorebox sprite
		RF_ForceRefresh();
	}
}

/*
=============================================================================
	STINGARU

	temp1 - internal state (see enum)
=============================================================================
*/

extern statetype s_stingaru;
extern statetype s_stingarujump;
extern statetype s_stingarustunned;

enum {
	STINGARU_WALK1, STINGARU_WALK2, STINGARU_WALK3, STINGARU_WALK4,
	STINGARU_JUMP1, STINGARU_JUMP2, STINGARU_JUMP3
};

void SpawnStingaru(Sint16 x, Sint16 y)
{
	GetNewObj(false);
	// Note: temp1 has been set to 0 (= STINGARU_WALK1) by GetNewObj
	new->obclass = stingaruobj;
	new->x = CONVERT_TILE_TO_GLOBAL(x);
	new->y = CONVERT_TILE_TO_GLOBAL(y) -15*PIXGLOBAL;
	if (US_RndT() < 0x80)
	{
		new->xdir = 1;
	}
	else
	{
		new->xdir = -1;
	}
	NewState(new, &s_stingarujump);
}

void StingaruThink(objtype *ob)
{
	Sint16 xdist, ydist;

	ob->ticcount += tics;

	switch (ob->temp1)
	{
	case STINGARU_WALK1:
	case STINGARU_WALK2:
	case STINGARU_WALK3:
	case STINGARU_WALK4:
		xdist = player->midx - ob->midx;
		ydist = player->bottom - ob->bottom;

		if (abs(ydist) <= 2*TILEGLOBAL && abs(xdist) <= 4*TILEGLOBAL)
		{
			ob->xdir = player->midx < ob->midx? -1 : 1;	// turn towards Keen
			ob->shapenum = STINGARUJUMPL1SPR + 4 + ob->xdir*4;
			ob->temp1 = STINGARU_JUMP1;
			SD_PlaySound(SND_STINGARUJUMP);
		}
		else
		{
			if (ob->ticcount >= 8)
			{
				ob->ticcount -= 8;
				tmp_xmove = ob->xdir * 8*PIXGLOBAL;

				if (ob->temp1 == STINGARU_WALK4)
				{
					ob->temp1 += STINGARU_WALK1-STINGARU_WALK4;
					ob->shapenum += STINGARUWALKL1SPR-STINGARUWALKL4SPR;
				}
				else
				{
					ob->temp1++;
					ob->shapenum++;
				}
			}
		}
		break;

	case STINGARU_JUMP1:
	case STINGARU_JUMP2:
		if (ob->ticcount >= 12)
		{
			ob->ticcount -= 12;
			ob->shapenum++;
			ob->temp1++;
			if (ob->temp1 == STINGARU_JUMP3)
			{
				ob->xspeed = 32 * ob->xdir;
				ob->yspeed = -32;
				ob->state = &s_stingarujump;
			}
		}
		break;
	}
}

void StingaruContact(objtype *ob, objtype *hit)
{
	switch (hit->obclass)
	{
	case keenobj:
		KillKeen();
		break;

	case shotobj:
		StunContact(ob, hit, &s_stingarustunned);
		ob->yspeed -= 24;
		break;
	}
}

void StingaruWalkReact(objtype *ob)
{
	if (!ob->hitnorth || ob->xdir < 0 && ob->hiteast || ob->xdir > 0 && ob->hitwest)
	{
		ob->x -= ob->xmove;
		ob->y -= ob->ymove;
		ob->xdir = -ob->xdir;
		ob->shapenum += 8*ob->xdir;
		ChangeState(ob, ob->state);
	}
	DrawReact(ob);
}

void StingaruJumpReact(objtype *ob)
{
	if (ob->hitsouth)
		ob->yspeed = 0;

	if (ob->hiteast && ob->xspeed < 0 || ob->hitwest && ob->xspeed > 0)
		ob->xspeed = 0;

	if (ob->hitnorth)
	{
		ob->temp1 = STINGARU_WALK1;
		ob->shapenum = STINGARUWALKL1SPR + 4 + ob->xdir*4;
		ob->xspeed = ob->yspeed = 0;
		ChangeState(ob, &s_stingaru);
	}
	DrawReact(ob);
}

statetype s_stingaru = {0, 0, think, false, push_down, 0, 0, 0, StingaruThink, StingaruContact, StingaruWalkReact, NULL};
statetype s_stingarujump = {STINGARUJUMPL3SPR, STINGARUJUMPR3SPR, think, false, push_none, 0, 0, 0, ProjectileThink4, StingaruContact, StingaruJumpReact, NULL};
statetype s_stingarustunned = {STINGARUSTUNNEDSPR, STINGARUSTUNNEDSPR, think, false, push_none, 0, 0, 0, ProjectileThink4, NULL, StunnedReact, NULL};
