/* Return To The Shadowlands Source Code
 * (Based On Reconstructed Commander Keen 4-6 Source Code)
 * Copyright (C) 2021 K1n9_Duk3
 *
 * The code in this file is loosely based on
 * Keen Dreams Source Code
 * Copyright (C) 2014 Javier M. Chavez
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
K4_LEVEL.C
==========

Contains (in this order):

- lump definition
- "Star Wars" crawl text
- level names & messages
- ScanInfoPlane() - for spawning the level objects and marking required sprites
- messages for Lindsey, Janitor, Oracle Members and more

- actor states & implementation for Scuba Keen

*/

#include "CK_DEF.H"

enum {
	CONTROLS_LUMP,			//  0
	KEEN_LUMP,				//  1
	BONUS100_LUMP,			//  2
	BONUS200_LUMP,			//  3
	BONUS500_LUMP,			//  4
	BONUS1000_LUMP,		//  5
	BONUS2000_LUMP,		//  6
	BONUS5000_LUMP,		//  7
	EXTRALIFE_LUMP,		//  8
	AMMO_LUMP,				//  9
	WOLRDKEEN_LUMP,		// 10
	SLUG_LUMP,				// 11
	MADMUSHROOM_LUMP,		// 12
	UNUSED1_LUMP,			// 13
	LINDSEY_LUMP,			// 14
	INCHWORM_LUMP,			// 15
	TREASUREEATER_LUMP,	// 16
	ORACLE_LUMP,			// 17
	EAGLE_LUMP,				// 18
	MIMROCK_LUMP,			// 19
	DOPEFISH_LUMP,			// 20
	SCHOOLFISH_LUMP,		// 21
	ARACHNUT_LUMP,			// 22
	SKYPEST_LUMP,			// 23
	WORMOUTH_LUMP,			// 24
	LICK_LUMP,				// 25
	PLATFORM_LUMP,			// 26
	BOUNDER_LUMP,			// 27
	THUNDERCLOUD_LUMP,	// 28
	BERKELOID_LUMP,		// 29
	KEYGEM_LUMP,			// 30
	DARTS_LUMP,				// 31
	SCUBAKEEN_LUMP,		// 32
	SPRITE_LUMP,			// 33
	MINE_LUMP,				// 34
	MOON_LUMP,				// 35
	EGG_LUMP,				// 36
	BADOING_LUMP,
	WITCHDOC_LUMP,
	JANITOR_LUMP,
	STINGARU_LUMP,
	NUMLUMPS					// 37
};

Uint16 lumpstart[NUMLUMPS] = {
	CONTROLS_LUMP_START,
	KEEN_LUMP_START,
	BONUS100_LUMP_START,
	BONUS200_LUMP_START,
	BONUS500_LUMP_START,
	BONUS1000_LUMP_START,
	BONUS2000_LUMP_START,
	BONUS5000_LUMP_START,
	EXTRALIFE_LUMP_START,
	AMMO_LUMP_START,
	WORLDKEEN_LUMP_START,
	SLUG_LUMP_START,
	MADMUSHROOM_LUMP_START,
	0,
	LINDSEY_LUMP_START,
	INCHWORM_LUMP_START,
	THIEF_LUMP_START,
	ORACLE_LUMP_START,
	EAGLE_LUMP_START,
	MIMROCK_LUMP_START,
	DOPEFISH_LUMP_START,
	SCHOOLFISH_LUMP_START,
	ARACHNUT_LUMP_START,
	SKYPEST_LUMP_START,
	WORMOUTH_LUMP_START,
	LICK_LUMP_START,
	PLATFORM_LUMP_START,
	BOUNDER_LUMP_START,
	THUNDERCLOUD_LUMP_START,
	BERKELOID_LUMP_START,
	KEYGEM_LUMP_START,
	DARTS_LUMP_START,
	SCUBAKEEN_LUMP_START,
	SPRITE_LUMP_START,
	MINE_LUMP_START,
	MOON_LUMP_START,
	EGG_LUMP_START,
	BADOING_LUMP_START,
	WITCHDOC_LUMP_START,
	JANITOR_LUMP_START,
	STINGARU_LUMP_START
};

Uint16 lumpend[NUMLUMPS] = {
	CONTROLS_LUMP_END,
	KEEN_LUMP_END,
	BONUS100_LUMP_END,
	BONUS200_LUMP_END,
	BONUS500_LUMP_END,
	BONUS1000_LUMP_END,
	BONUS2000_LUMP_END,
	BONUS5000_LUMP_END,
	EXTRALIFE_LUMP_END,
	AMMO_LUMP_END,
	WORLDKEEN_LUMP_END,
	SLUG_LUMP_END,
	MADMUSHROOM_LUMP_END,
	0,
	LINDSEY_LUMP_END,
	INCHWORM_LUMP_END,
	THIEF_LUMP_END,
	ORACLE_LUMP_END,
	EAGLE_LUMP_END,
	MIMROCK_LUMP_END,
	DOPEFISH_LUMP_END,
	SCHOOLFISH_LUMP_END,
	ARACHNUT_LUMP_END,
	SKYPEST_LUMP_END,
	WORMOUTH_LUMP_END,
	LICK_LUMP_END,
	PLATFORM_LUMP_END,
	BOUNDER_LUMP_END,
	THUNDERCLOUD_LUMP_END,
	BERKELOID_LUMP_END,
	KEYGEM_LUMP_END,
	DARTS_LUMP_END,
	SCUBAKEEN_LUMP_END,
	SPRITE_LUMP_END,
	MINE_LUMP_END,
	MOON_LUMP_END,
	EGG_LUMP_END,
	BADOING_LUMP_END,
	WITCHDOC_LUMP_END,
	JANITOR_LUMP_END,
	STINGARU_LUMP_END
};

boolean lumpneeded[NUMLUMPS];

char far crawltext[] =
        "Commander Keen\n"
        "in\n"
        "Return to the Shadowlands\n"
        "\n"
        "While working in his\n"
        "clubhouse Billy hears\n"
        "a loud noise from the\n"
        "yard and decides to\n"
        "investigate.\n"
        "\n"
        "Someone has kidnapped\n"
        "his pet Yorp, Spot!\n"
        "\n"
        "Rushing to Gnosticus IV\n"
        "Keen hopes to use\n"
        "the Oracle to find\n"
        "out who took his\n"
        "beloved pet.\n"
        "\n"
        "But the Oracle Council\n"
        "seems to have other\n"
        "plans for our hero...\n";

char far levelname00[] = "Far Shadowlands";
char far levelname01[] = "Dunesville";
char far levelname02[] = "Glacier Gate";
char far levelname03[] = "The Ruins of Sand Yego";
char far levelname04[] = "Backwoods Blockade";
char far levelname05[] = "Frigid Fortress";
char far levelname06[] = "Frozone";
char far levelname07[] = "Bounder Town";
char far levelname08[] = "Reflectalus";
char far levelname09[] = "Lost Longhouse";
char far levelname10[] = "Pyramid of Fear";
char far levelname11[] = "Pyramid of Footsteps";
char far levelname12[] = "Whispering Woods";
char far levelname13[] = "Pyramid of the Sun";
char far levelname14[] = "Isle of the Ancients";
char far levelname15[] = "Isle of Caves";
char far levelname16[] = "Isle of Ice";
char far levelname17[] = "Sunken Sinkhole";
char far levelname18[] = "Jungri-la";
char far levelname20[] = "The Bean-with-Bacon\nMegarocket";

char far levelmessage00[] = "Keen returns to the\nShadowlands";
char far levelmessage01[] = "Keen treads into\nDunesville";
char far levelmessage02[] = "Keen skates into\nGlacier Gate";
char far levelmessage03[] = "Keen may\nbe the cause of\nThe Ruins of Sand Yego";
char far levelmessage04[] = "Keen assaults the\nBackwoods Blockade";
char far levelmessage05[] = "Keen invades\nthe Frigid Fortress";
char far levelmessage06[] = "Keen chills out\nin the Frozone";
char far levelmessage07[] = "Keen bounces into\nBounder Town";
char far levelmessage08[] = "Keen sees visions\nin Reflectalus";
char far levelmessage09[] = "Keen discovers\nthe Lost Longhouse";
char far levelmessage10[] = "Keen creeps into the\nPyramid of Fear";
char far levelmessage11[] = "Keen is followed\ninto the\nPyramid of Footsteps";
char far levelmessage12[] = "Keen is enchanted by\nthe Whispering Woods";
char far levelmessage13[] = "Keen gazes into the\nPyramid of the Sun";
char far levelmessage14[] = "Keen irreverntly tours\nthe Isle of Ancients";
char far levelmessage15[] = "Keen descends\nthe Isle of Caves";
char far levelmessage16[] = "Keen slips into\nthe Isle of Ice";
char far levelmessage17[] = "Keen plops into\nthe Sunken Sinkhole";
char far levelmessage18[] = "Keen basks in\nJungri-la";
char far levelmessage20[] = "Keen launches into the\nBean-with-Bacon\nMegarocket";

char far *levelnames[GAMELEVELS] = {
	levelname00,
	levelname01,
	levelname02,
	levelname03,
	levelname04,
	levelname05,
	levelname06,
	levelname07,
	levelname08,
	levelname09,
	levelname10,
	levelname11,
	levelname12,
	levelname13,
	levelname14,
	levelname15,
	levelname16,
	levelname17,
	levelname18,
	"",		// level 19 (high scores) gets a blank name to avoid crashes
	levelname20
};

char far *levelmessages[GAMELEVELS] = {
	levelmessage00,
	levelmessage01,
	levelmessage02,
	levelmessage03,
	levelmessage04,
	levelmessage05,
	levelmessage06,
	levelmessage07,
	levelmessage08,
	levelmessage09,
	levelmessage10,
	levelmessage11,
	levelmessage12,
	levelmessage13,
	levelmessage14,
	levelmessage15,
	levelmessage16,
	levelmessage17,
	levelmessage18,
	"",		// level 19 (high scores) gets a blank message to avoid crashes
	levelmessage20
};

Uint16 bonuslumps[] = {
	KEYGEM_LUMP, KEYGEM_LUMP, KEYGEM_LUMP, KEYGEM_LUMP,
	BONUS100_LUMP, BONUS200_LUMP, BONUS500_LUMP,
	BONUS1000_LUMP, BONUS2000_LUMP, BONUS5000_LUMP,
	EXTRALIFE_LUMP, AMMO_LUMP
};

void ScanInfoPlane(void)
{
	objtype *ob;
	Uint16 i, x, y, chunk, info;
	Uint16 far *map;

	InitObjList();
	memset(lumpneeded, 0, sizeof(lumpneeded));
	map = mapsegs[2];

	for (y=0; y<mapheight; y++)
	{
		for (x=0; x<mapwidth; x++)
		{
			info = *map++;

			if (info == 0)
				continue;

			switch (info)
			{
			case 1:
				SpawnKeen(x, y, 1);
				SpawnScore();
				lumpneeded[KEEN_LUMP] = true;
				CA_MarkGrChunk(SCOREBOXSPR);
				break;

			case 2:
				SpawnKeen(x, y, -1);
				SpawnScore();
				lumpneeded[KEEN_LUMP] = true;
				CA_MarkGrChunk(SCOREBOXSPR);
				break;

			case 3:
				SpawnWorldKeen(x, y);
				SpawnScore();
				lumpneeded[WOLRDKEEN_LUMP] = true;
				CA_MarkGrChunk(SCOREBOXSPR);
				break;

			case 4:
				SpawnOracle(x, y);
				lumpneeded[ORACLE_LUMP] = true;
				break;

			case 50:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 49:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 5:
				SpawnBerkeloid(x, y);
				lumpneeded[BERKELOID_LUMP] = true;
				break;

			case 6:
				SpawnLindsey(x, y);
				lumpneeded[LINDSEY_LUMP] = true;
				break;

			case 52:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 51:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 7:
				SpawnWormouth(x, y);
				lumpneeded[WORMOUTH_LUMP] = true;
				break;

			case 46:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 45:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 8:
				SpawnSkypest(x, y);
				lumpneeded[SKYPEST_LUMP] = true;
				break;

			case 9:
				SpawnThundercloud(x, y);
				lumpneeded[THUNDERCLOUD_LUMP] = true;
				break;

			case 10:
				SpawnFoot(x, y);
				lumpneeded[INCHWORM_LUMP] = true;	// lump includes the foot sprite
				// Note: The smoke sprites aren't actually required for the foot!
				for (i=SMOKE1SPR; i<=SMOKE4SPR; i++)
				{
					CA_MarkGrChunk(i);
				}
				break;

			case 11:
				SpawnInchworm(x, y);
				lumpneeded[INCHWORM_LUMP] = true;
				for (i=SMOKE1SPR; i<=SMOKE4SPR; i++)
				{
					CA_MarkGrChunk(i);
				}
				break;

			case 12:
				SpawnBounder(x, y);
				lumpneeded[BOUNDER_LUMP] = true;
				break;

			case 13:
				SpawnEgg(x, y);
				lumpneeded[EAGLE_LUMP] = true;
				lumpneeded[EGG_LUMP] = true;
				break;

			case 48:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 47:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 14:
				SpawnLick(x, y);
				lumpneeded[LICK_LUMP] = true;
				break;

			case 88:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 87:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 15:
				SpawnDopefish(x, y);
				lumpneeded[DOPEFISH_LUMP] = true;
				break;

			case 16:
				SpawnSchoolfish(x, y);
				lumpneeded[SCHOOLFISH_LUMP] = true;
				break;

			case 24:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 23:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 17:
				SpawnSprite(x, y);
				lumpneeded[SPRITE_LUMP] = true;
				break;

			case 18:
				SpawnTreasureeater(x, y);
				lumpneeded[TREASUREEATER_LUMP] = true;
				break;

			case 19:
				SpawnMimrock(x, y);
				lumpneeded[MIMROCK_LUMP] = true;
				break;

			case 74:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 73:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 20:
				SpawnArachnut(x, y);
				lumpneeded[ARACHNUT_LUMP] = true;
				break;

			case 21:
				SpawnMadMushroom(x, y);
				lumpneeded[MADMUSHROOM_LUMP] = true;
				break;

			case 44:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 43:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 22:
				SpawnSlug(x, y);
				lumpneeded[SLUG_LUMP] = true;
				break;

			case 25:
				RF_SetScrollBlock(x, y, 1);
				break;

			case 26:
				RF_SetScrollBlock(x, y, 0);
				break;

			case 27:
			case 28:
			case 29:
			case 30:
				SpawnPlatform(x, y, info-27);
				lumpneeded[PLATFORM_LUMP] = true;;
				break;

			case 32:
				SpawnLoosePlatform(x, y);
				lumpneeded[PLATFORM_LUMP] = true;
				break;

			case 33:
				SpawnJanitor(x, y);
				lumpneeded[JANITOR_LUMP] = true;
				break;

			case 34:
				if (gamestate.ammo < 5)
				{
					SpawnBonus(x, y, 11);
					lumpneeded[bonuslumps[11]] = true;
				}
				break;

			case 35:
				SpawnWetsuit(x, y);
				CA_MarkGrChunk(WETSUITSPR);
				break;

			case 38:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 37:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 36:
				SpawnBadoing(x, y);
				lumpneeded[BADOING_LUMP] = true;
				break;

			case 41:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 40:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 39:
				SpawnWitchDoctor(x, y);
				lumpneeded[WITCHDOC_LUMP] = true;
				break;

			case 42:
				SpawnScubaKeen(x, y);
				SpawnScore();
				lumpneeded[SCUBAKEEN_LUMP] = true;
				//mark pickup shapes:
				for (i=BONUS100SPR; i<=BONUSAMMOSPR; i++)
				{
					CA_MarkGrChunk(i);
				}
				CA_MarkGrChunk(SCOREBOXSPR);
				break;

			case 83:
			case 84:
			case 85:
			case 86:
				if (gamestate.difficulty < gd_Hard)
					break;
				SpawnDartShooter(x, y, info-83);
				lumpneeded[DARTS_LUMP] = true;
				break;

			case 79:
			case 80:
			case 81:
			case 82:
				if (gamestate.difficulty < gd_Normal)
					break;
				SpawnDartShooter(x, y, info-79);
				lumpneeded[DARTS_LUMP] = true;
				break;

			case 53:
			case 54:
			case 55:
			case 56:
				SpawnDartShooter(x, y, info-53);
				lumpneeded[DARTS_LUMP] = true;
				break;

			case 57:
			case 58:
			case 59:
			case 60:
			case 61:
			case 62:
			case 63:
			case 64:
			case 65:
			case 66:
			case 67:
			case 68:
				SpawnBonus(x, y, info-57);
				lumpneeded[bonuslumps[info-57]] = true;
				break;

			case 69:
			case 70:
			case 71:
			case 72:
				SpawnMine(x, y, info-69);
				lumpneeded[MINE_LUMP] = true;
				break;

			case 75:
				lumpneeded[MOON_LUMP] = true;
				break;

			case 78:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 77:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 76:
				SpawnEagle(x, y);
				lumpneeded[EAGLE_LUMP] = true;
				break;

			case 107:
				if (gamestate.difficulty < gd_Hard)
					break;
			case 106:
				if (gamestate.difficulty < gd_Normal)
					break;
			case 89:
				SpawnStingaru(x, y);
				lumpneeded[STINGARU_LUMP] = true;
				break;
			}
		}
	}

	for (ob = player; ob; ob = ob->next)
	{
		if (ob->active != allways)
			ob->active = no;
	}

	for (i = 0; i < NUMLUMPS; i++)
	{
		if (lumpneeded[i])
		{
			for (chunk = lumpstart[i]; chunk <= lumpend[i]; chunk++)
			{
				CA_MarkGrChunk(chunk);
			}
		}
	}
}

//============================================================================

void CenterText(char *text)
{
	Uint16 w, h;
	MeasureTextBlock(text, &w, &h);
	PrintY = WindowY + (signed)(WindowH-h)/2;
	US_CPrint(text);
}
#define US_CPrint CenterText	// I was too lazy to copy and paste...

//============================================================================

char *lindseyText1 =
	"Commander Keen!\n"
	"I'm afraid I went\n"
        "looking for the scouts\n"
        "and got into a bit of\n"
        "a pinch. Thank you\n"
        "for rescuing me!";

char *keenText1 =
	"Any time, your highness!\n"
	"Now, let's find the rest\n"
	"of those scouts!";

char *lindseyText2 =
	"The Oracle Janitor has\n"
	"gone to find the scouts.\n"
	"There's trouble afoot!";

char *keenText2 =
	"Don't worry, princess,\n"
	"I'll find them!";

void LindseyDialog(void)
{
	SD_WaitSoundDone();
	StopMusic();
	CA_UpLevel();
	CA_ClearMarks();
	CA_MarkGrChunk(LINDSEYPIC);
	CA_MarkGrChunk(KEENTALKPIC);
	CA_MarkGrChunk(KEENTHUMBPIC);
	CA_CacheMarks(NULL);
	VW_FixRefreshBuffer();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX, WindowY, LINDSEYPIC);
//	PrintY += 3;
	WindowW -= 48;
	WindowX += 48;
	if (mapon == 12)
	{
		US_CPrint(lindseyText1);
	}
	else
	{
		US_CPrint(lindseyText2);
	}
	VW_UpdateScreen();
	SD_PlaySound(SND_MAKEFOOT);
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
	WindowW -= 48;
//	PrintY += 20;
	if (mapon == 12)
	{
		US_CPrint(keenText1);
	}
	else
	{
		US_CPrint(keenText2);
	}
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	VWB_DrawPic(WindowX+WindowW, WindowY, KEENTHUMBPIC);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	CA_DownLevel();
	//StartMusic(gamestate.mapon);	// level ends anyway...

	//reset scorebox (sprite may have been re-cached by CA_DownLevel)
	scoreobj->temp2 = -1;
	scoreobj->temp1 = -1;
	scoreobj->temp3 = -1;
	scoreobj->temp4 = -1;

	RF_NewPosition(originxglobal, originyglobal);	// force a refresh, but not right now
}

//============================================================================

char far janitorText1[] =
	"I've searched for those\n"
	"darn scouts for days!\n"
	"Finally someone has come\n"
	"to MY rescue!";

char far janitorText2[] =
	"I'll leave the rest\n"
	"to you, Captain, I'm\n"
	"going on vacation!";

char far janitorText3[] =
	"Now wait a minute,\n"
	"you mean I have to\n"
	"clean up YOUR mess\n"
	"by myself?";

char far janitorText4[] =
	"Here's my mop.";

void JanitorDialog(void)
{
	char str[200];

	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();
	CA_MarkGrChunk(JANITORPIC);
	CA_MarkGrChunk(KEENTALKPIC);
	CA_MarkGrChunk(KEENMADPIC);
	CA_CacheMarks(NULL);
	VW_FixRefreshBuffer();
	StartMusic(-1);
	
	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX, WindowY, JANITORPIC);
//	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	_fstrcpy(str, janitorText1);
	US_CPrint(str);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX, WindowY, JANITORPIC);
//	PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	_fstrcpy(str, janitorText2);
	US_CPrint(str);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
//	PrintY += 12;
	WindowW -= 48;
	_fstrcpy(str, janitorText3);
	US_CPrint(str);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX, WindowY, JANITORPIC);
//	PrintY += 26;
	WindowW -= 48;
	WindowX += 48;
	_fstrcpy(str, janitorText4);
	US_CPrint(str);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
	VWB_DrawPic(WindowX+WindowW-40, WindowY+24, KEENMADPIC);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	StopMusic();
	CA_DownLevel();
	StartMusic(gamestate.mapon);

	//BUG: scorebox needs to be reset here (sprite may have been re-cached by CA_DownLevel)
}

//============================================================================

void CantSwimMessage(void)
{
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_CacheGrChunk(KEENTALKPIC);

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
	WindowW -= 48;
	//PrintY += 12;
	US_CPrint(
		"I don't have\n"
		"my wetsuit!"
		);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();
	CA_DownLevel();

	//Note: scorebox sprite has NOT been re-cached here - we didn't use CA_CacheMarks.
}

//============================================================================

void WetsuitMessage(void)
{
	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();
	CA_MarkGrChunk(KEENTALKPIC);
	CA_MarkGrChunk(KEENTHUMBPIC);
	CA_CacheMarks(NULL);

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
	WindowW -= 48;
	//PrintY += 12;
	US_CPrint(
		"I wonder who leaves\n"
		"all these wetsuits\n"
		"lying around?"
		);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	VWB_DrawPic(WindowX+WindowW, WindowY, KEENTHUMBPIC);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	CA_DownLevel();

	//Note: scorebox sprite may have been re-cached by CA_DownLevel, but the level ends after this anyway
}

//============================================================================

void OracleDialog(void)
{
	register char *scouttext, *keentext;

	SD_WaitSoundDone();
	CA_UpLevel();
	CA_ClearMarks();
	CA_MarkGrChunk(ORACLEPIC);
	CA_MarkGrChunk(KEENTALKPIC);
	CA_MarkGrChunk(KEENTHUMBPIC);
	CA_CacheMarks(NULL);
	StartMusic(-1);
	VW_FixRefreshBuffer();

	switch (mapon)
	{
	case 3:
		scouttext = "Captain Keen!\nRemember me?\nIt's an honor to be\nsaved by you!";
		keentext = "You again?\nTry to stay\nout of trouble.";
		break;

	case 5:
		scouttext = "Oh Comrade Keen! Can I\nhave your autograph?";
		keentext = keentext = "You're even more\nobnoxious than\nthe last one...";
		break;

	case 8:
		scouttext = "You don't happen to\nhave any snacks with\nyou, do you, Captain?";
		keentext = "Meet the others\nback at the Megarocket\nfor some candy.";
		break;

	case 10:
		scouttext = "Help! I'm afraid\nof the dark!";
		keentext = "How did you get\nall the way down\nhere anyway!?";
		break;

	case 13:
		scouttext = "Did you see it, Captain?\nIt was THIS big!\nAlmost took my hand off!";
		keentext = "You should be more\ncareful, kid.";
		break;

	case 14:
		scouttext = "Aw schucks, I was\nhoping to be rescued\nby the princess!";
		keentext = "I could just\nleave you here...";
		break;

	case 16:
		scouttext = "Oh my, Corporal Keen,\nI'm unworthy!";
		keentext = "I can't disagree\nwith that!";
		break;

	default:	// level 17 message is also the default message - just to make sure our text pointers are initialized
		scouttext = "Commodore Keen!\nI'm so excited I don't\nknow what to do\nwith my hands.";
		keentext = "It's Comman-\noh forget it!";
		break;
	}

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX, WindowY, ORACLEPIC);
	//PrintY += 6;
	WindowW -= 48;
	WindowX += 48;
	US_CPrint(scouttext);
	VW_UpdateScreen();
	VW_WaitVBL(60);
	IN_ClearKeysDown();
	IN_Ack();

	US_CenterWindow(26, 8);
	VWB_DrawPic(WindowX+WindowW-48, WindowY, KEENTALKPIC);
	WindowW -= 48;
	//PrintY += 12;
	US_CPrint(keentext);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	VWB_DrawPic(WindowX+WindowW, WindowY, KEENTHUMBPIC);
	VW_UpdateScreen();
	VW_WaitVBL(30);
	IN_ClearKeysDown();
	IN_Ack();

	gamestate.rescued++;
	CA_DownLevel();
	StopMusic();

	//Note: scorebox sprite may have been re-cached by CA_DownLevel, but the level has already ended when this is executed
}

//============================================================================
//============================================================================

statetype s_scubakeen1    = {SCUBAKEENL1SPR, SCUBAKEENR1SPR, stepthink, false, push_none, 50, 0, 0, KeenScubaThink, KeenScubaContact, KeenScubaReact, &s_scubakeen2};
statetype s_scubakeen2    = {SCUBAKEENL2SPR, SCUBAKEENR2SPR, stepthink, false, push_none, 50, 0, 0, KeenScubaThink, KeenScubaContact, KeenScubaReact, &s_scubakeen1};
statetype s_scubakeenfast1 = {SCUBAKEENL1SPR, SCUBAKEENR1SPR, stepthink, false, push_none, 50, 0, 0, KeenScubaFastThink, KeenScubaContact, KeenScubaReact, &s_scubakeen2};	//never actually used
statetype s_scubakeenfast2 = {SCUBAKEENL2SPR, SCUBAKEENR2SPR, stepthink, false, push_none, 50, 0, 0, KeenScubaFastThink, KeenScubaContact, KeenScubaReact, &s_scubakeen1};	//never actually used
//Note: the die states for scubakeen are in CK_KEEN.C and K4_ACT3.C (dopefish section)

statetype s_smallbubble1  = {SMALLBUBBLE1SPR, SMALLBUBBLE1SPR, think, false, push_none, 20, 0, 24, BubbleThink, NULL, DrawReact, &s_smallbubble1};
statetype s_smallbubble2  = {SMALLBUBBLE2SPR, SMALLBUBBLE2SPR, think, false, push_none, 20, 0, 24, BubbleThink, NULL, DrawReact, &s_smallbubble2};
statetype s_smallbubble3  = {SMALLBUBBLE3SPR, SMALLBUBBLE3SPR, think, false, push_none, 20, 0, 24, BubbleThink, NULL, DrawReact, &s_smallbubble3};
statetype s_smallbubble4  = {SMALLBUBBLE4SPR, SMALLBUBBLE4SPR, think, false, push_none, 20, 0, 24, BubbleThink, NULL, DrawReact, &s_smallbubble4};

void SpawnScubaKeen(Sint16 x, Sint16 y)
{
	player->obclass = keenobj;
	player->active = allways;
	player->priority = 1;
	player->x = CONVERT_TILE_TO_GLOBAL(x);
	player->y = CONVERT_TILE_TO_GLOBAL(y);
	player->xdir = 1;
	player->ydir = 1;
	player->needtoclip = FULLCLIP;
	NewState(player, &s_scubakeen1);
}

void SpawnBubble(objtype *ob)
{
	ob->temp4 = 0;
	GetNewObj(true);
	if (ob->xdir == -1)
	{
		new->x = ob->x;
	}
	else
	{
		new->x = ob->x + 24*PIXGLOBAL;
	}
	new->y = ob->y;
	new->obclass = decoobj;
	new->priority = 3;
	new->active = removable;
	new->needtoclip = false;
	new->yspeed = -24;
	new->xspeed = 4;
	switch (US_RndT() / 64)
	{
	case 0:
		NewState(new, &s_smallbubble1);
		break;
	case 1:
		NewState(new, &s_smallbubble2);
		break;
	case 2:
		NewState(new, &s_smallbubble3);
		break;
	case 3:
		NewState(new, &s_smallbubble4);
		break;
	}
	SD_PlaySound(SND_BLUB);
}

void KeenScubaThink(objtype *ob)
{
	Sint32 i;
	Sint16 vx, vy, xc, yc;

	xc = ob->xspeed < 0;
	yc = ob->yspeed < 4;

	ob->temp4 = ob->temp4 + tics;
	if (ob->temp4 > 60)
		SpawnBubble(ob);

	if (button0down && !button0held)
	{
		button0held = true;
		if (c.xaxis)
			ob->xspeed = c.xaxis * 18;
		if (c.yaxis)
			ob->yspeed = c.yaxis * 18;
		ob->state = ob->state->nextstate;
	}
	if (c.xaxis)
		ob->xdir = c.xaxis;

	for (i = lasttimecount-tics; i < lasttimecount; i++)
	{
		if ((i & 7) == 0)
		{
			if (ob->xspeed > 12)
			{
				vx = -3;
			}
			else if (ob->xspeed > 0)
			{
				vx = -1;
			}
			else if (ob->xspeed > -12)
			{
				vx = 1;
			}
			else
			{
				vx = 3;
			}
			vx += c.xaxis;
			vx += c.xaxis;
			ob->xspeed += vx;

			if (c.xaxis == 0 && (ob->xspeed < 0) != xc)
				ob->xspeed = 0;

			if (ob->yspeed > 12)
			{
				vy = -3;
			}
			else if (ob->yspeed > 4)
			{
				vy = -1;
			}
			else if (ob->yspeed > -12)
			{
				vy = 1;
			}
			else
			{
				vy = 3;
			}
			vy += c.yaxis;
			vy += c.yaxis;
			ob->yspeed += vy;

			if (c.yaxis == 0 && ob->yspeed > 4 && yc)
				ob->yspeed = 0;
		}
		tmp_xmove += ob->xspeed;
		tmp_ymove += ob->yspeed;
	}
}

void KeenScubaFastThink(objtype *ob)	//never actually used
{
	ob->temp4 = ob->temp4 + tics;
	if (ob->temp4 > 60)
		SpawnBubble(ob);

	if (button0down && !button0held)
	{
		button0held = true;
		ob->xspeed = c.xaxis * 18;
		if (c.yaxis)
			ob->yspeed = c.yaxis * 18;

		if (ob->state == &s_scubakeenfast1)
		{
			ob->state = &s_scubakeenfast2;
		}
		else
		{
			ob->state = &s_scubakeenfast1;
		}
	}

	tmp_xmove = tmp_xmove + ob->xspeed * tics;
	tmp_ymove = tmp_ymove + ob->yspeed * tics;
	if (tmp_xmove > 0)
	{
		ob->xdir = 1;
	}
	else if (tmp_ymove < 0)
	{
		ob->xdir = -1;
	}

	tmp_ymove = tmp_ymove + tics*4;
}

void KeenScubaContact(objtype *ob, objtype *hit)
{
	switch (hit->obclass)
	{
	case bonusobj:
		switch (hit->temp1)
		{
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
			SD_PlaySound(bonussound[hit->temp1]);
			hit->obclass = decoobj;
			hit->priority = 3;
			hit->shapenum = pickupshape[hit->temp1];
			GivePoints(bonuspoints[hit->temp1]);
			if (hit->temp1 < 4)
			{
				gamestate.keys[hit->temp1]++;
			}
			else if (hit->temp4 == 10)
			{
				gamestate.lives++;
			}
			else if (hit->temp1 == 11)
			{
				gamestate.ammo += ammotable[gamestate.difficulty];
			}
			ChangeState(hit, &s_bonusrise);
			break;
		}
		break;
	case oracleobj:
		playstate = ex_rescued;
		break;
	}
	ob++;
}

void KeenScubaReact(objtype *ob)
{
	if (ob->hiteast && ob->xspeed < 0 || ob->hitwest && ob->xspeed > 0)
		ob->xspeed = 0;

	if (ob->hitnorth && ob->yspeed > 0 || ob->hitsouth && ob->yspeed < 0)
		ob->yspeed = 0;

	RF_PlaceSprite(&ob->sprite, ob->x, ob->y, ob->shapenum, 0, ob->priority);
}
